# qubicj-computor-api

Provides infrastracture to work with Qubic in Java projects. It contains code related to Qubic crypto operations (generate IDs, public key, private key, sign, verify) and for communicating with Qubic nodes (computors).

For communicating with nodes an asynchronous reactive implementation (using [project reactor](https://projectreactor.io/)) is provided, but you can use this implementation in blocking code, too.
The library has been programmed with spring dependency injection in mind, but you can use it without spring. For examples see below and the other projects in this repository.

Currently, this project is WIP and not all operations are added yet. Contact me, if you need certain functionality or a release version. Heavy refactoring might take place, 
but I am happy to assist with migration. If the documentation here is outdated have a look at the code.

1. [Build](#Build)
2. [Usage](#Usage)
3. [Installing shared library](#install-shared-library)

## Build

Before building or using you need to **install the dependencies** (external shared library), because not all relevant crypto functions are available in Java yet.
See section [Installing shared library](#install-shared-library) below. Then follow the [build instructions in the parent project](../README.md#build).

If you want to include the dependency in your own project then you do not need to build the library. Simply
add the dependency (see next section).

## Usage

### Add dependency

The project is hosted on maven central. Therefore, you can easily include it in your project via a dependency definition.
For maven add the following dependency to your `pom.xml`:

```xml
<dependency>
  <groupId>at.qubic</groupId>
  <artifactId>qubicj-computor-api</artifactId>
  <version>x.y.z</version>
</dependency>
```

For gradle add

```groovy
implementation 'at.qubic:qubicj-computor-api:x.y.z'
```

Replace `x.y.z` with the version number (take latest version if possible, see releases). 
You need the shared library dependency installed on your system, if you want to access qubic crypto functions.

### Examples

This section contains examples for qubic operations like retrieving the balance of an entity (ID) or sending a transaction. For more details have a look at the demo projects.
They are always up-to-date while this documentation here might be outdated soon.

As the API is WIP documentation might not be accurrate. It is recommended to have a look at the test classes for usage examples.

#### Wiring objects together

To create the objects you can either use spring config classes (see `SpringBootTestConfiguration` in tests for example) or simply use the constructors.

You decide how much framework you want to use. You can usa a single node, a pool of fixed nodes or a pool of all available public nodes.
If you want to use a single node do the following (most basic way):

```java
Node node = new Node(1.2.3.4);
// do some call (blocking)
TickInfo info = node.getCurrentTickInfo().block();
System.out.println("Latest tick: " + info.getTick());

// async version
node.getCurrentTickInfo()
    .subscribe(ti -> System.out.println("Latest tick: " + ti.getTick()));
```

A simple service calling a certain node could be created like this (it is preferable to use the constructor with more parameters):

```java
Nodes nodes = new Nodes();
nodes.addNode("1.2.3.4"); // node
nodes.addNode("1.2.3.5"); // node
nodes.addNode("1.2.3.6"); // node

ComputorService computorService = new ComputorService(nodes); // (optional) wrapper for more comfortable access
```

If you want to use all available public peers and not only your own nodes you can initialize the Nodes class like this:

```java
import at.qubic.api.network.*;

NetworkStatus networkStatus = new HealthCheckNetworkStatus(); // queries node health
NodesManagementStrategy strat = new PublicNodesStrategy(networkStatus); // uses most current (public) peers
Nodes nodes = new Nodes(strat);
nodes.addNode("1.2.3.4"); // add several seed nodes

ComputorService computorService = new ComputorService(nodes); // (optional) wrapper for more comfortable access
```

The public nodes strat needs some time to warm up until all nodes are retrieved and status checked. Therefore, the first request will take a bit longer.

If you need seed nodes and do not know how to get them you can have a look in the default qubicj-shell main-net 
[application properties](../qubicj-shell/src/main/resources/application-mainnet.properties) file (but this list might
be out of date) or use a public list like [this one](https://app.qubic.li/network/live). 

#### Query balance

Create a computor service and call the query balance method to retrieve entities. 

```java
computorService.queryBalance("MMPYLHEBCADPHFOXPTTSBRWJRNSBHEVFXCXEHRBIXFDTQCLJRFRKZGGFFQUM") // ID
        .map(entity -> entity.getIncomingAmount() - entity.getOutgoingAmount()) // calculate balance of entity
        .subscribe(balance -> log.info("Current balance: [{}].", balance));
```

The blocking version for retrieving the entity could look like this:

```java
Entity entity = computorService
        .queryBalance("MMPYLHEBCADPHFOXPTTSBRWJRNSBHEVFXCXEHRBIXFDTQCLJRFRKZGGFFQUM")
        .block();
```

#### Do transaction

```java
computorService.sendTransaction("abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabc", // seed
            "TESTIOYEZSAUQBUNKJEGXOILIBICZYWAMVSWYCYFHDCSZOAXPTGSOKFCCUNK", // destination ID
            1_000_000) // amount
        .subscribe(tx -> log.info("Sent transaction [{}]. Target tick: [{}].", tx.getHash(), tx.getTick()));
```

The blocking version for sending the transactionc could look like this:

```java
SignedTransaction tx = computorService.sendTransaction("abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabc",
            "TESTIOYEZSAUQBUNKJEGXOILIBICZYWAMVSWYCYFHDCSZOAXPTGSOKFCCUNK",
            1_000_000)
        .block();
```
 
#### Create a raw transaction

If you do not want to use the qubicj client but another API or your own implementation you might want to crate a raw transaction (binary). That's how:

```java
TransactionService transactionService = new TransactionService();
// int tick = ...
// String seed = "..."
// String destinationIdentity = "..."
// long amount = ...        
SignedTransaction transaction = transactionService.createTransaction(tick, seed, destinationIdentity, amount);
byte[] rawTransaction = transaction.toBytes();
```

If you need the binary representation not only of the transaction but the complete qubic message (including the header) for sending a transaction, you can create it like this:

```java
// byte[] signedTransaction = ...
QubicHeader qubicHeader = QubicHeader.builder()
    .type(BROADCAST_TRANSACTION)
    .dejavu(0)    
    .size(152) // header + transaction + signature
    .build();

QubicMessage message = QubicMessage.builder()
    .header(qubicHeader)
    .payload(signedTransaction) // transaction + signature
    .build();

byte[] rawQubicMessage = message.toBytes();
```

This binary version can be sent directly to a computor node.

#### Other operations

Check the `Node` and/or `ComputorService` classes for available calls. If calls are missing please open an issue.

## Install shared library

Unfortunately not all relevant crypto functions are available in Java yet. Therefore, we need to use a shared C library. 
All needed operations are implemented in the qubic crypto library.

### Qubic crypto lib

We use the shared library from [qubic-cli](https://github.com/qubic/qubic-cli) as default for the API code.

#### Installation Instructions

1. Clone [repository](https://github.com/qubic/qubic-cli)
2. Compile
3. Install shared library

For compilation, it's important to make a 'relase' build because that makes the code about 5 times faster. For example:

```shell
mkdir build
cd build
cmake -DCMAKE_BUILD_TYPE=Release ..
cmake --build .
```

Install the library

```shell
sudo cmake --install .
sudo ldconfig
```

XOR manually

```shell
sudo cp libfourq-qubic.so /usr/local/lib/
sudo ldconfig
```

#### Additional information

Qubic uses fourq. The sign and verify part of qubic is different from the [FourQ standard implementation](https://github.com/microsoft/FourQlib) as it uses
KangarooTwelve instead of SHA-512. For the vanity id generator the faster standard implementation is better suited as sign and verify is not needed.

## Known Issues

### Build error because of missing shared library

```
[ERROR]   TransactionServiceTest.<init>:25 » UnsatisfiedLink Unable to load library 'fourq-qubic':
libfourq-qubic.so: cannot open shared object file: No such file or directory
libfourq-qubic.so: cannot open shared object file: No such file or directory
Native library (linux-x86-64/libfourq-qubic.so) not found in resource path
```

This means that the shared library is not installed properly on your system and can't be found. There are different
reasons this can happen.

1. Make sure to run `ldconfig` after installation. To further debug check, if ldconfig finds the library:

    ```shell
    sudo ldconfig -p | grep qubic
    ```

2. Some systems do not include `/usr/local/lib` in their library path (for example Arch Linux). To heck if library path 
includes default installation directory run 

    ```shell
    cat /etc/ld.so.conf.d/* | grep "/usr/local"`. 
    ```
    
    If it is not included there you should install the shared library to another location. For example add a link from your
    local lib directory to the installed version in `/usr/local/lib` or modify the cmake installation directory 
    to install to `/usr/lib/` instead:
    
    ```shell
    cmake -DCMAKE_BUILD_TYPE=Release -DCMAKE_INSTALL_PREFIX=/usr ..
    cmake --build .
    cmake --install .
    ```
