package at.qubic.api.crypto;

import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import at.qubic.api.exception.InvalidIdentityException;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.function.BooleanSupplier;
import java.util.function.Supplier;
import java.util.regex.Pattern;

@Slf4j
public class IdentityUtil {

    private static final char UPPER_CASE_OFFSET = 'A';
    private static final char LOWER_CASE_OFFSET = 'a';
    private final boolean verify;
    private static final Pattern IDENTITY_PATTERN = Pattern.compile("[A-Z]{60}");
    private static final Pattern SEED_PATTERN = Pattern.compile("[a-z]{55}");

    private final FourQ fourQ;

    public IdentityUtil(boolean doVerification, FourQ fourQ) {
        this.verify = doVerification;
        this.fourQ = fourQ;
    }

    public IdentityUtil() {
        this(true, new QubicFourQ());
    }

    public byte[] getSubSeedFromSeed(String seed) {
        if (verify) verifyIdentity(() -> isValidSeed(seed), () -> String.format("Invalid seed: [%s].",seed));
        final byte[] seedArr = toSeedBytes(seed);
        return KangarooTwelve.hash(seedArr, 32);
    }

    public byte[] getPrivateKeyFromSubSeed(byte[] subSeed) {
        if (verify) verifyIdentity(() -> ArrayUtils.getLength(subSeed) == 32, () -> "Invalid sub seed length.");
        return KangarooTwelve.hash(subSeed, 32);
    }

    public byte[] getPublicKeyFromPrivateKey(byte[] privateKey) {
        if (verify) verifyIdentity(() -> ArrayUtils.getLength(privateKey) == 32, () -> "Invalid private key length.");
        return fourQ.ecc_mul_fixed(privateKey);
    }

    public byte[] getPublicKeyFromSeed(String seed) {
        return getPublicKeyFromSubSeed(getSubSeedFromSeed(seed));
    }

    public byte[] getPublicKeyFromSubSeed(byte[] subSeed) {
        return getPublicKeyFromPrivateKey(getPrivateKeyFromSubSeed(subSeed));
    }

    /**
     * <p>
     *     Public key is compressed version of identity. The public key is used to communicate with computors about IDs.
     *     Little endian.
     * </p>
     * <p>
     *     Chris D.: <cite>32 bytes long public key can be represented by 4 big integers, 8 bytes long each (4 * 8 = 32).
     *     Last 4 chars of ID string are checksum, so we are left with 56 chars to decode. We have 4 big ints, each one fits 56 / 4 = 14 chars.
     *     Think of each big int having 26 ^ 13 + 25 possible values (has more but not interesting in this case of encoding).
     *     Each time we decode one of the 14 chars we find the character's value (by subtracting A, so that A=0, B=1,..., Z=25),
     *     we multiply previous big int value (or 0) by 26 and add to it the value of the character
     *     (char value is also the result of big int % 26 + A, when encoding).</cite>
     * </p>
     * <p>
     *     TL;DR => 56 characters put into four 8 byte values (positive long is good enough).
     * </p>
     * @param identity The identity to convert.
     * @return The public key as a byte[32] array.
     */
    public byte[] getPublicKeyFromIdentity(String identity) {
        if (verify) verifyIdentity(() -> isValidIdentityString(identity), () -> String.format("Invalid identity: [%s]", identity));
        final byte[] publicKey = convertIdentityToPublicKey(identity);
        verifyIdentityChecksum(identity, publicKey);
        return publicKey;
    }

    public String getIdentityFromPublicKey(byte[] publicKey) {
        return getIdentityFromPublicKey(publicKey, UPPER_CASE_OFFSET);
    }

    public String getTransactionHash(byte[] signedTxDigest) {
        return getIdentityFromPublicKey(signedTxDigest, LOWER_CASE_OFFSET);
    }

    public boolean isValidSeed(String seed) {
        return SEED_PATTERN.matcher(StringUtils.defaultString(seed)).matches();
    }

    public boolean isValidIdentity(String identity) {
        if (isValidIdentityString(identity)) {
            try {
                final byte[] publicKey = convertIdentityToPublicKey(identity);
                return isValidChecksum(identity, publicKey);
            } catch (InvalidIdentityException iie) {
                log.warn(iie.getMessage());
                return false;
            }
        } else {
            return false;
        }
    }

    private static boolean isValidIdentityString(String identity) {
        return IDENTITY_PATTERN.matcher(StringUtils.defaultString(identity)).matches();
    }

    private static byte[] convertIdentityToPublicKey(String identity) {
        ByteBuffer buf = ByteBuffer.allocate(32);
        buf.order(ByteOrder.LITTLE_ENDIAN);

        long[] publicKeyLongs = new long[4];
        for (int i = 0; i < 4; i++) {

            for (int j = 14; j-- > 0;) {
                final char character = identity.charAt(i * 14 + j);
                if (character < UPPER_CASE_OFFSET || character > 'Z') {
                    throw new InvalidIdentityException("Only characters A-Z are allowed.");
                }
                publicKeyLongs[i] = publicKeyLongs[i] * 26 + (character - UPPER_CASE_OFFSET);
            }
            buf.putLong(publicKeyLongs[i]);
        }

        return buf.array();
    }

    private String getIdentityFromPublicKey(byte[] publicKey, char offset) {
        if (verify) verifyIdentity(() -> ArrayUtils.getLength(publicKey) == 32, () -> "Invalid public key length.");

        StringBuilder identity = new StringBuilder();
        ByteBuffer publicKeyBuf = ByteBuffer.wrap(publicKey);
        publicKeyBuf.order(ByteOrder.LITTLE_ENDIAN);

        for (int i = 0; i < 4; i++) {
            // we need unsigned longs, but they are not available in java natively. negative longs won't work.
            // we can either use the Long.remainderUnsigned/divideUnsigned methods or BigInteger, but it's
            // ugly to go via string constructor new BigInteger(Long.toUnsignedString(publicKeyBuf.getLong()))
            long publicKeyFragment = publicKeyBuf.getLong();

            for (int j = 0; j < 14; j++) {
                long character = Long.remainderUnsigned(publicKeyFragment, 26);
                identity.append((char) (character + offset));
                publicKeyFragment = Long.divideUnsigned(publicKeyFragment, 26);
            }
        }

        publicKeyBuf.clear();

        identity.append(getChecksum(publicKey, offset));
        return identity.toString();
    }

    private static String getPublicKeyChecksum(final byte[] publicKey) {
        return getChecksum(publicKey, UPPER_CASE_OFFSET);
    }

    private static String getChecksum(final byte[] publicKey, char offset) {
        final byte[] checksumBytes = KangarooTwelve.hash(publicKey, 4);
        int checksumInt = (fromByteArray(checksumBytes) & 0x3FFFF);
        log.trace("Checksum int value [{}]. Bits: [{}].", checksumInt, Integer.toBinaryString(checksumInt));
        // four characters are stored in the integer
        StringBuilder checksum = new StringBuilder();
        for (int i = 0; i < 4; i++) {
            checksum.append((char) ((checksumInt % 26) + offset)); // decode
            checksumInt /= 26; // next character
        }
        return checksum.toString();
    }

    private static boolean isValidChecksum(String identity, byte[] publicKey) {
        final String checksum = getPublicKeyChecksum(publicKey);
        boolean valid = checksum.equals(identity.substring(56));
        if (!valid) {
            log.warn("Identity [{}] has invalid checksum. Expected [{}]", identity, checksum);
        }
        return valid;
    }

    private static void verifyIdentityChecksum(String identity, byte[] publicKey) {
        verifyIdentity(() -> isValidChecksum(identity, publicKey),
                () -> String.format("Checksum of identity [%s] is not correct. Expected [%s] but was [%s].",
                        identity, getPublicKeyChecksum(publicKey), StringUtils.substring(identity, 56)));
    }

    private static byte[] toSeedBytes(String seed) {
        byte[] seedArr = new byte[seed.length()];
        for (int i = 0; i < seedArr.length; i++) {
            seedArr[i] = (byte) (seed.charAt(i) - LOWER_CASE_OFFSET);
        }
        return seedArr;
    }

    private static void verifyIdentity(BooleanSupplier condition, Supplier<String> messageSupplier) {
        if (!condition.getAsBoolean()) {
            String msg = messageSupplier.get();
            log.error(msg);
            throw new InvalidIdentityException(msg);
        }
    }

    private static int fromByteArray(byte[] bytes) {
        // packing an array of 4 bytes to an int, little endian,
        return ((bytes[3] & 0xFF) << 24) |
                ((bytes[2] & 0xFF) << 16) |
                ((bytes[1] & 0xFF) << 8 ) |
                ((bytes[0] & 0xFF));
    }

}
