package at.qubic.api.domain;

import lombok.Builder;
import lombok.EqualsAndHashCode;
import lombok.Getter;
import lombok.ToString;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.codec.binary.Hex;
import org.apache.commons.lang3.SerializationException;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.concurrent.ThreadLocalRandom;

@Slf4j
@Builder
@Getter
@EqualsAndHashCode
@ToString
public class QubicHeader {

    public static final int HEADER_LENGTH = 8; // header is 8 bytes without payload

    @Builder.Default
    private final int size = HEADER_LENGTH; // byte[3]
    private final MessageType type; // byte
    @Builder.Default
    private final int dejavu = Math.abs(ThreadLocalRandom.current().nextInt()); // unsigned int

    public static QubicHeader fromBytes(ByteBuffer buf) {
        try {
            if (log.isTraceEnabled()) {
                byte[] headerBytes = new byte[HEADER_LENGTH];
                int position = buf.position();
                buf.get(headerBytes);
                log.trace("Header bytes: {}", Hex.encodeHexString(headerBytes));
                buf.position(position);
            }
            byte[] size = new byte[3];
            buf.get(size);
            return QubicHeader.builder()
                    .size(bytesToSize(size))
                    .type(MessageType.fromCode(buf.get()))
                    .dejavu(buf.getInt())
                    .build();
        } catch (RuntimeException re) {
            String message = "Error parsing message header.";
            log.warn(message);
            throw new SerializationException(message, re);
        }
    }

    public ByteBuffer toBytes(int capacity) {
        ByteBuffer buf = ByteBuffer.allocate(capacity);
        buf.put(sizeToBytes(getSize()));
        buf.put(getType().getCode());
        buf.putInt(getDejavu());
        return buf;
    }

    private static int bytesToSize(byte[] sizeArr) {
        assert sizeArr.length == 3;
        return ByteBuffer.allocate(4)
                .order(ByteOrder.LITTLE_ENDIAN)
                .put(sizeArr) // put 3 bytes, last one is always zero
                .rewind() // go to beginning
                .getInt(); // return 4 bytes int
    }

    static byte[] sizeToBytes(int intSize) {
        if (intSize < 0) {
            String msg = String.format("Only positive values up to 3 bytes are supported. Value: [%d].", intSize);
            log.error(msg);
            throw new IllegalArgumentException(msg);
        }
        byte[] size = new byte[3];
        ByteBuffer buf = ByteBuffer.allocate(4);
        buf.order(ByteOrder.LITTLE_ENDIAN); // stored as little endian
        buf.putInt(intSize);
        buf.get(0, size, 0, 3); // only take first three bytes
        return size;
    }

}
