package at.qubic.api.domain;

import lombok.Builder;
import lombok.EqualsAndHashCode;
import lombok.Getter;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.codec.binary.Hex;

import java.nio.ByteBuffer;

/**
 * A qubic message can be used as response and request. It consists out of a header and it's
 * payload. In case of a response the payload needs to be deserialized according to the header
 * information (type and size).
 */
@Slf4j
@Builder
@Getter
@EqualsAndHashCode
public class QubicMessage {

    private final QubicHeader header;

    @Builder.Default
    private final byte[] payload = new byte[0];

    public byte[] toBytes() {
        return header.toBytes(header.getSize())
                .put(payload)
                .array();
    }

    public static QubicMessage fromBytes(ByteBuffer buf) {
        QubicHeader header = QubicHeader.fromBytes(buf);
        log.debug("Read header: {}.", header.getType());

        final int payloadSize = header.getSize() - QubicHeader.HEADER_LENGTH;
        if (payloadSize > buf.remaining()) {
            String message = String.format("Unprocessable byte buffer. Expected size: [%s]. Remaining bytes: [%s]", header.getSize(), buf.remaining());
            log.error(message);
            throw new IllegalStateException(message);
        } else {
            log.debug("Received payload with [{}] bytes.", payloadSize);
        }
        byte[] payload = new byte[payloadSize];
        if (payloadSize > 0) {
            buf.get(payload);
            log.trace("Payload (bytes): {}", payload);
            log.trace("Payload (hex): {}", Hex.encodeHexString(payload));
        } else if (header.getType() != MessageType.END_RESPONSE) { // we don't want to warn about end responses
            log.warn("Payload is empty. Header: {}.", header);
        }
        return QubicMessage.builder()
                .header(header)
                .payload(payload)
                .build();
    }

}
