package at.qubic.api.domain.qx.response;

import lombok.Builder;
import lombok.EqualsAndHashCode;
import lombok.Getter;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.codec.binary.Hex;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import at.qubic.api.util.BufferUtil;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

@Slf4j
@Builder
@Getter
@EqualsAndHashCode
public class EntityOrder {

    public static final int ORDER_COUNT = 256;
    public static final int ORDER_SIZE_BYTES = 56; // 32 + 8 + 8 + 8

    private final byte[] issuer; // uint8_t[32] entity public key
    private final byte[] assetName; // uint64_t assetName;
    private final long price; // long long
    private final long numberOfShares; // long long

    // qx call responds with max 256 asset orders
    public static List<EntityOrder> getEntityOrders(byte[] input) {
        ByteBuffer buffer = ByteBuffer.wrap(input);
        buffer.order(ByteOrder.LITTLE_ENDIAN);
        List<EntityOrder> entityOrders = new ArrayList<>(ORDER_COUNT);
        EntityOrder entityOrder = EntityOrder.fromBytes(buffer);
        // skip empty orders at end
        while (!entityOrder.isEmpty() && entityOrders.size() < ORDER_COUNT-1) {
            entityOrders.add(entityOrder);
            entityOrder = EntityOrder.fromBytes(buffer);
        }
        return entityOrders;
    }

    private static EntityOrder fromBytes(ByteBuffer buffer) {
        return EntityOrder.builder()
                .issuer(BufferUtil.getByteArray(buffer, 32))
                .assetName(BufferUtil.getByteArray(buffer, 8))
                .price(buffer.getLong())
                .numberOfShares(buffer.getLong())
                .build();
    }

    public boolean isEmpty() {
        return price == 0 || numberOfShares == 0
                || ArrayUtils.isEmpty(assetName) || Arrays.equals(assetName, new byte[32]);
    }

    @Override
    public String toString() {
        return "EntityOrder{" +
                "issuer=" + (issuer == null ? null : Hex.encodeHexString(issuer)) +
                ", assetName=" + getAssetNameString() +
                ", price=" + price +
                ", numberOfShares=" + numberOfShares +
                '}';
    }

    public String getAssetNameString() {
        return assetName == null ? StringUtils.EMPTY : new String(assetName, StandardCharsets.US_ASCII).trim();
    }

}
