package at.qubic.api.domain.std;

import lombok.Builder;
import lombok.Getter;
import lombok.experimental.SuperBuilder;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.codec.binary.Hex;
import org.apache.commons.lang3.builder.ToStringBuilder;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;

@Slf4j
@Getter
@SuperBuilder
public class Transaction {

    public static final int TRANSACTION_BYTES_LENGTH = 80;

    @Builder.Default
    private final byte[] sourcePublicKey = new byte[32]; // char[32] => byte[32]
    @Builder.Default
    private final byte[] destinationPublicKey = new byte[32]; // char[32] => byte[32]
    private final long amount; // 8 bytes
    private final int tick; // 4 bytes
    @Builder.Default
    private final short inputType = 0; // 2 bytes
    @Builder.Default
    private final short inputSize = 0; // 2 bytes
    @Builder.Default
    private final byte[] extraData = new byte[0]; // optional byte[inputSize]

    protected ByteBuffer toByteBuffer(int capacity) {
        ByteBuffer buf = ByteBuffer.allocate(capacity);
        buf.order(ByteOrder.LITTLE_ENDIAN);
        buf.put(sourcePublicKey);
        buf.put(destinationPublicKey);
        buf.putLong(amount);
        buf.putInt(tick);
        buf.putShort(inputType);
        buf.putShort(inputSize);
        assert inputSize == extraData.length;
        if (inputSize > 0) {
            //noinspection ConstantValue // extra data set by builder
            if (inputSize != extraData.length) {
                log.error("Input size mismatch. Expected [{}] extra data bytes but got [{}].", inputSize, extraData.length);
            }
            buf.put(extraData);
        }
        return buf;
    }

    public byte[] toBytes() {
        return toByteBuffer(TRANSACTION_BYTES_LENGTH + inputSize)
                .array();
    }

    protected static Transaction fromByteBuffer(ByteBuffer buf, boolean signed) {
        // if signed 64 bytes signature are at end
        int signatureSize = signed ? 64 : 0;
        Transaction tx = Transaction.builder()
                .sourcePublicKey(getBytesArray(buf, 32))
                .destinationPublicKey(getBytesArray(buf, 32))
                .amount(buf.getLong())
                .tick(buf.getInt())
                .inputType(buf.getShort())
                .inputSize(buf.getShort())
                .extraData(getBytesArray(buf, buf.remaining() - signatureSize))
                .build();
        assert buf.remaining() == signatureSize;
        return tx;
    }

    protected int getBytesLength() {
        return TRANSACTION_BYTES_LENGTH + inputSize;
    }

    public String toString() {
        return new ToStringBuilder(this)
                .append("sourcePublicKey", Hex.encodeHexString(sourcePublicKey))
                .append("destinationPublicKey", Hex.encodeHexString(destinationPublicKey))
                .append("amount", amount)
                .append("tick", tick)
                .append("inputType", inputType)
                .append("inputSize", inputSize)
                .build();
    }

    private static byte[] getBytesArray(ByteBuffer buf, int size) {
        byte[] arr = new byte[size];
        buf.get(arr);
        return arr;
    }

}
