package at.qubic.api.domain.std.request;

import at.qubic.api.exception.ValidationException;
import at.qubic.api.util.AssetUtil;
import lombok.Builder;
import lombok.Getter;
import org.apache.commons.lang3.StringUtils;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;

import static at.qubic.api.domain.std.request.RequestAssets.RequestType.*;

@Getter
@Builder
public class RequestAssetsByFilter extends RequestAssets {

    // flags
    protected static final short NO_FLAG = 0;
    @SuppressWarnings("unused") // currently not supported
    protected static final short FLAG_ANY_SIBLINGS = 0b1;
    protected static final short FLAG_ANY_ISSUER = 0b10;
    protected static final short FLAG_ANY_ASSET_NAME = 0b100;
    protected static final short FLAG_ANY_OWNER = 0b1000;
    protected static final short FLAG_ANY_OWNERSHIP_MANAGING_CONTRACT = 0b10000;
    protected static final short FLAG_ANY_POSESSOR = 0b100000;
    protected static final short FLAG_ANY_POSSESSION_MANAGING_CONTRACT = 0b1000000;

    protected final RequestType assetRequestType; // unsigned short // 2b
    // protected final short flags; // unsigned short // 2b
    protected final short ownershipManagingContract; // unsigned short // 2b
    protected final short possessionManagingContract; // unsigned short // 2b

    protected final byte[] issuer; // unsigned char [32]
    protected final String assetName; // long long => long or byte[8] (7 characters zero terminated)
    protected final byte[] owner; // unsigned char [32]
    protected final byte[] possessor; // unsigned char [32]

    @Override
    protected byte[] getMessagePayload() {

        short flags;
        if (assetRequestType == REQUEST_TYPE_ISSUANCE_RECORDS) {
            flags = (short) (getIssuerFlag() | getAssetNameFlag());
        } else {
            validateNonNullProperty(assetName, "assetName");
            validateNonNullProperty(issuer, "issuer");
            flags = (short) (getOwnerFlag() |
                    getPossessorFlag() |
                    getOwnershipManagingContractFlag() |
                    getPossessionManagingContractFlag());
        }

        return ByteBuffer.allocate(SIZE)
                .order(ByteOrder.LITTLE_ENDIAN)
                .putShort(assetRequestType.asShort())
                .putShort(flags)
                .putShort(ownershipManagingContract)
                .putShort(possessionManagingContract)
                .put(issuer == null ? new byte[32] : issuer)
                .put(StringUtils.isBlank(assetName) ? new byte[8] : AssetUtil.nameTo8Bytes(assetName))
                .put(owner == null ? new byte[32] : owner)
                .put(possessor == null ? new byte[32] : possessor)
                .array();
    }

    protected void validateNonNullProperty(Object notNull, String propertyName) {
        if (notNull == null) {
            throw new ValidationException(String.format("Property [%s] must not be null.", propertyName));
        }
    }

    protected short getIssuerFlag() {
        return issuer == null ? FLAG_ANY_ISSUER : NO_FLAG;
    }

    protected short getAssetNameFlag() {
        return assetName == null ? FLAG_ANY_ASSET_NAME : NO_FLAG;
    }


    protected short getOwnerFlag() {
        return owner == null ? FLAG_ANY_OWNER : NO_FLAG;
    }

    protected short getOwnershipManagingContractFlag() {
        return ownershipManagingContract == 0 ? FLAG_ANY_OWNERSHIP_MANAGING_CONTRACT : NO_FLAG;
    }

    protected short getPossessorFlag() {
        return possessor == null || assetRequestType != REQUEST_TYPE_POSSESSION_RECORDS ? FLAG_ANY_POSESSOR : NO_FLAG;
    }

    protected short getPossessionManagingContractFlag() {
        return possessionManagingContract == 0 || assetRequestType != REQUEST_TYPE_POSSESSION_RECORDS ? FLAG_ANY_POSSESSION_MANAGING_CONTRACT : NO_FLAG;
    }

}
