package at.qubic.api.domain.std.response;

import lombok.Builder;
import lombok.EqualsAndHashCode;
import lombok.Getter;
import org.apache.commons.codec.binary.Hex;
import org.apache.commons.lang3.builder.ToStringBuilder;
import org.apache.commons.lang3.builder.ToStringStyle;
import at.qubic.api.util.BufferUtil;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.Objects;

@Builder
@Getter
@EqualsAndHashCode
public class Entity {

    // FIXME use or output correct data types because of overflows
    //      (long long, unsigned int, ...)

    private final byte[] publicKey; // char[32] => byte[32]
    private final long incomingAmount; // long long - 8 bytes
    private final long outgoingAmount; // long long - 8 bytes
    private final int numberOfIncomingTransfers; // unsigned int - 4 bytes
    private final int numberOfOutgoingTransfers; // unsigned int - 4 bytes
    private final int latestIncomingTransferTick; // unsigned int - 4 bytes
    private final int latestOutgoingTransferTick; // unsigned int - 4 bytes
    private final int tick; // 4 bytes
    private final int spectrumIndex; // 4 bytes
    private final byte[][] spectrum; // 24 * 32 = 768 bytes

    public static Entity fromBytes(byte[] input) {
        ByteBuffer buf = ByteBuffer.wrap(input);
        buf.order(ByteOrder.LITTLE_ENDIAN);
        Entity entity = Entity.builder()
                .publicKey(BufferUtil.getByteArray(buf, 32))
                .incomingAmount(buf.getLong())
                .outgoingAmount(buf.getLong())
                .numberOfIncomingTransfers(buf.getInt())
                .numberOfOutgoingTransfers(buf.getInt())
                .latestIncomingTransferTick(buf.getInt())
                .latestOutgoingTransferTick(buf.getInt())
                .tick(buf.getInt())
                .spectrumIndex(buf.getInt())
                .spectrum(getSpectrum(buf))
                .build();
        assert buf.remaining() == 0;
        return entity;
    }

    public ByteBuffer toByteBuffer() {
        Objects.requireNonNull(publicKey);
        Objects.requireNonNull(spectrum);
        ByteBuffer buf = ByteBuffer.allocate(64 + 4 + 4 + spectrum.length * 32);
        buf.order(ByteOrder.LITTLE_ENDIAN);
        buf.put(publicKey);
        buf.putLong(incomingAmount);
        buf.putLong(outgoingAmount);
        buf.putInt(numberOfIncomingTransfers);
        buf.putInt(numberOfOutgoingTransfers);
        buf.putInt(latestIncomingTransferTick);
        buf.putInt(latestOutgoingTransferTick);
        buf.putInt(tick);
        buf.putInt(spectrumIndex);
        for (byte[] hash : spectrum) {
            buf.put(hash);
        }
        return buf;
    }

    private static byte[][] getSpectrum(ByteBuffer buf) {
        final byte[][] spectrum = new byte[24][32];
        for (byte[] bytes : spectrum) {
            buf.get(bytes);
        }
        return spectrum;
    }

    public String toString() {
        return new ToStringBuilder(this, ToStringStyle.SHORT_PREFIX_STYLE)
                .append("publicKey", publicKey == null ? null : Hex.encodeHexString(publicKey))
                .append("incomingAmount", Long.toUnsignedString(incomingAmount))
                .append("outgoingAmount", Long.toUnsignedString(outgoingAmount))
                .append("numberOfIncomingTransfers", Integer.toUnsignedLong(numberOfIncomingTransfers))
                .append("numberOfOutgoingTransfers", Integer.toUnsignedLong(numberOfOutgoingTransfers))
                .append("latestIncomingTransferTick", Integer.toUnsignedLong(latestIncomingTransferTick))
                .append("latestOutgoingTransferTick", Integer.toUnsignedLong(latestOutgoingTransferTick))
                .append("tick", Integer.toUnsignedLong(tick))
                .build();
    }

}
