package at.qubic.api.domain.std.response;

import lombok.Builder;
import lombok.EqualsAndHashCode;
import lombok.Getter;
import org.apache.commons.codec.binary.Hex;
import org.apache.commons.lang3.builder.ToStringBuilder;
import org.apache.commons.lang3.builder.ToStringStyle;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.time.Instant;
import java.time.OffsetDateTime;
import java.time.ZoneOffset;
import java.time.format.DateTimeFormatter;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

@Builder
@Getter
@EqualsAndHashCode
public class TickData {

    // NUMBER_OF_TRANSACTIONS_PER_TICK = 1024
    // SIGNATURE_SIZE = 64

    private final short computorIndex; // unsigned short // 2B
    private final short epoch; // unsigned short // 2B
    private final int tick; // unsigned int // 4B

    private final short millisecond; // unsigned short // 2B
    private final byte second; // unsigned char // 1B
    private final byte minute; // unsigned char // 1B
    private final byte hour; // unsigned char // 1B
    @Builder.Default
    private final byte day = 1; // unsigned char // 1B
    @Builder.Default
    private final byte month = 1; // unsigned char // 1B
    private final byte year; // unsigned char // 1B

    @Builder.Default
    private final byte[] timelock = new byte[32];  // byte[32] => unsigned char[32] // 32B

    @Builder.Default
    private final byte[][] transactionDigests = new byte[1024][32]; // byte[NUMBER_OF_TRANSACTIONS_PER_TICK][32] // 1024*32 = 32768B

    @Builder.Default
    private final long[] contractFees = new long[1024]; //  long[1024] => long long contractFees[1024] // 1024*8 = 8192B

    @Builder.Default
    private final byte[] signature = new byte[64]; // byte[64] => unsigned char signature[SIGNATURE_SIZE] // 64B

    public static TickData fromBytes(byte[] input) {
        ByteBuffer buf = ByteBuffer.wrap(input)
                .order(ByteOrder.LITTLE_ENDIAN);
        TickData tickData = TickData.builder()
                .computorIndex(buf.getShort())
                .epoch(buf.getShort())
                .tick(buf.getInt())
                .millisecond(buf.getShort())
                .second(buf.get())
                .minute(buf.get())
                .hour(buf.get())
                .day(buf.get())
                .month(buf.get())
                .year(buf.get())
                .timelock(getTimelock(buf))
                .transactionDigests(getTransactionDigests(buf))
                .contractFees(getContractFees(buf))
                .signature(getSignature(buf))
                .build();
        assert buf.remaining() == 0;
        return tickData;
    }

    private static byte[] getSignature(ByteBuffer buf) {
        byte[] signature = new byte[64];
        buf.get(signature);
        return signature;
    }

    private static byte[] getTimelock(ByteBuffer buf) {
        byte[] timelock = new byte[32];
        buf.get(timelock);
        return timelock;
    }

    private static byte[][] getTransactionDigests(ByteBuffer buf) {
        final byte[][] digests = new byte[1024][32];
        for (byte[] digest : digests) {
            buf.get(digest);
        }
        return digests;
    }

    private static long[] getContractFees(ByteBuffer buf) {
        final long[] contractFees = new long[1024];

        for (int i = 0; i < contractFees.length; i++) {
            contractFees[i] = buf.getLong();
        }
        // does not increment position: buf.asLongBuffer().get(contractFees);
        return contractFees;
    }

    private final static DateTimeFormatter FORMATTER = DateTimeFormatter.ofPattern("20uu-MM-dd'T'HH:mm:ss.nX");

    public Instant getTimestamp() {
        return OffsetDateTime.of(2000 + year, month, day, hour, minute, second, Short.toUnsignedInt(millisecond) * 1000 * 1000, ZoneOffset.UTC).toInstant();
    }

    public String toString() {
        return new ToStringBuilder(this, ToStringStyle.SHORT_PREFIX_STYLE)
                .append("computorIndex", computorIndex)
                .append("epoch", epoch)
                .append("tick", tick)
                .append("timestamp", getTimestamp())
                .append("transactionDigests", toHexStrings(transactionDigests))
                .append("signature", Hex.encodeHexString(signature))
                .build();
    }

    private List<String> toHexStrings(byte[][] digests) {
        return Arrays.stream(digests)
                .filter(this::notEmpty) // filter empty (not easy for byte[] in java therefore convert to ints first)
                .map(Hex::encodeHexString)
                .collect(Collectors.toList());
    }

    private boolean notEmpty(byte[] bytes) {
        for (byte b : bytes) {
            if (b != 0) {
                return true;
            }
        }
        return false;
    }

}
