package at.qubic.api.network;

import at.qubic.api.domain.QubicRequest;
import at.qubic.api.domain.MessageType;
import at.qubic.api.domain.QubicMessage;
import at.qubic.api.exception.InvalidResponseException;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.codec.binary.Hex;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;
import reactor.netty.Connection;

import java.net.ConnectException;
import java.net.NoRouteToHostException;
import java.nio.ByteBuffer;
import java.time.Duration;
import java.util.concurrent.TimeoutException;
import java.util.function.Predicate;

@Slf4j
public class BasicNodeOperations {

    /**
     * Send the request message and return the first response message after the peer exchange.
     *
     * @param node The node to use for the operation.
     * @param request The message to send.
     * @param messageType The type of the expected messages. Other messages are not to be expected and will be ignored.
     * @return The response message.
     */
    public Mono<QubicMessage> sendAndReceiveNext(Node node, Mono<? extends QubicRequest> request, MessageType messageType) {
        // for simple singular responses (typically without end message)
        return node.getClient().connect()
                .flatMap(conn -> exchangeMessage(node, request, conn)
                        .filter(isType(messageType))
                        .next()
                )
                .switchIfEmpty(Mono.defer(() -> Mono.error(new InvalidResponseException("Did not receive expected message of type: " + messageType))))
                .timeout(Duration.ofSeconds(Node.REQUEST_TIMEOUT))
                .doOnError(e -> logNodeCallError(node, e));
    }

    /**
     * Send the request message and return the response messages after the peer exchange.
     * In case there is no end response the caller needs to terminate the flux.
     *
     * @param node The node to use for the operation.
     * @param request The message to send.
     * @param type The type of the expected messages. The flux will terminate if another message is received.
     * @return The response messages (without the peer exchange and the optional end message).
     * @throws InvalidResponseException in case no end response is retrieved.
     */
    public Flux<QubicMessage> sendAndReceive(Node node, Mono<? extends QubicRequest> request, MessageType type) {
        Predicate<QubicMessage> takeWhilePredicate = type == null
                ? isType(MessageType.END_RESPONSE).negate()
                : isType(type);

        return node.getClient().connect()
                .flatMapMany(conn -> exchangeMessage(node, request, conn)
                        .takeUntil(isType(MessageType.END_RESPONSE))
                )
                .collectList()
                .flatMapMany(list -> {
                    // TODO replace quick and dirty workaround. There must be a better way.
                    //      find solution or maybe return Mono<List<QubicMessage>>
                    if (list.isEmpty() || list.getLast().getHeader().getType() != MessageType.END_RESPONSE) {
                        throw new InvalidResponseException("No end response received");
                    } else {
                        return Flux.fromIterable(list);
                    }
                })
                .doOnNext(qm -> log.debug("Received [{}].", qm.getHeader()))
                .filter(takeWhilePredicate)
                .timeout(Duration.ofSeconds(Node.REQUEST_TIMEOUT))
                .doOnError(e -> logNodeCallError(node, e));
    }

    /**
     * Send the request message but do not process a response.
     *
     * @param node The node to use for the operation.
     * @param requestMessage The message to send.
     * @return The connection for additional operations if needed.
     */
    public Mono<? extends Connection> send(Node node, Mono<QubicRequest> requestMessage) {
        final Mono<byte[]> requestBytes = requestMessage
                .doOnNext(msg -> log.debug("[{}] Sending message of type: [{}].", node.getName(), msg.getMessageType()))
                .map(req -> req.toMessage().toBytes())
                .doOnNext(bytes -> log.trace("Sending bytes: [{}]", Hex.encodeHexString(bytes)));
        return node.getClient().handle((in, out) -> out.sendByteArray(requestBytes)) // handle disposes connection
                .connect()
                .switchIfEmpty(Mono.defer(() -> Mono.error(new InvalidResponseException("Unexpected 'empty' result when sending to node: " + node.getName()))))
                .doOnError(e -> logNodeCallError(node, e));
    }

    private Flux<QubicMessage> exchangeMessage(Node node, Mono<? extends QubicRequest> request, Connection conn) {
        return exchangeMessages(conn, getRequestBytes(request))
                .map(this::readMessage)
                .doOnNext(node::updatePublicPeers) // peer exchange is processed by node
                .skipWhile(isType(MessageType.EXCHANGE_PUBLIC_PEERS));
    }

    private Mono<byte[]> getRequestBytes(Mono<? extends QubicRequest> request) {
        return convertToMessageBytes(request.doOnNext(r -> log.debug("Sending [{}].", r.toMessage().getHeader().getType())));
    }

    private void logNodeCallError(Node node, Throwable err) {
        if (err instanceof TimeoutException || err instanceof ConnectException || err instanceof NoRouteToHostException) {
            log.warn("Node [{}] call errored: {}", node.getName(), err.getMessage());
        } else {
            log.error("Node [{}] call errored.", node.getName(), err);
        }
    }

    private static Flux<ByteBuffer> exchangeMessages(final Connection conn, Mono<byte[]> requestBytes) {
        return conn.outbound()
                .sendByteArray(requestBytes)
                .then()
                .thenMany(conn.inbound()
                        .receive()
                        .asByteBuffer()
                        .doOnTerminate(conn::dispose) // we need to dispose here as the inner publisher completes
                );
    }

    private QubicMessage readMessage(ByteBuffer buf) {
        QubicMessage msg = QubicMessage.fromBytes(buf);
        log.trace("Payload: {}", msg.getPayload());

        if (buf.hasRemaining()) {
            byte[] remaining = new byte[buf.remaining()];
            log.warn("Received [{}] bytes more than expected. Message: {}, remaining: {}.",
                    buf.remaining(), msg, buf.get(remaining));
            assert !buf.hasRemaining();
        }
        if (isType(MessageType.END_RESPONSE).test(msg)) {
            log.debug("Received header with type [{}].", MessageType.END_RESPONSE);
        }
        return msg;
    }

    private Mono<byte[]> convertToMessageBytes(Mono<? extends QubicRequest> request) {
        return request.map(req -> req.toMessage().toBytes())
                .doOnNext(bytes -> log.trace("Sending bytes [{}]", Hex.encodeHexString(bytes)));
    }

    private static Predicate<QubicMessage> isType(MessageType message) {
        return qubicMessage -> qubicMessage.getHeader().getType() == message;
    }

}
