package at.qubic.api.network;

import lombok.extern.slf4j.Slf4j;
import at.qubic.api.domain.std.response.TickInfo;
import reactor.core.publisher.Mono;

import java.time.Instant;
import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;

@Slf4j
public class HealthCheckNetworkStatus implements NetworkStatus {

    private static final TickComparator TICK_COMPARATOR = new TickComparator();

    private Instant previousUpdate = Instant.EPOCH;
    private Instant latestUpdate = Instant.EPOCH;
    private TickInfo latestTick = TickInfo.builder()
            .epoch(INITIAL_EPOCH)
            .tick(INITIAL_TICK)
            .build();

    /**
     * Returns the latest known tick.
     * @param nodes nodes to ask for latest tick.
     * @return one of the most current tick infos queried from the nodes.
     */
    @Override
    public Mono<TickInfo> update(Collection<Node> nodes) {
        Iterable<Mono<TickInfo>> tickInfos = nodes.stream()
                .map(n -> Mono.defer(n::getCurrentTickInfo)
                        .defaultIfEmpty(TickInfo.builder().build()) // must not return empty
                        .onErrorReturn(TickInfo.builder().build())) // must not return error
                .collect(Collectors.toList());
        return Mono.zipDelayError(tickInfos, getLatestTickFunction()) // attention: does not support empty mono!
                .doFirst(this::setLatestUpdateTime) // avoid further updates
                .map(ti -> (TickInfo) ti) // cast
                .doOnNext(this::setLatestTick)
                .doOnSuccess(ti -> logStats(nodes))
                .doOnError(t -> log.error("Health check failed.", t)); // error should not happen.
    }

    @Override
    public Mono<TickStatus> getTickStatus() {
        return Mono.just(TickStatus.builder()
                .latestUpdate(latestUpdate)
                .tickInfo(latestTick)
                .build());
    }

    private void setLatestTick(TickInfo currentTick)  {
        log.info("Current tick: {}", currentTick);
        if (epoch(currentTick) >= epoch(latestTick) && tick(currentTick) >= tick(latestTick)) {
            this.latestUpdate = Instant.now();
            this.latestTick = currentTick;
            if (tick(currentTick) > tick(latestTick)) {
                log.info("Updated network status to tick [{}] at [{}].", tick(latestTick), latestUpdate);
            }
        } else {
            resetLatestUpdateTime();
            log.warn("Not updating epoch [{}] and tick [{}] because current tick info is older: {}",
                    epoch(latestTick), tick(latestTick), currentTick);
        }
    }

    protected void logStats(Collection<Node> nodes) {
        Map<Integer, SortedSet<String>> nodesAtTick = new TreeMap<>(Comparator.reverseOrder()); // sorted
        for (Node node : nodes) {
            final int tick = Objects.requireNonNull(node.getTickInfo()).getTick();
            nodesAtTick.computeIfAbsent(tick, x -> new TreeSet<>()); // sorted
            nodesAtTick.get(tick).add(Objects.requireNonNull(node.getName()));
        }
        int currentTick = this.latestTick.getTick();
        for (Integer i : nodesAtTick.keySet()) {
            final SortedSet<String> n = nodesAtTick.get(i);
            String message = String.format("[%d] of [%d] node(s) at tick [%d]: %s", n.size(), nodes.size(), i, n);
            if (currentTick <= i + 1) {
                log.info(message);
            } else {
                log.debug(message);
            }
        }
    }

    private static Function<Object[], Object> getLatestTickFunction() {
        return tis -> Arrays.stream(tis).max((o1, o2) -> TICK_COMPARATOR.compare((TickInfo) o1, (TickInfo) o2)).orElseGet(() -> TickInfo.builder().build());
    }

    private int tick(TickInfo tickInfo) {
        return tickInfo.getTick();
    }

    private short epoch(TickInfo tickInfo) {
        return tickInfo.getEpoch();
    }

    private void setLatestUpdateTime() {
        this.previousUpdate = this.latestUpdate;
        this.latestUpdate = Instant.now();
    }

    private void resetLatestUpdateTime() {
        this.latestUpdate = this.previousUpdate;
    }

}
