package at.qubic.api.network;

import at.qubic.api.domain.MessageType;
import at.qubic.api.domain.QubicMessage;
import at.qubic.api.domain.QubicRequest;
import at.qubic.api.domain.RequestContractFunction;
import at.qubic.api.domain.qearn.request.*;
import at.qubic.api.domain.qearn.response.*;
import at.qubic.api.domain.qx.request.QxGetAssetOrders;
import at.qubic.api.domain.qx.request.QxGetEntityOrders;
import at.qubic.api.domain.qx.request.QxGetFees;
import at.qubic.api.domain.qx.response.AssetOrder;
import at.qubic.api.domain.qx.response.EntityOrder;
import at.qubic.api.domain.qx.response.QxFees;
import at.qubic.api.domain.std.SignedTransaction;
import at.qubic.api.domain.std.request.*;
import at.qubic.api.domain.std.response.*;
import at.qubic.api.exception.InvalidResponseException;
import at.qubic.api.tcp.TcpClientFactory;
import io.github.resilience4j.circuitbreaker.CircuitBreaker;
import io.github.resilience4j.circuitbreaker.CircuitBreakerConfig;
import io.github.resilience4j.reactor.circuitbreaker.operator.CircuitBreakerOperator;
import lombok.AccessLevel;
import lombok.Getter;
import lombok.ToString;
import lombok.extern.slf4j.Slf4j;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;
import reactor.netty.Connection;
import reactor.netty.tcp.TcpClient;

import java.net.InetSocketAddress;
import java.time.Duration;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.function.Function;

@Slf4j
@ToString(onlyExplicitlyIncluded = true)
public class Node {

    public static final int REQUEST_TIMEOUT = 2; // TODO make configurable

    // Create a custom configuration for a CircuitBreaker // TODO make configurable
    private static final CircuitBreakerConfig CIRCUIT_BREAKER_CONFIG = CircuitBreakerConfig.custom()
            .failureRateThreshold(75)
            .slidingWindowSize(10)
            .minimumNumberOfCalls(4)
            .slowCallDurationThreshold(Duration.ofSeconds(1))
            .permittedNumberOfCallsInHalfOpenState(1)
            .waitDurationInOpenState(Duration.ofMinutes(5))
            .build();

    @Getter
    private Instant latestTickUpdate = Instant.EPOCH;
    @Getter
    private TickInfo tickInfo = TickInfo.builder().build();

    @Getter(value = AccessLevel.PACKAGE)
    private final TcpClient client;

    @Getter
    @ToString.Include
    private final String name;

    @Getter(AccessLevel.MODULE)
    private final List<PeerListener> peerListeners = Collections.synchronizedList(new ArrayList<>(2));

    private final CircuitBreaker circuitBreaker;

    // delegator
    private final BasicNodeOperations basicNodeOperations = new BasicNodeOperations();

    public Node(String host, int port) {
        this(TcpClientFactory.tcpClient(host, port));
    }

    protected Node(TcpClient tcpClient) {
        this.client = tcpClient;
        this.name = getHostName(tcpClient);
        this.circuitBreaker = CircuitBreaker.of(this.name, CIRCUIT_BREAKER_CONFIG);
    }

    // TODO test

    /**
     * Get the current tick info.
     * @return Information about the node's current tick.
     */
    public Mono<TickInfo> getCurrentTickInfo() {
        final Mono<QubicRequest> request = Mono.just(RequestCurrentTickInfo.builder().build());
        return sendAndReceiveNext(request, MessageType.RESPOND_CURRENT_TICK_INFO)
                .map(response -> TickInfo.fromBytes(response.getPayload()))
                .switchIfEmpty(Mono.defer(() -> invalidResponseMono(name, "get current tick info")))
                .doOnNext(ti -> log.info("[{}] Received tick info: {}", name, ti.getTick()))
                .doOnNext(this::updateTickInfo);
    }
    public Mono<TickData> getTickData(int tick) {
        final Mono<QubicRequest> request = Mono.just(RequestTickData.builder().tick(tick).build());
        return sendAndReceiveNext(request, MessageType.BROADCAST_FUTURE_TICK_DATA)
                .switchIfEmpty(Mono.defer(() -> invalidResponseMono(name, "get tick data")))
                .map(message -> TickData.fromBytes(message.getPayload()))
                .doOnNext(td -> log.debug("[{}] Got tick data for tick [{}]: [{}].", name, tick, td != null));
    }

    /**
     * Get the entity information (balance) for the specified public key. An entity is identified by
     * its 'ID' (textual representation of the public key).
     * @param publicKey The public key of the entity to get the information of.
     * @return The entity information. See {@link Entity}.
     */
    public Mono<Entity> getEntity(Mono<byte[]> publicKey) {
        final Mono<QubicRequest> requestMessage = publicKey.map(pk -> RequestEntity.builder().publicKey(pk).build());
        return sendAndReceiveNext(requestMessage, MessageType.RESPOND_ENTITY)
                .map(response -> Entity.fromBytes(response.getPayload()))
                .switchIfEmpty(Mono.defer(() -> invalidResponseMono(name, "get entity")))
                .doOnNext(e -> log.debug("[{}]: {}", name, e));
    }

    /**
     * Sends a signed transaction by broadcasting it to the node.
     * @param transaction The transaction to be sent. Make sure that the correct future tick is set.
     * @return The submitted transaction for further processing (submitting is fire and forget).
     */
    public Mono<SignedTransaction> broadcastTransaction(Mono<SignedTransaction> transaction) {
        return transaction
                .doOnNext(stx -> log.info("[{}] Sending transaction [{}].", name, stx.getTransactionHash()))
                .map(tx -> BroadcastTransaction.builder().transaction(tx).build())
                .flatMap(req -> send(Mono.just(req)).map(conn -> req.getTransaction()));
    }

    /**
     * Get all transactions of a tick.
     *
     *
     * @param tick The tick to get the transactions from.
     * @return A flux of signed transactions. The transaction hash will get calculated. Attention: will return empty
     *         flux and not error in case the computor node is not at the specified tick yet.
     */
    public Flux<SignedTransaction> getTickTransactions(int tick) {
        final Mono<QubicRequest> request = Mono.just(RequestTickTransactions.builder().tick(tick).build());
        return sendAndReceive(request, MessageType.BROADCAST_TRANSACTION)
                .map(message -> SignedTransaction.fromBytes(message.getPayload()));
    }

    public Mono<AssetIssuanceResponse> getIssuedAssetByUniverseIndex(int index) {
        // attention: node returns a list but by index there should only be one
        final Mono<QubicRequest> request = Mono.just(RequestAssetsByIndex.builder().universeIndex(index).build());
        return sendAndReceiveNext(request, MessageType.RESPOND_ASSETS)
                .map(response -> AssetIssuanceResponse.fromBytes(response.getPayload()))
                .doOnNext(r -> log.debug("[{}]: {}", name, r));
    }

    public Flux<AssetIssuanceResponse> getIssuedAssetsByFilter(RequestAssetsByFilter request) {
        assert request.getAssetRequestType() == RequestAssets.RequestType.REQUEST_TYPE_ISSUANCE_RECORDS;
        return sendAndReceive(Mono.just(request), MessageType.RESPOND_ASSETS)
                .map(response -> AssetIssuanceResponse.fromBytes(response.getPayload()))
                .doOnNext(r -> log.debug("[{}]: {}", name, r));
    }

    public Flux<AssetOwnershipResponse> getOwnedAssetsByFilter(RequestAssetsByFilter request) {
        assert request.getAssetRequestType() == RequestAssets.RequestType.REQUEST_TYPE_OWNERSHIP_RECORDS;
        return sendAndReceive(Mono.just(request), MessageType.RESPOND_ASSETS)
                .map(response -> AssetOwnershipResponse.fromBytes(response.getPayload()))
                .doOnNext(r -> log.debug("[{}]: {}", name, r));
    }

    public Flux<AssetPosessionResponse> getPossessedAssetsByFilter(RequestAssetsByFilter request) {
        assert request.getAssetRequestType() == RequestAssets.RequestType.REQUEST_TYPE_POSSESSION_RECORDS;
        return sendAndReceive(Mono.just(request), MessageType.RESPOND_ASSETS)
                .map(response -> AssetPosessionResponse.fromBytes(response.getPayload()))
                .doOnNext(r -> log.debug("[{}]: {}", name, r));
    }

    /**
     * Get all owned assets of an entity.
     * @param publicKey The public key of the entity to get the owned assets of.
     * @return The assets the entity owns.
     */
    public Flux<RespondOwnedAssets> getOwnedAssets(Mono<byte[]> publicKey) {
        final Mono<QubicRequest> requestMono = publicKey.map(pk -> RequestOwnedAssets.builder().publicKey(pk).build());
        return sendAndReceive(requestMono, MessageType.RESPOND_OWNED_ASSETS)
                .map(response -> RespondOwnedAssets.fromBytes(response.getPayload()))
                .doOnNext(r -> log.debug("[{}]: {}", name, r));
    }

    /**
     * Get all possessed assets of an entity.
     * @param publicKey The public key of the entity to get the possessed assets of.
     * @return The assets the entity possesses.
     */
    public Flux<RespondPossessedAssets> getPossessedAssets(Mono<byte[]> publicKey) {
        final Mono<QubicRequest> requestMono = publicKey.map(pk -> RequestPossessedAssets.builder().publicKey(pk).build());
        return sendAndReceive(requestMono, MessageType.RESPOND_POSSESSED_ASSETS)
                .map(response -> RespondPossessedAssets.fromBytes(response.getPayload()))
                .doOnNext(r -> log.debug("[{}]: {}", name, r));
    }

    /**
     * Get QX fees.
     * @return Fees for QX contract operations.
     */
    public Mono<QxFees> getQxFees() {
        return callContractFunction(new QxGetFees(), bytes -> QxFees.fromBytes(bytes));
    }

    /**
     * Get QX orders for one asset (order book).
     * @return list of orders of an asset. List can be empty but never <code>null</code>.
     */
    public Mono<List<AssetOrder>> getQxAssetOrders(QxGetAssetOrders request) {
        return callContractFunction(request, bytes -> AssetOrder.getAssetOrders(bytes));
    }

    /**
     * Get all QX orders for one entity.
     * @return list of orders for an entity. List can be empty but never <code>null</code>.
     */
    public Mono<List<EntityOrder>> getQxEntityOrders(QxGetEntityOrders request) {
        return callContractFunction(request, bytes -> EntityOrder.getEntityOrders(bytes));
    }

    public Mono<UserLockInfo> getQearnUserLockInfo(GetUserLockInfo request) {
        return callContractFunction(request, bytes -> UserLockInfo.fromBytes(bytes));
    }

    public Mono<UserLockStatus> getQearnUserLockStatus(GetUserLockStatus request) {
        return callContractFunction(request, bytes -> UserLockStatus.fromBytes(bytes));
    }

    public Mono<LockInfoForEpoch> getQearnLockInfoForEpoch(GetLockInfoForEpoch request) {
        return callContractFunction(request, bytes -> LockInfoForEpoch.fromBytes(bytes));
    }

    public Mono<StateOfRound> getQearnStateOfRound(GetStateOfRound request) {
        return callContractFunction(request, bytes -> StateOfRound.fromBytes(bytes));
    }

    public Mono<EndedStatus> getQearnEndedStatus(GetEndedStatus request) {
        return callContractFunction(request, bytes -> EndedStatus.fromBytes(bytes));
    }

    public Mono<StatsForEpoch> getQearnStatsForEpoch(GetStatsForEpoch request) {
        return callContractFunction(request, bytes -> StatsForEpoch.fromBytes(bytes));
    }

    public Mono<BurnedAndBoostedStats> getQearnBurnedAndBoostedStats(GetBurnedAndBoostedStats request) {
        return callContractFunction(request, bytes -> BurnedAndBoostedStats.fromBytes(bytes));
    }

    public Mono<BurnedAndBoostedStatsForEpoch> getQearnBurnedAndBoostedStatsForEpoch(GetBurnedAndBoostedStatsForEpoch request) {
        return callContractFunction(request, bytes -> BurnedAndBoostedStatsForEpoch.fromBytes(bytes));
    }

    public <I extends RequestContractFunction, O> Mono<O> callContractFunction(I request, Function<byte[], O> createFunction) {
        return sendAndReceiveNext(Mono.just(request), MessageType.RESPOND_CONTRACT_FUNCTION)
                .map(response -> createFunction.apply(response.getPayload()))
                .switchIfEmpty(Mono.defer(() -> invalidResponseMono(name, callContractErrorMessage(request)))) // probably not needed
                .onErrorMap(original -> invalidResponseException(name, callContractErrorMessage(request), original))
                .doOnNext(r -> log.debug("[{}]: {}", name, r));
    }

    private static InvalidResponseException invalidResponseException(String node, String details, Throwable cause) {
        String m = createErrorMessage(node, details);
        return new InvalidResponseException(m, cause);
    }

    private static <T> Mono<T> invalidResponseMono(String node, String details) {
        String m = createErrorMessage(node, details);
        return Mono.error(new InvalidResponseException(m));
    }

    private static <I extends RequestContractFunction> String callContractErrorMessage(I request) {
        return String.format(
                "calling contract [%d] function [%d]",
                request.getContractIndex(), request.getInputType());
    }

    private static String createErrorMessage(String node, String details) {
        String m = String.format("[%s] Received invalid response %s.", node, details);
        log.error(m);
        return m;
    }

    /**
     * Send the request message and return the first response message after the peer exchange.
     *
     * @param request The message to send.
     * @param messageType The type of the expected messages.
     * @return The response message.
     */
    @SuppressWarnings("SameParameterValue")
    private Mono<QubicMessage> sendAndReceiveNext(Mono<? extends QubicRequest> request, MessageType messageType) {
        return basicNodeOperations.sendAndReceiveNext(this, request, messageType)
                .transform(CircuitBreakerOperator.of(circuitBreaker));
    }

    /**
     * Send the request message and return the response messages after the peer exchange.
     * In case there is no end response the caller needs to terminate the flux.
     *
     * @param request The message to send.
     * @param type The type of the expected messages. The flux will terminate if another message is received.
     * @return The response messages (without the peer exchange and the optional end message).
     */
    private Flux<QubicMessage> sendAndReceive(Mono<? extends QubicRequest> request, MessageType type) {
        return basicNodeOperations.sendAndReceive(this, request, type)
                .transform(CircuitBreakerOperator.of(circuitBreaker));
    }

    /**
     * Send the request message but do not process a response.
     *
     * @param requestMessage The message to send.
     * @return The connection for additional operations if needed.
     */
    private Mono<? extends Connection> send(Mono<QubicRequest> requestMessage) {
        return basicNodeOperations.send(this, requestMessage)
                .transform(CircuitBreakerOperator.of(circuitBreaker));
    }

    private void updateTickInfo(TickInfo tickInfo) {
        this.latestTickUpdate = Instant.now();
        this.tickInfo = tickInfo;
        log.debug("[{}] Updating tick info to {}.", name, tickInfo);
    }

    void updatePublicPeers(QubicMessage qubicMessage) {
        if (MessageType.EXCHANGE_PUBLIC_PEERS == qubicMessage.getHeader().getType()) {
            final String[] peers = ExchangePublicPeers.fromBytes(qubicMessage.getPayload()).getPeers();
            peerListeners.forEach(pl -> pl.publishPeers(peers));
        } // else skip
    }

    public void addPeerListener(PeerListener peerListener) {
        this.peerListeners.add(peerListener);
    }

    private static String getHostName(TcpClient tcpClient) {
        Objects.requireNonNull(tcpClient);
        Objects.requireNonNull(tcpClient.configuration());
        return ((InetSocketAddress) tcpClient.configuration().remoteAddress().get()).getHostName();
    }

}
