package at.qubic.api.network;

import lombok.extern.slf4j.Slf4j;
import at.qubic.api.domain.std.response.TickInfo;
import at.qubic.api.properties.NetworkProperties;
import org.apache.commons.lang3.exception.ExceptionUtils;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;
import reactor.core.scheduler.Schedulers;

import java.time.Duration;
import java.time.Instant;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

@Slf4j
public abstract class NodesManagementAdapter implements NodesManagementStrategy {

    private static final int MAX_TICK_DELAY = 1;
    private static final int DEFAULT_PORT = 21841;
    private static final Duration REFRESH_INTERVAL = Duration.ofSeconds(60);

    protected final int port;
    protected final int maxTickDelay;
    protected final Duration refreshInterval;
    protected final Map<String, Node> nodes = new ConcurrentHashMap<>();
    protected final NetworkStatus networkStatus;

    public NodesManagementAdapter(NetworkStatus networkStatus) {
        this(networkStatus, new NetworkProperties(DEFAULT_PORT, MAX_TICK_DELAY, REFRESH_INTERVAL));
    }

    public NodesManagementAdapter(NetworkStatus networkStatus, NetworkProperties networkProperties) {
        this.networkStatus = networkStatus;
        log.info("Network status implementation: {}", networkStatus.getClass().getSimpleName());
        this.port = networkProperties.getPort();
        this.maxTickDelay = networkProperties.getMaxTickDelay();
        this.refreshInterval = networkProperties.getRefreshInterval();
        log.info("Using port [{}], max tick delay [{}], refresh interval [{}]", port, maxTickDelay, refreshInterval);
    }

    @Override
    public Node addNode(String host) {
        return nodes.computeIfAbsent(host, h -> newNode(h, port));
    }

    protected void addNode(Node node) {
        nodes.put(node.getName(), node);
    }

    @Override
    public Flux<Node> getNodes(int max) {
        return Mono.defer(networkStatus::getTickStatus)
                .filter(ts -> ts.isValid() && getCurrentNodes(ts.getTickInfo(), maxTickDelay).size() >= max)
                .doOnNext(this::updateNetworkStatusAsync) // everything ok if not empty
                .map(TickStatus::getTickInfo)
                .switchIfEmpty(Mono.defer(this::updateNetworkStatus)) // no valid network status
                .flatMapMany(ti -> getShuffledCurrentNodes(ti, maxTickDelay))
                .take(max)
                .doOnError(e -> log.error("Error getting nodes: {}", ExceptionUtils.getMessage(e)));
    }

    protected Node newNode(String host, int port) {
        log.info("Creating new node [{}].", host);
        return new Node(host, port);
    }

    protected Flux<Node> getShuffledCurrentNodes(TickInfo ti, int offset) {
        List<Node> currentNodes = getCurrentNodes(ti, offset);
        Collections.shuffle(currentNodes);
        return Flux.fromIterable(currentNodes);
    }

    protected List<Node> getCurrentNodes(TickInfo ti, int offset) {
        return nodes.values()
                .stream()
                .filter(n -> isAtTick(n.getTickInfo(), ti.getEpoch(),ti.getTick() - offset))
                .collect(Collectors.toList());
    }

    @Override
    public Mono<Instant> forceNetworkUpdate(Instant updateIfBefore) {
        return Mono.defer(networkStatus::getTickStatus)
                .map(TickStatus::getLatestUpdate)
                .flatMap(updatedAt -> {
                    if (updatedAt.isBefore(updateIfBefore)) {
                        // force update
                        return networkStatus.update(nodes.values())
                                .map(ti -> Instant.now());
                    } else {
                        // do nothing
                        return Mono.just(updatedAt);
                    }
                });
    }


    protected Mono<TickInfo> updateNetworkStatus() {
        return networkStatus.update(nodes.values())
                .doFirst(() -> log.info("Network status update..."))
                .doOnError(e -> log.error("Error updating network status: {}", ExceptionUtils.getMessage(e)));
    }

    protected void updateNetworkStatusAsync(TickStatus ts) {
        if (!ts.isValid() || ts.lastUpdatedBefore(Instant.now().minus(refreshInterval))) {
            networkStatus.update(nodes.values())
                    .doFirst(() -> log.info("Asynchronous network status update..."))
                    .subscribeOn(Schedulers.boundedElastic())
                    .subscribe(x -> {},
                            e -> log.error("Error doing async network status update: {}", e.getMessage()));
        }
    }

    private static boolean isAtTick(TickInfo tickInfo, short epoch, int tick) {
        // minimum is 1. zero is invalid tick.
        return tickInfo.getEpoch() >= Math.max(1, epoch) && tickInfo.getTick() >= Math.max(1, tick);
    }



}
