package at.qubic.api.service;

import at.qubic.api.domain.qearn.request.*;
import at.qubic.api.domain.qearn.response.*;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.codec.binary.Hex;
import org.apache.commons.lang3.StringUtils;
import at.qubic.api.crypto.IdentityUtil;
import at.qubic.api.domain.std.SignedTransaction;
import at.qubic.api.domain.std.Transaction;
import at.qubic.api.domain.qx.request.*;
import at.qubic.api.domain.qx.response.AssetOrder;
import at.qubic.api.domain.qx.response.EntityOrder;
import at.qubic.api.domain.qx.response.QxFees;
import at.qubic.api.domain.std.response.*;
import at.qubic.api.network.Node;
import at.qubic.api.network.Nodes;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.time.Instant;
import java.util.List;
import java.util.function.Function;

@Slf4j
public class ComputorService {

    private final Nodes nodes;
    private final IdentityUtil idu;
    private final TransactionService transactionService;

    @SuppressWarnings("unused")
    public ComputorService(Nodes nodes) {
        this(nodes, new IdentityUtil());
    }

    public ComputorService(Nodes nodes, IdentityUtil identityUtil) {
        this(nodes, identityUtil, new TransactionService(identityUtil));
    }

    public ComputorService(Nodes nodes, IdentityUtil idu, TransactionService transactionService) {
        this.nodes = nodes;
        this.idu = idu;
        this.transactionService = transactionService;
    }

    public Mono<TickInfo> getCurrentTickInfo() {
        return nodes.getNode().flatMap(Node::getCurrentTickInfo);
    }

    public Mono<Entity> getEntity(String identity) {
        Mono<byte[]> publicKey = Mono.just(identity)
                .map(idu::getPublicKeyFromIdentity);

        return nodes.getNode()
                .doOnNext(n -> log.info("[{}] Query id [{}].", n.getName(), identity))
                .flatMap(n -> n.getEntity(publicKey));
    }

    public Mono<Entity> getEntity(byte[] publicKey) {
        return nodes.getNode()
                .doOnNext(n -> log.debug("[{}] Query id by public key.", n.getName()))
                .flatMap(n -> n.getEntity(Mono.just(publicKey)));
    }

    public Mono<SignedTransaction> sendQuTransferTransaction(String seed, String targetIdentity, long amount, int tickOffset) {
        return sendTransaction(createSignedTransferTransaction(seed, targetIdentity, amount, tickOffset));
    }

    public Mono<SignedTransaction> sendTransaction(SignedTransaction signedTransaction) {
        return nodes.getNode().flatMap(node -> node
                .broadcastTransaction(Mono.just(signedTransaction))
                .doOnSuccess(tx -> log.info("[{}] Sent {}", node.getName(), getTransactionLogString(tx)))
        );
    }

    public Mono<SignedTransaction> sendTransaction(Function<Integer, SignedTransaction> createTransactionFunction) {
        return nodes.getNode().flatMap(node -> node
                .getCurrentTickInfo()
                .map(TickInfo::getTick)
                .map(createTransactionFunction)
                .flatMap(stx -> node.broadcastTransaction(Mono.just(stx)))
                .doOnSuccess(stx -> log.info(getTransactionLogString(stx)))
        );
    }

    private Function<Integer, SignedTransaction> createSignedTransferTransaction(String seed, String targetIdentity, long amount, int tickOffset) {
        return tick -> transactionService.createTransaction(tick + tickOffset, seed, targetIdentity, amount);
    }

    public Mono<TickData> getTickData(int tick) {
        return nodes.getNode().flatMap(node -> node.getTickData(tick)
                        .doOnNext(td -> log.info("{}", td))
        );
    }

    public Flux<SignedTransaction> getTickTransactions(int tick) {
        return getNode(tick).flatMapMany(node -> node
                .getTickTransactions(tick)
                .map(tx -> {
                    tx.setTransactionHash(transactionService.calculateTransactionHash(tx));
                    log.debug(getTransactionLogString(tx));
                    return tx;
                })
        );
    }

    private Mono<Node> getNode(int minTick) {
        return nodes.getNode().flatMap(node -> {
            if (node.getTickInfo().getTick() >= minTick) { // cached tick info is good enough
                return Mono.just(node);
            } else { // query node for current tick info
                return node.getCurrentTickInfo()
                        .filter(ti -> ti.getTick() >= minTick)
                        .map(ti -> node)
                        .switchIfEmpty(Mono.defer(() -> {
                            String message = String.format("Node [%s] is at tick [%d]. Required tick [%d].",
                                    node.getName(), node.getTickInfo().getTick(), minTick);
                            log.warn(message);
                            return Mono.error(new IllegalStateException(message));
                        }));
            }
        });
    }

    // TODO test
    public Flux<RespondOwnedAssets> getOwnedAssets(String identity) {
        Mono<byte[]> publicKey = Mono.just(identity)
                .map(idu::getPublicKeyFromIdentity);
        return nodes.getNode()
                .doOnNext(n -> log.info("[{}] Get owned assets for [{}].", n.getName(), identity))
                .flatMapMany(n -> n.getOwnedAssets(publicKey));
    }

    // TODO test
    public Flux<RespondPossessedAssets> getPossessedAssets(String identity) {
        Mono<byte[]> publicKey = Mono.just(identity)
                .map(idu::getPublicKeyFromIdentity);
        return nodes.getNode()
                .doOnNext(n -> log.info("[{}] Get possessed assets for [{}].", n.getName(), identity))
                .flatMapMany(n -> n.getPossessedAssets(publicKey));
    }

    public Mono<QxFees> getQxFees() {
        return nodes.getNode()
                .doOnNext(n -> log.info("[{}] Get qx fees.", n.getName()))
                .flatMap(Node::getQxFees);
    }

    public Mono<List<AssetOrder>> getQxAskAssetOrders(String issuer, String assetName, long offset) {
        QxGetAssetAskOrders request = new QxGetAssetAskOrders(idu.getPublicKeyFromIdentity(issuer), assetName, offset);
        return getQxAssetOrders(request);
    }

    public Mono<List<AssetOrder>> getQxBidAssetOrders(String issuer, String assetName, long offset) {
        QxGetAssetBidOrders request = new QxGetAssetBidOrders(idu.getPublicKeyFromIdentity(issuer), assetName, offset);
        return getQxAssetOrders(request);
    }

    private Mono<List<AssetOrder>> getQxAssetOrders(QxGetAssetOrders request) {
        return nodes.getNode()
                .doOnNext(n -> log.info("[{}] Get qx orders for asset [{}]. Offset: [{}].", n.getName(), request.getAssetName(), request.getOffset()))
                .flatMap(n -> n.getQxAssetOrders(request));
    }

    public Mono<List<EntityOrder>> getQxAskEntityOrders(String identity, long offset) {
        QxGetEntityAskOrders request = new QxGetEntityAskOrders(idu.getPublicKeyFromIdentity(identity), offset);
        return getQxEntityOrders(request, identity);
    }

    public Mono<List<EntityOrder>> getQxBidEntityOrders(String identity, long offset) {
        QxGetEntityBidOrders request = new QxGetEntityBidOrders(idu.getPublicKeyFromIdentity(identity), offset);
        return getQxEntityOrders(request, identity);
    }

    private Mono<List<EntityOrder>> getQxEntityOrders(QxGetEntityOrders request, String identity) {
        return nodes.getNode()
                .doOnNext(n -> log.info("[{}] Get qx orders for entity [{}]. Offset: [{}].", n.getName(), identity, request.getOffset()))
                .flatMap(n -> n.getQxEntityOrders(request));
    }

    public Mono<UserLockInfo> getQearnUserLockInfo(String identity, int epoch) {
        return nodes.getNode()
                .doOnNext(n -> log.info("[{}] Get qearn lock info for identity [{}] and epoch [{}].", n.getName(), identity, epoch))
                .flatMap(n -> n.getQearnUserLockInfo(GetUserLockInfo.builder()
                        .publicKey(idu.getPublicKeyFromIdentity(identity))
                        .epoch(epoch)
                        .build()));
    }

    public Mono<UserLockStatus> getQearnUserLockStatus(String identity) {
        return nodes.getNode()
                .doOnNext(n -> log.info("[{}] Get qearn lock status for identity [{}].", n.getName(), identity))
                .flatMap(n -> n.getQearnUserLockStatus(GetUserLockStatus.builder()
                        .publicKey(idu.getPublicKeyFromIdentity(identity))
                        .build()));
    }

    public Mono<LockInfoForEpoch> getQearnLockInfoForEpoch(int epoch) {
        return nodes.getNode()
                .doOnNext(n -> log.info("[{}] Get qearn lock info for epoch [{}].", n.getName(), epoch))
                .flatMap(n -> n.getQearnLockInfoForEpoch(GetLockInfoForEpoch.builder()
                        .epoch(epoch)
                        .build()));
    }

    public Mono<StateOfRound> getQearnStateOfRound(int epoch) {
        return nodes.getNode()
                .doOnNext(n -> log.info("[{}] Get qearn state of round for epoch [{}].", n.getName(), epoch))
                .flatMap(n -> n.getQearnStateOfRound(GetStateOfRound.builder()
                        .epoch(epoch)
                        .build()));
    }

    public Mono<EndedStatus> getQearnEndedStatus(String identity) {
        return nodes.getNode()
                .doOnNext(n -> log.info("[{}] Get qearn ended status for identity [{}].", n.getName(), identity))
                .flatMap(n -> n.getQearnEndedStatus(GetEndedStatus.builder()
                        .publicKey(idu.getPublicKeyFromIdentity(identity))
                        .build()));
    }

    public Mono<StatsForEpoch> getQearnStatsForEpoch(int epoch) {
        return nodes.getNode()
                .doOnNext(n -> log.info("[{}] Get qearn stats for epoch [{}].", n.getName(), epoch))
                .flatMap(n -> n.getQearnStatsForEpoch(GetStatsForEpoch.builder()
                        .epoch(epoch)
                        .build()));
    }

    public Mono<BurnedAndBoostedStats> getQearnBurnedAndBoostedStats() {
        return nodes.getNode()
                .doOnNext(n -> log.info("[{}] Get qearn burned and boosted stats.", n.getName()))
                .flatMap(n -> n.getQearnBurnedAndBoostedStats(new GetBurnedAndBoostedStats()));
    }

    public Mono<BurnedAndBoostedStatsForEpoch> getQearnBurnedAndBoostedStatsForEpoch(int epoch) {
        return nodes.getNode()
                .doOnNext(n -> log.info("[{}] Get qearn burned and boosted stats for epoch [{}].", n.getName(), epoch))
                .flatMap(n -> n.getQearnBurnedAndBoostedStatsForEpoch(GetBurnedAndBoostedStatsForEpoch.builder()
                        .epoch(epoch)
                        .build()));
    }

    public Mono<Instant> forceNetworkUpdateIfBefore(Instant updateIfBefore) {
        return nodes.forceNetworkUpdate(updateIfBefore);
    }

    private String getTransactionLogString(SignedTransaction signedTransaction) {
        Transaction transaction = signedTransaction.getTransaction();
        return String.format("tx [%s], tick [%d], [%s] => [%s], qu [%d], hash [%s], input type [%s], input size [%s], extra data [%s]",
                StringUtils.abbreviate(signedTransaction.getTransactionHash(), "", 10),
                transaction.getTick(),
                idu.getIdentityFromPublicKey(transaction.getSourcePublicKey()),
                idu.getIdentityFromPublicKey(transaction.getDestinationPublicKey()),
                transaction.getAmount(),
                signedTransaction.getTransactionHash(),
                transaction.getInputType(),
                transaction.getInputSize(),
                transaction.getExtraData() == null ? null : Hex.encodeHexString(transaction.getExtraData())
        );
    }

}
