package at.qubic.api.service;

import org.apache.commons.lang3.ArrayUtils;
import at.qubic.api.crypto.IdentityUtil;
import at.qubic.api.crypto.KangarooTwelve;
import at.qubic.api.crypto.QubicFourQ;
import at.qubic.api.crypto.SchnorrQubic;
import at.qubic.api.domain.std.SignedTransaction;
import at.qubic.api.domain.std.Transaction;

import java.util.Objects;

public class TransactionService {

    private final IdentityUtil identityUtil;
    private final SchnorrQubic signing;

    public TransactionService() {
        this(new IdentityUtil());
    }

    public TransactionService(IdentityUtil identityUtil) {
        this(identityUtil, new QubicFourQ());
    }

    public TransactionService(IdentityUtil identityUtil, SchnorrQubic schnorrQubic) {
        this.identityUtil = identityUtil;
        this.signing = schnorrQubic;
    }

    public SignedTransaction createTransaction(int tick, String seed, String destinationIdentity, long amount) {
        return createTransaction(tick, seed, destinationIdentity, amount, (short) 0, new byte[0]);
    }

    public SignedTransaction createTransaction(int tick, String seed, String destinationIdentity, long amount, short inputType, byte[] extraData) {
        byte[] subSeed = identityUtil.getSubSeedFromSeed(seed);
        byte[] sourcePublicKey = identityUtil.getPublicKeyFromPrivateKey(identityUtil.getPrivateKeyFromSubSeed(subSeed));
        byte[] destinationPublicKey = identityUtil.getPublicKeyFromIdentity(destinationIdentity);
        Transaction transaction = createTransaction(tick, amount, sourcePublicKey, destinationPublicKey, inputType, extraData);
        final byte[] signature = createSignatureWithSubSeed(subSeed, sourcePublicKey, transaction);
        return createSignedTransaction(transaction, signature);
    }

    public SignedTransaction createTransactionWithNonceK(int tick, byte[] nonceK, String sourceIdentity, String destinationIdentity, long amount) {
        return createTransactionWithNonceK(tick, nonceK, sourceIdentity, destinationIdentity, amount, (short) 0, new byte[0]);
    }

    public SignedTransaction createTransactionWithNonceK(int tick, byte[] nonceK, String sourceIdentity, String destinationIdentity, long amount, short inputType, byte[] extraData) {
        if (ArrayUtils.getLength(nonceK) != 64) {
            throw new IllegalArgumentException("Nonce k must be 64 bytes");
        }
        byte[] sourcePublicKey = identityUtil.getPublicKeyFromIdentity(sourceIdentity);
        byte[] destinationPublicKey = identityUtil.getPublicKeyFromIdentity(destinationIdentity);
        Transaction transaction = createTransaction(tick, amount, sourcePublicKey, destinationPublicKey, inputType, extraData);
        final byte[] signature = createSignatureWithNonceK(nonceK, sourcePublicKey, transaction);
        return createSignedTransaction(transaction, signature);
    }

    private static Transaction createTransaction(int tick, long amount, byte[] sourcePublicKey, byte[] destinationPublicKey, short inputType, byte[] extraData) {
        Objects.requireNonNull(extraData, "Extra data cannot be null. Pass byte[0] instead.");
        short inputSize = (short) extraData.length;
        return Transaction.builder()
                .sourcePublicKey(sourcePublicKey)
                .destinationPublicKey(destinationPublicKey)
                .amount(amount)
                .inputType(inputType)
                .inputSize(inputSize)
                .extraData(extraData)
                .tick(tick)
                .build();
    }

    private byte[] createSignatureWithSubSeed(byte[] subSeed, byte[] publicKey, Transaction transaction) {
        byte[] txDigest = KangarooTwelve.hash(transaction.toBytes(), 32);
        return signing.sign(subSeed, publicKey, txDigest);
    }

    private byte[] createSignatureWithNonceK(byte[] nonceK, byte[] publicKey, Transaction transaction) {
        byte[] txDigest = KangarooTwelve.hash(transaction.toBytes(), 32);
        return signing.signWithNonceK(nonceK, publicKey, txDigest);
    }

    private SignedTransaction createSignedTransaction(Transaction transaction, byte[] signature) {
        final byte[] transactionBytes = transaction.toBytes();
        return SignedTransaction.builder()
                .transaction(transaction)
                .signature(signature)
                .transactionHash(calculateTransactionHash(ArrayUtils.addAll(transactionBytes, signature)))
                .build();
    }

    public String calculateTransactionHash(SignedTransaction transaction) {
        return calculateTransactionHash(transaction.toBytes());
    }

    private String calculateTransactionHash(byte[] signedTransactionBytes) {
        final byte[] signedTxDigest = KangarooTwelve.hash(signedTransactionBytes, 32);
        return identityUtil.getTransactionHash(signedTxDigest);
    }

    public boolean verify(SignedTransaction signedTransaction) {
        Objects.requireNonNull(signedTransaction.getTransaction());
        final byte[] transactionBytes = signedTransaction.getTransaction().toBytes();
        final byte[] signedTxDigest = KangarooTwelve.hash(transactionBytes, 32);
        final byte[] sourcePublicKey = signedTransaction.getTransaction().getSourcePublicKey();
        final byte[] signature = signedTransaction.getSignature();
        return signing.verify(sourcePublicKey, signedTxDigest, signature);
    }

}
