package at.qubic.api.util;

import org.apache.commons.lang3.StringUtils;

import java.nio.charset.StandardCharsets;

public class AssetUtil {

    /**
     * Creates an ascii string out of the specified input byte array. Null characters are trimmed.
     * @param name The byte array containing the name.
     * @return The byte array converted to a string.
     */
    public static String nameToString(byte[] name) {
        return name == null ? null : new String(name, StandardCharsets.US_ASCII).trim();
    }

    /**
     * Return input string as byte[8]. The last byte is zero (null terminated string).
     * @param name The input string. Shouldn't be more than 7 bytes. Other bytes are ignored.
     * @return The name as null terminated byte array of length 8.
     */
    public static byte[] nameTo8Bytes(String name) {
        byte[] assetName = new byte[8];
        return toNameBytes(name, assetName);
    }

    /**
     * Return input string as byte[8]. The last byte is zero (null terminated string).
     * @param name The input string. Shouldn't be more than 7 bytes. Other bytes are ignored.
     * @return The name as null terminated byte array of length 8.
     */
    public static byte[] nameTo7Bytes(String name) {
        byte[] assetName = new byte[7];
        return toNameBytes(name, assetName);
    }

    private static byte[] toNameBytes(String name, byte[] assetName) {
        for (int i = 0; i < Math.min(StringUtils.length(name), 7); i++) {
            assetName[i] = (byte) name.charAt(i);
        }
        return assetName;
    }

    /**
     * Converts a unit of measurement string to bytes. Letters are mapped to their corresponding byte value.
     * "00000020" is converted to byte {0,0,0,0,0,0,2,0} for example.
     * @param unitOfMeasurement String describing the units of measurement for an asset.
     * @return The units of measurement in byte array format.
     */
    public static byte[] unitOfMeasurementTo8Bytes(String unitOfMeasurement) {
        byte[] units = new byte[8];
        toUnitOfMeasurementBytes(unitOfMeasurement, units);
        return units;
    }

    /**
     * Converts a unit of measurement string to bytes. Letters are mapped to their corresponding byte value.
     * "00000020" is converted to byte {0,0,0,0,0,0,2} for example.
     * @param unitOfMeasurement String describing the units of measurement for an asset.
     * @return The units of measurement in byte array format.
     */
    public static byte[] unitOfMeasurementTo7Bytes(String unitOfMeasurement) {
        byte[] units = new byte[7];
        toUnitOfMeasurementBytes(unitOfMeasurement, units);
        return units;
    }

    private static void toUnitOfMeasurementBytes(String unitOfMeasurement, byte[] units) {
        for (int i = 0; i < Math.min(7, StringUtils.length(unitOfMeasurement)); i++) {
            units[i] = (byte) (unitOfMeasurement.charAt(i) - 48);
        }
    }

    /**
     * Converts a units of measurement array to a readable string. byte values are mapped to their corresponding
     * ascii letters, for example 0 is mapped to the '0' characters.
     * @param input The units of measurement bytes. Only the first 7 bytes are converted.
     * @return An ascii string representing the units of measurement.
     */
    public static String unitOfMeasurementToString(byte[] input) {
        byte[] converted = new byte[7];
        // convert 0-9 to ASCII 0-9 (decimal 48-57) by adding 48. 0000020 is then shown as "0000020" for example.
        for (int i = 0; i < Math.min(7, input.length); i++) { converted[i] = (byte) (input[i] + 48); }
        return new String(converted, StandardCharsets.US_ASCII).trim();
    }

}
