package at.qubic.api;

import lombok.extern.slf4j.Slf4j;
import at.qubic.api.crypto.FourQ;
import at.qubic.api.crypto.IdentityUtil;
import at.qubic.api.crypto.QubicFourQ;
import at.qubic.api.crypto.SchnorrQubic;
import at.qubic.api.network.*;
import at.qubic.api.properties.ComputorProperties;
import at.qubic.api.properties.NetworkProperties;
import at.qubic.api.service.ComputorService;
import at.qubic.api.service.QxTransactionService;
import at.qubic.api.service.TransactionService;
import org.springframework.boot.SpringBootConfiguration;
import org.springframework.boot.context.properties.ConfigurationProperties;
import org.springframework.boot.context.properties.EnableConfigurationProperties;
import org.springframework.context.annotation.Bean;
import org.springframework.core.env.Environment;

import java.util.Arrays;

@SpringBootConfiguration
@EnableConfigurationProperties
@Slf4j
public class SpringBootTestConfiguration {

    @Bean
    @ConfigurationProperties(value = "computor", ignoreUnknownFields = false)
    ComputorProperties computorProperties() {
        return new ComputorProperties();
    }

    @Bean
    @ConfigurationProperties(value = "network", ignoreUnknownFields = false)
    NetworkProperties networkProperties() {
        return new NetworkProperties();
    }

    @Bean
    NetworkStatus networkStatus(Environment environment) {
        return environment.getRequiredProperty("check.health", Boolean.class)
                ? new HealthCheckNetworkStatus()
                : new NoopNetworkStatus();
    }

    @Bean
    NodesManagementStrategy nodesManagementStrategy(NetworkProperties properties, NetworkStatus networkStatus, Environment environment) {
        boolean usePublicPeers = environment.getRequiredProperty("use.public.peers", Boolean.class);
        log.info("Using port [{}].", properties.getPort());
        return usePublicPeers
                ? new PublicNodesStrategy(networkStatus, properties)
                : new FixNodesStrategy(networkStatus, properties);
    }

    @Bean
    Nodes nodes(NodesManagementStrategy nodesManagementStrategy, ComputorProperties properties) {
        Nodes nodes = new Nodes(nodesManagementStrategy);
        log.info("Using computors {}.", Arrays.toString(properties.getHosts()));
        for (String host : properties.getHosts()) {
            nodes.addNode(host);
        }
        return nodes;
    }

    @Bean
    QubicFourQ crypto(Environment env) {
        String fourQ = env.getRequiredProperty("fourq.lib.name");
        log.info("FourQ library: [{}].", fourQ);
        return new QubicFourQ(fourQ);
    }

    @Bean
    IdentityUtil identityUtil(Environment env, FourQ fourQ) {
        boolean verify = env.getRequiredProperty("id.verifications", Boolean.class);
        return new IdentityUtil(verify, fourQ);
    }

    @Bean
    TransactionService transactionService(IdentityUtil identityUtil, SchnorrQubic signing) {
        return new TransactionService(identityUtil, signing);
    }

    @Bean
    QxTransactionService qxTransactionService(TransactionService transactionService) {
        return new QxTransactionService(transactionService);
    }

    @Bean
    ComputorService qubicComputorService(Nodes nodes, IdentityUtil identityUtil, TransactionService transactionService) {
        return new ComputorService(nodes, identityUtil, transactionService);
    }

}
