package at.qubic.api.crypto;

import org.apache.commons.codec.binary.Hex;
import org.junit.jupiter.api.Test;
import at.qubic.api.exception.InvalidIdentityException;

import java.nio.charset.StandardCharsets;

import static org.apache.commons.codec.binary.Hex.decodeHex;
import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatThrownBy;

@SuppressWarnings("SpellCheckingInspection")
class IdentityUtilTest {

    private final IdentityUtil util = new IdentityUtil();

    /*  Examples for testing

        Seed: sjklwvnfvsqrutwiixfjchqecqdnafujozygbzcctijynbytgrdihzq
        Private key: 969ea207cc1e1a0e1801b3c4941af919c67ab4f4b3cc6bf5a680709bae5d378e
        Public key: 7fac844c525271704f35a9f5a5702a5c93b17ed8a5c0a6287e1dbdd321111f1b
        Identity: BLCFPJBEMLNXGDVJQOUVKCCHLPRCHQRRCMLBHECSEBGAMTHRTJTWLMUAKQPG

        Seed: abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabc
        Private key: 1d335317cb32506c484aa582c78b89053deddc7ea3a600749a69515a800aa861
        Public key: 88b5035d1ba0860b1c26ae5a4a070a52f1596d10d816bad98862eb75f77f9fad
        Identity: CZPGJVFHPHHSIAMUOALHMCYFQYJCJGNPOYLLDYMKIGYYQEMJJNFMQCBFAVBB
     */

    @Test
    void getPublicKeyFromIdentity() throws Exception {
        assertThat(util.getPublicKeyFromIdentity("BZBQFLLBNCXEMGLOBHUVFTLUPLVCPQUASSILFABOFFBCADQSSUPNWLZBQEXK"))
                .isEqualTo(decodeHex("1f590d03e613bdded38b4c0820ac44615f91af12435980b3ede3c08c315a2544"));

        assertThat(util.getPublicKeyFromIdentity("AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAFXIB"))
                .isEqualTo(new byte[32]);

        assertThat(util.getPublicKeyFromIdentity("BAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAARMID"))
                .isEqualTo(new byte[] { 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0});

        assertThat(util.getPublicKeyFromIdentity("BLCFPJBEMLNXGDVJQOUVKCCHLPRCHQRRCMLBHECSEBGAMTHRTJTWLMUAKQPG"))
                .isEqualTo(decodeHex("7fac844c525271704f35a9f5a5702a5c93b17ed8a5c0a6287e1dbdd321111f1b"));

        assertThat(util.getPublicKeyFromIdentity("CZPGJVFHPHHSIAMUOALHMCYFQYJCJGNPOYLLDYMKIGYYQEMJJNFMQCBFAVBB"))
                .isEqualTo(decodeHex("88b5035d1ba0860b1c26ae5a4a070a52f1596d10d816bad98862eb75f77f9fad"));

        assertThat(util.getPublicKeyFromIdentity("TESTIOYEZSAUQBUNKJEGXOILIBICZYWAMVSWYCYFHDCSZOAXPTGSOKFCCUNK"))
                .isEqualTo(decodeHex("bba07fde537ea438760811910f2d874f017974eadcfdde70aa771533aa94064c"));

        assertThat(util.getPublicKeyFromIdentity("ITXUCKTOAXGLKFPOQMBDZJDSHUKCKZNNCACNONPWZESXDCILUKUMLJRAPPPD"))
                .isEqualTo(decodeHex("f219230a0b56fbb94b3d73d3879f245390fccc3b23c1fdab0482f59b14a9fe16"));

    }

    @Test
    void getPublicKeyFromIdentity_verifyChecksum() {
        assertThat(util.getPublicKeyFromIdentity("BZBQFLLBNCXEMGLOBHUVFTLUPLVCPQUASSILFABOFFBCADQSSUPNWLZBQEXK")).isNotEmpty();
        assertThat(util.getPublicKeyFromIdentity("BLCFPJBEMLNXGDVJQOUVKCCHLPRCHQRRCMLBHECSEBGAMTHRTJTWLMUAKQPG")).isNotEmpty();
        assertThat(util.getPublicKeyFromIdentity("CZPGJVFHPHHSIAMUOALHMCYFQYJCJGNPOYLLDYMKIGYYQEMJJNFMQCBFAVBB")).isNotEmpty();

        assertThatThrownBy(() -> util.getPublicKeyFromIdentity("BZBQFLLBNCXEMGLOBHUVFTLUPLVCPQUASSILFABOFFBCADQSSUPNWLZBQEXZ"))
                .isInstanceOf(InvalidIdentityException.class).hasMessageContaining("Checksum");
    }

    @Test
    void getSubSeedFromSeed() throws Exception {
        assertThat(util.getSubSeedFromSeed("sjklwvnfvsqrutwiixfjchqecqdnafujozygbzcctijynbytgrdihzq"))
                .isEqualTo(decodeHex("be55905cb197d59bdf935b977637e2afd468271113de497adcaf0dfa3f5d55b2"));

        assertThat(util.getSubSeedFromSeed("abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabc"))
                .isEqualTo(decodeHex("58aabff86c0cd56874851343be2a7023b83c4760d0b70ec3e1e2ccffa5c5a8ae"));
    }

    @Test
    void getPrivateKeyFromSubSeed() throws Exception {
        assertThat(util.getPrivateKeyFromSubSeed(decodeHex("be55905cb197d59bdf935b977637e2afd468271113de497adcaf0dfa3f5d55b2")))
                .isEqualTo(decodeHex("969ea207cc1e1a0e1801b3c4941af919c67ab4f4b3cc6bf5a680709bae5d378e"));

        assertThat(util.getPrivateKeyFromSubSeed(decodeHex("58aabff86c0cd56874851343be2a7023b83c4760d0b70ec3e1e2ccffa5c5a8ae")))
                .isEqualTo(decodeHex("1d335317cb32506c484aa582c78b89053deddc7ea3a600749a69515a800aa861"));
    }

    @Test
    void getPublicKeyFromPrivateKey() throws Exception {
        assertThat(util.getPublicKeyFromPrivateKey(decodeHex("969ea207cc1e1a0e1801b3c4941af919c67ab4f4b3cc6bf5a680709bae5d378e")))
                .isEqualTo(decodeHex("7fac844c525271704f35a9f5a5702a5c93b17ed8a5c0a6287e1dbdd321111f1b"));

        assertThat(util.getPublicKeyFromPrivateKey(decodeHex("1d335317cb32506c484aa582c78b89053deddc7ea3a600749a69515a800aa861")))
                .isEqualTo(decodeHex("88b5035d1ba0860b1c26ae5a4a070a52f1596d10d816bad98862eb75f77f9fad"));
    }

    @Test
    void getPublicKeyFromSeed() throws Exception {
        assertThat(util.getPublicKeyFromSeed("sjklwvnfvsqrutwiixfjchqecqdnafujozygbzcctijynbytgrdihzq"))
                .isEqualTo(decodeHex("7fac844c525271704f35a9f5a5702a5c93b17ed8a5c0a6287e1dbdd321111f1b"));

        assertThat(util.getPublicKeyFromSeed("abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabc"))
                .isEqualTo(decodeHex("88b5035d1ba0860b1c26ae5a4a070a52f1596d10d816bad98862eb75f77f9fad"));

    }

    @Test
    void getPublicKeyFromSubSeed() throws Exception {
        assertThat(util.getPublicKeyFromSubSeed(decodeHex("be55905cb197d59bdf935b977637e2afd468271113de497adcaf0dfa3f5d55b2")))
                .isEqualTo(decodeHex("7fac844c525271704f35a9f5a5702a5c93b17ed8a5c0a6287e1dbdd321111f1b"));

        assertThat(util.getPublicKeyFromSubSeed(decodeHex("58aabff86c0cd56874851343be2a7023b83c4760d0b70ec3e1e2ccffa5c5a8ae")))
                .isEqualTo(decodeHex("88b5035d1ba0860b1c26ae5a4a070a52f1596d10d816bad98862eb75f77f9fad"));
    }

    @Test
    void getIdentityFromPublicKey() throws Exception {
        assertThat(util.getIdentityFromPublicKey(decodeHex("1f590d03e613bdded38b4c0820ac44615f91af12435980b3ede3c08c315a2544")))
                .isEqualTo("BZBQFLLBNCXEMGLOBHUVFTLUPLVCPQUASSILFABOFFBCADQSSUPNWLZBQEXK");

        assertThat(util.getIdentityFromPublicKey(new byte[32]))
                .isEqualTo("AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAFXIB");


        assertThat(util.getIdentityFromPublicKey(decodeHex("7fac844c525271704f35a9f5a5702a5c93b17ed8a5c0a6287e1dbdd321111f1b")))
                .isEqualTo("BLCFPJBEMLNXGDVJQOUVKCCHLPRCHQRRCMLBHECSEBGAMTHRTJTWLMUAKQPG");

        assertThat(util.getIdentityFromPublicKey(decodeHex("88b5035d1ba0860b1c26ae5a4a070a52f1596d10d816bad98862eb75f77f9fad")))
                .isEqualTo("CZPGJVFHPHHSIAMUOALHMCYFQYJCJGNPOYLLDYMKIGYYQEMJJNFMQCBFAVBB");
    }

    @Test // roundtrip
    void roundtripFromSeedToIdentity() {
        verifyId("sjklwvnfvsqrutwiixfjchqecqdnafujozygbzcctijynbytgrdihzq", "BLCFPJBEMLNXGDVJQOUVKCCHLPRCHQRRCMLBHECSEBGAMTHRTJTWLMUAKQPG");
        verifyId("abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabc", "CZPGJVFHPHHSIAMUOALHMCYFQYJCJGNPOYLLDYMKIGYYQEMJJNFMQCBFAVBB");
        verifyId("tcpykuzkmivjjsjqfcdnekqeggrooenxhwbfwgfigjkivbgthythbpk", "TESTHHKPOCSTJGBDJGUFHOKARUFAMUMGDFTQEVAHGHNABPPSENXMZTVDJRRA");
        verifyId("ewsvzoaeqhlxxoebzawvlvvsrgubvlyxvquucvdtjrekqxhmniuszqi", "TESTIOYEZSAUQBUNKJEGXOILIBICZYWAMVSWYCYFHDCSZOAXPTGSOKFCCUNK");
        verifyId("tnvquwhrpkrrxekrfivsxagmhqtrjrgbacdpkjfqegpbsvztuzyrier", "TESTJFDGICVGFHLRAZCKJKQYVEUBVYWTMRYYTPFXRESOQJGTKHCIHDJGFQZJ");
        verifyId("woosnyntviegnodobfanoettfcsxrbacadgnutrzxvlcbmpinhvuage", "KJAENNYZIQUNMCABZCQVVTSDAKODKJZFBCKLXLKMMELELMXNEILNFHCFOOIA");
    }

    private void verifyId(String seed, String identity) {
        final byte[] subSeed = util.getSubSeedFromSeed(seed);
        final byte[] privateKey = util.getPrivateKeyFromSubSeed(subSeed);
        final byte[] publicKey = util.getPublicKeyFromPrivateKey(privateKey);
        assertThat(util.getIdentityFromPublicKey(publicKey)).isEqualTo(identity);
    }

    @Test
    void getTransactionHash() throws Exception {
        byte[][] signedTransactions = {
                Hex.decodeHex("f219230a0b56fbb94b3d73d3879f245390fccc3b23c1fdab0482f59b14a9fe1604929d794d9b2919060274135e716e07b2c1a8835cf13f04179914d6ff0d43d80100000000000000aed5ab0000000000e91738e096fa762ecea2eaf5bb16d462be6ee4be787a70bc890f79d42d8df386526800072e5e3d138ceed8d28abdaeb1f1bc7b6f60b46d8987a260da75600400"),
                Hex.decodeHex("88b5035d1ba0860b1c26ae5a4a070a52f1596d10d816bad98862eb75f77f9fad04929d794d9b2919060274135e716e07b2c1a8835cf13f04179914d6ff0d43d8010000000000000010daab000000000016d18a2f1a9f3403a5549f9fa3f0242458c077b22867d0db7bb6b8a23824fe1fae90b8df0b81858eb77058bf0928ad55247fa027e2cf6ee440b91ad839042300")
        };

        String[] txHashes = {
                "lvbcmkgebiwykboktuekddkpssdavggmatctivrsugpgsxhrlwuaplbdbhgg",
                "ovlwchpiapdcmedcusakiatkvovcprsigtdfizgvjbklyoxitmwofysblwfk"
        };

        for (int i = 0; i < txHashes.length; i++) {
            final byte[] signedTxHash = KangarooTwelve.hash(signedTransactions[i], 32);
            String txHash = util.getTransactionHash(signedTxHash);
            assertThat(txHash).isEqualTo(txHashes[i]);
        }
    }

    // tests for errors

    @Test
    void getPrivateKeyFromSubSeedFromSeed_invalidLength_throw() {
        assertThatThrownBy(
                () -> util.getPrivateKeyFromSubSeed(decodeHex("be55905cb197d59bdf935b977637e2afd468271113de497adcaf0dfa3f5d55"))
        ).isInstanceOf(InvalidIdentityException.class);
    }

    @Test
    void getPublicKeyFromPrivateKey_invalidLength_throw() {
        assertThatThrownBy(
                () -> util.getPrivateKeyFromSubSeed(decodeHex("969ea207cc1e1a0e1801b3c4941af919c67ab4f4b3cc6bf5a680709bae5d37"))
        ).isInstanceOf(InvalidIdentityException.class);
    }

    @Test
    void isValidIdentityString() {
        assertThat(util.isValidIdentity("BZBQFLLBNCXEMGLOBHUVFTLUPLVCPQUASSILFABOFFBCADQSSUPNWLZBQEXK")).isTrue();

        assertThat(util.isValidIdentity(null)).isFalse();
        assertThat(util.isValidIdentity("BZBQFLLBNCXEMGLOBHUVxyzUPLVCPQUASSILFABOFFBCADQSSUPNWLZBQEXK")).isFalse();
        assertThat(util.isValidIdentity("BZBQFLLBNCXEMGLOBHUV1234PLVCPQUASSILFABOFFBCADQSSUPNWLZBQEXK")).isFalse();
        assertThat(util.isValidIdentity("BZB-FLLBNCXEMGLOBHUV1234PLVCPQUASSILFABOFFBCADQSSUPNWLZBQEXK")).isFalse();
        assertThat(util.isValidIdentity("BZB.FLLBNCXEMGLOBHUV1234PLVCPQUASSILFABOFFBCADQSSUPNWLZBQEXK")).isFalse();
        assertThat(util.isValidIdentity("BZBQFLLBNCXEMGLOBHUVFTLUPLVCPQUASSILFABOFFBCADQSSUPNWLZBQEX")).isFalse();
    }

    @Test
    void isValidIdentityChecksum() {
        assertThat(util.isValidIdentity("BZBQFLLBNCXEMGLOBHUVFTLUPLVCPQUASSILFABOFFBCADQSSUPNWLZBQEXK")).isTrue();
        assertThat(util.isValidIdentity("TESTHHKPOCSTJGBDJGUFHOKARUFAMUMGDFTQEVAHGHNABPPSENXMZTVDJRRA")).isTrue();

        assertThat(util.isValidIdentity("TESTABCPOCSTJGBDJGUFHOKARUFAMUMGDFTQEVAHGHNABPPSENXMZTVDJRRA")).isFalse();
        assertThat(util.isValidIdentity("BZBQFLLBNCXEMGLOBHUVFTLUPLVCPQUASSILFABOFFBCADQSSUPNWLZBQEXZ")).isFalse();
    }

    @Test
    void isValidSeed() {
        assertThat(util.isValidSeed("sjklwvnfvsqrutwiixfjchqecqdnafujozygbzcctijynbytgrdihzq")).isTrue();

        assertThat(util.isValidSeed("")).isFalse();
        assertThat(util.isValidSeed(null)).isFalse();
        assertThat(util.isValidSeed("sjklwvnfvsqru-wiixfjchqecqdnafujozygbzcctijynbytgrdihzq")).isFalse();
        assertThat(util.isValidSeed("1jklwvnfvsqrutwiixfjchqecqdnafujozygbzcctijynbytgrdihzq")).isFalse();
        assertThat(util.isValidSeed("sjklwvnfvsqrutwiixfjchqecq2nafujozygbzcctijynbytgrdihzq")).isFalse();
        assertThat(util.isValidSeed("sJKLwvnfvsqrutwiixfjchqecq2nafujozygbzcctijynbytgrdihzq")).isFalse();
        assertThat(util.isValidSeed("sjklwvnfvsqrutwiixfjchqecqdnafujozygbzcctijynbytgrdihzqs")).isFalse();
        assertThat(util.isValidSeed("sjklwvnfvsqrutwiixfjchqecqdnafujozygbzcctijynbytgrdihz")).isFalse();


    }

    /*
        1 elbow / 2 mechanic / 3 infant / 4 you / 5 unveil / 6 model / 7 often / 8 random /
        9 invest / 10 arctic / 11 sauce / 12 search / 13 corn / 14 human / 15 shine / 16 pilot /
        17 grain / 18 cushion / 19 resist / 20 stem / 21 estate / 22 convince / 23 awake / 24 angry
     */
    @Test
    void convertBIP39ToSubSeed() {
        String words = "elbow mechanic infant you unveil model often random invest arctic sauce search corn human shine pilot grain cushion resist stem estate convince awake angry";
        byte[] subSeed = KangarooTwelve.hash(words.getBytes(StandardCharsets.US_ASCII), 32);
        byte[] publicKey = util.getPublicKeyFromSubSeed(subSeed);
        String identity = util.getIdentityFromPublicKey(publicKey);
        assertThat(identity).isEqualTo("RQPNRDIGLOSKIFNIJLXBJEKMXQSDGTUQHEOTHYEBHHGMZDBFHKGHBEZFBCVG");

    }

}