package at.qubic.api.crypto;

import org.apache.commons.codec.binary.Hex;
import org.junit.jupiter.api.Test;
import at.qubic.api.domain.std.Transaction;

import static org.apache.commons.codec.binary.Hex.decodeHex;
import static org.assertj.core.api.Assertions.assertThat;

class QubicFourQTest {

    private final QubicFourQ qubicFourQ = new QubicFourQ();
    private final IdentityUtil idu = new IdentityUtil(true, qubicFourQ);

    @Test
    void sign() throws Exception {
        byte[] sourceSubSeed = idu.getSubSeedFromSeed("abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabc");
        byte[] sourcePublicKey = idu.getPublicKeyFromPrivateKey(idu.getPrivateKeyFromSubSeed(sourceSubSeed));
        byte[] targetPublicKey = idu.getPublicKeyFromIdentity("GNRDTDTIEAAATAUPJZEQEEZHXPFAOQRAWIRJWGLFDANVSMQAWURDTHHGXGHC");
        Transaction tx = Transaction.builder()
                .sourcePublicKey(sourcePublicKey)
                .destinationPublicKey(targetPublicKey)
                .amount(1)
                .tick(11262480)
                .build();

        final byte[] txBytes = tx.toBytes();
        byte[] digest = KangarooTwelve.hash(txBytes, 32);

        final byte[] signature = qubicFourQ.sign(sourceSubSeed, sourcePublicKey, digest);
        assertThat(signature)
                .isEqualTo(Hex.decodeHex("16d18a2f1a9f3403a5549f9fa3f0242458c077b22867d0db7bb6b8a23824fe1fae90b8df0b81858eb77058bf0928ad55247fa027e2cf6ee440b91ad839042300"));
    }

    @Test
    void verify() throws Exception {
        byte[] sourceSubSeed = idu.getSubSeedFromSeed("abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabc");
        byte[] sourcePublicKey = idu.getPublicKeyFromPrivateKey(idu.getPrivateKeyFromSubSeed(sourceSubSeed));
        byte[] targetPublicKey = idu.getPublicKeyFromIdentity("GNRDTDTIEAAATAUPJZEQEEZHXPFAOQRAWIRJWGLFDANVSMQAWURDTHHGXGHC");
        final byte[] rawTx = Transaction.builder()
                .sourcePublicKey(sourcePublicKey)
                .destinationPublicKey(targetPublicKey)
                .amount(1)
                .tick(11262480)
                .build()
                .toBytes();
        byte[] digest = KangarooTwelve.hash(rawTx, 32);
        byte[] signature = Hex.decodeHex("16d18a2f1a9f3403a5549f9fa3f0242458c077b22867d0db7bb6b8a23824fe1fae90b8df0b81858eb77058bf0928ad55247fa027e2cf6ee440b91ad839042300");

        assertThat(qubicFourQ.verify(sourcePublicKey, digest, signature)).isTrue();
        assertThat(qubicFourQ.verify(targetPublicKey, digest, signature)).isFalse();
    }

    @Test
    void ecc_mul_fixed() throws Exception {
        final byte[] privateKey = decodeHex("969ea207cc1e1a0e1801b3c4941af919c67ab4f4b3cc6bf5a680709bae5d378e");
        final byte[] publicKey = qubicFourQ.ecc_mul_fixed(privateKey);
        assertThat(publicKey).isEqualTo(decodeHex("7fac844c525271704f35a9f5a5702a5c93b17ed8a5c0a6287e1dbdd321111f1b"));
    }

    @Test
    void ecc_mul_fixed_with_out_array() throws Exception {
        final byte[] privateKey = decodeHex("969ea207cc1e1a0e1801b3c4941af919c67ab4f4b3cc6bf5a680709bae5d378e");
        final byte[] publicKey = new byte[32];
        qubicFourQ.ecc_mul_fixed(privateKey, publicKey);
        assertThat(publicKey).isEqualTo(decodeHex("7fac844c525271704f35a9f5a5702a5c93b17ed8a5c0a6287e1dbdd321111f1b"));
    }

}