package at.qubic.api.domain.std;

import lombok.extern.slf4j.Slf4j;
import org.apache.commons.codec.binary.Hex;
import org.apache.commons.io.IOUtils;
import org.junit.jupiter.api.Test;

import java.util.Arrays;
import java.util.Objects;

import static org.apache.commons.codec.binary.Hex.decodeHex;
import static org.assertj.core.api.Assertions.assertThat;
import static at.qubic.api.domain.QubicHeader.HEADER_LENGTH;

@SuppressWarnings("SpellCheckingInspection")
@Slf4j
class SignedTransactionTest {

    @Test
    void toBytes() throws Exception {
        Transaction transaction = Transaction.builder()
                .sourcePublicKey(decodeHex("bba07fde537ea438760811910f2d874f017974eadcfdde70aa771533aa94064c"))
                .destinationPublicKey(decodeHex("f219230a0b56fbb94b3d73d3879f245390fccc3b23c1fdab0482f59b14a9fe16"))
                .amount(1)
                .tick(11270810)
                .build();
        SignedTransaction signedTransaction = SignedTransaction.builder()
                .transaction(transaction)
                .signature(Hex.decodeHex("1aefa6a4bf8ed746b618bbd1522a8d1e5ee47890b4dce9f694e6a32357ce9e90540a4fab145520149b71a95526a7deedf7fb528ee32b5a6b32181550f14d0a00"))
                .transactionHash("foo")
                .build();

        assertThat(signedTransaction.toBytes()).isEqualTo(Hex.decodeHex("bba07fde537ea438760811910f2d874f017974eadcfdde70aa771533aa94064cf219230a0b56fbb94b3d73d3879f245390fccc3b23c1fdab0482f59b14a9fe1601000000000000009afaab00000000001aefa6a4bf8ed746b618bbd1522a8d1e5ee47890b4dce9f694e6a32357ce9e90540a4fab145520149b71a95526a7deedf7fb528ee32b5a6b32181550f14d0a00"));

    }

    @Test
    void toBytes_withoutSignature() throws Exception {
        Transaction transaction = Transaction.builder()
                .sourcePublicKey(decodeHex("bba07fde537ea438760811910f2d874f017974eadcfdde70aa771533aa94064c"))
                .destinationPublicKey(decodeHex("f219230a0b56fbb94b3d73d3879f245390fccc3b23c1fdab0482f59b14a9fe16"))
                .amount(1)
                .tick(11270810)
                .build();

        SignedTransaction signedTransaction = SignedTransaction.builder()
                .transaction(transaction)
                .transactionHash("foo")
                .build();

        assertThat(signedTransaction.toBytes()).isEqualTo(transaction.toBytes());
    }

    @Test
    void fromBytes_givenNoExtraData() throws Exception {
        byte[] messageBytes = IOUtils.toByteArray(Objects.requireNonNull(this.getClass().getResourceAsStream("/testdata/domain/broadcast-transactions-cndabrkx.bin")));
        SignedTransaction signedTransaction = SignedTransaction.fromBytes(Arrays.copyOfRange(messageBytes, HEADER_LENGTH, messageBytes.length)); // skip header
        assertThat(signedTransaction.getTransactionHash()).isNull();
        assertThat(signedTransaction.getTransaction()).isNotNull();
        assertThat(signedTransaction.getTransaction().getInputSize()).isZero();
        assertThat(signedTransaction.getTransaction().getExtraData()).isEmpty();
        assertThat(signedTransaction.getSignature()).isEqualTo(Hex.decodeHex("83fef04f8b62e09e18ba1d97ae52221530dfd14f3dbcc4e527a79573be0e2c9662e58b153a694c4baf0ad08efd56b993f3237a79ee054b605b951a09b6922900"));
    }

    @Test
    void fromBytes_givenExtraData() throws Exception {
        byte[] messageBytes = IOUtils.toByteArray(Objects.requireNonNull(this.getClass().getResourceAsStream("/testdata/domain/broadcast-transactions-daomgxxs.bin")));
        SignedTransaction signedTransaction = SignedTransaction.fromBytes(Arrays.copyOfRange(messageBytes, HEADER_LENGTH, messageBytes.length)); // skip header
        assertThat(signedTransaction.getTransactionHash()).isNull();
        assertThat(signedTransaction.getTransaction()).isNotNull();
        assertThat(signedTransaction.getTransaction().getInputSize()).isPositive();
        assertThat(signedTransaction.getTransaction().getExtraData()).isNotEmpty();
        assertThat(signedTransaction.getSignature()).isEqualTo(Hex.decodeHex("8d4c2c037381731f4bcc753810d5563cbe499067aeefa29dfc357631ccb182738e951aa1c6e43091bd9dc4854e7bd64e968546d24c5e422277229a5e7d420f00"));
    }

    @Test
    void toString_printProperties() throws Exception {
        SignedTransaction transaction = SignedTransaction.builder()
                .transactionHash("foo")
                .signature(Hex.decodeHex("ffffff"))
                .build();

        String toString = transaction.toString();
        log.info(toString);
        assertThat(toString).contains("transactionHash=foo");
        assertThat(toString).contains("signature=ffffff");
    }

}