package at.qubic.api.domain.std;

import lombok.extern.slf4j.Slf4j;
import org.apache.commons.codec.binary.Hex;
import org.apache.commons.io.IOUtils;
import org.junit.jupiter.api.Test;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.Arrays;
import java.util.Objects;

import static org.apache.commons.codec.binary.Hex.decodeHex;
import static org.assertj.core.api.Assertions.assertThat;
import static at.qubic.api.domain.QubicHeader.HEADER_LENGTH;

@SuppressWarnings("SpellCheckingInspection")
@Slf4j
class TransactionTest {

    @Test
    void toBytes() throws Exception {
        Transaction transaction = Transaction.builder()
                .sourcePublicKey(decodeHex("bba07fde537ea438760811910f2d874f017974eadcfdde70aa771533aa94064c"))
                .destinationPublicKey(decodeHex("f219230a0b56fbb94b3d73d3879f245390fccc3b23c1fdab0482f59b14a9fe16"))
                .amount(1)
                .tick(11270810)
                .build();

        assertThat(transaction.toBytes()).isEqualTo(Hex.decodeHex("bba07fde537ea438760811910f2d874f017974eadcfdde70aa771533aa94064cf219230a0b56fbb94b3d73d3879f245390fccc3b23c1fdab0482f59b14a9fe1601000000000000009afaab0000000000"));
    }

    @Test
    void toString_thenPrintProperties() throws Exception {

        Transaction transaction = Transaction.builder()
                .sourcePublicKey(decodeHex("bba07fde537ea438760811910f2d874f017974eadcfdde70aa771533aa94064c"))
                .destinationPublicKey(decodeHex("f219230a0b56fbb94b3d73d3879f245390fccc3b23c1fdab0482f59b14a9fe16"))
                .amount(1)
                .tick(42)
                .inputSize((short) 2)
                .inputType((short) 3)
                .build();

        String toString = transaction.toString();
        log.info(toString);
        assertThat(toString).contains("sourcePublicKey=bba07fde537ea438760811910f2d874f017974eadcfdde70aa771533aa94064c");
        assertThat(toString).contains("destinationPublicKey=f219230a0b56fbb94b3d73d3879f245390fccc3b23c1fdab0482f59b14a9fe16");
        assertThat(toString).contains("tick=42");
        assertThat(toString).contains("amount=1");
        assertThat(toString).contains("inputSize=2");
        assertThat(toString).contains("inputType=3");
    }

    @Test
    void fromBytes() throws Exception {
        byte[] messageBytes = IOUtils.toByteArray(Objects.requireNonNull(this.getClass().getResourceAsStream("/testdata/domain/broadcast-transactions-cndabrkx.bin")));

        ByteBuffer buf = ByteBuffer
                .wrap(Arrays.copyOfRange(messageBytes, HEADER_LENGTH, messageBytes.length - 64)) // skip header and exclude signature
                .order(ByteOrder.LITTLE_ENDIAN);

        Transaction transaction = Transaction.fromByteBuffer(buf, false);
        assertThat(transaction.getSourcePublicKey()).isEqualTo(Hex.decodeHex("128eaf1eb10f663f3f9ce13564fc3f7f6da2777022b87923ac68a7a9077800da"));
        assertThat(transaction.getDestinationPublicKey()).isEqualTo(Hex.decodeHex("94b26b1e930f3c3b074029028c7d583eea3e3dcfaacde948bd5b09f73683b30f"));
        assertThat(transaction.getAmount()).isEqualTo(439278011);
        assertThat(transaction.getInputType()).isEqualTo((short) 0);
        assertThat(transaction.getInputSize()).isEqualTo((short) 0);
        assertThat(transaction.getExtraData()).isEmpty();
    }

    @Test
    void fromBytes_givenExtraData() throws Exception {
        byte[] messageBytes = IOUtils.toByteArray(Objects.requireNonNull(this.getClass().getResourceAsStream("/testdata/domain/broadcast-transactions-daomgxxs.bin")));

        ByteBuffer buf = ByteBuffer
                .wrap(Arrays.copyOfRange(messageBytes, HEADER_LENGTH, messageBytes.length - 64)) // skip header and exclude signature
                .order(ByteOrder.LITTLE_ENDIAN);

        Transaction transaction = Transaction.fromByteBuffer(buf, false);
        assertThat(transaction.getSourcePublicKey()).isEqualTo(Hex.decodeHex("c44a7743cb72ecfbda8e56a13fbf1902dae32c2e0d3939d437758a48fa25180a"));
        assertThat(transaction.getDestinationPublicKey()).isEqualTo(Hex.decodeHex("9e1a100cfb556def7bcc6252e47ddf0985428637c3d1b3caa16f33fd98438d94"));
        assertThat(transaction.getAmount()).isEqualTo(0);
        assertThat(transaction.getInputType()).isEqualTo((short) 0);
        assertThat(transaction.getInputSize()).isEqualTo((short) 32);
        assertThat(transaction.getExtraData()).isEqualTo(Hex.decodeHex("eecca563a7f35db009e0345e33e9f86fb24ba6ca4e20c069172869b8b9129cf1"));

    }

    @Test
    void fromBytes_thenIgnoreSignature() throws Exception {
        byte[] messageBytes = IOUtils.toByteArray(Objects.requireNonNull(this.getClass().getResourceAsStream("/testdata/domain/broadcast-transactions-cndabrkx.bin")));

        ByteBuffer buf = ByteBuffer
                .wrap(Arrays.copyOfRange(messageBytes, HEADER_LENGTH, messageBytes.length)) // skip header and include signature
                .order(ByteOrder.LITTLE_ENDIAN);

        Transaction transaction = Transaction.fromByteBuffer(buf, true);
        assertThat(transaction).isNotNull();
    }

}