package at.qubic.api.network;

import at.qubic.api.exception.InvalidResponseException;
import org.junit.jupiter.api.Test;
import at.qubic.api.domain.MessageType;
import at.qubic.api.domain.QubicMessage;
import at.qubic.api.domain.std.request.BroadcastTransaction;
import at.qubic.api.domain.QubicRequest;
import at.qubic.api.util.TestMessageFactory;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;
import reactor.netty.ByteBufFlux;
import reactor.netty.Connection;
import reactor.netty.NettyInbound;
import reactor.netty.NettyOutbound;
import reactor.netty.tcp.TcpClient;
import reactor.netty.tcp.TcpClientConfig;
import reactor.test.StepVerifier;

import java.net.ConnectException;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.nio.ByteBuffer;
import java.time.Duration;
import java.util.concurrent.TimeoutException;
import java.util.function.Supplier;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

class BasicNodeOperationsTest {

    private final TcpClient tcpClient = mock();
    private final Node node = createNode();
    private final BasicNodeOperations nodeOperations = new BasicNodeOperations();

    @Test
    void sendAndReceiveNext() {
        final NettyInbound inbound = mockInbound(ByteBuffer.wrap(messageBytes(MessageType.UNKNOWN, new byte[0])));
        final NettyOutbound outbound = mockOutbound();
        mockConnection(inbound, outbound);

        QubicRequest request = mock();
        QubicMessage message = mock();
        when(request.toMessage()).thenReturn(message);
        when(message.toBytes()).thenReturn(new byte[0]);

        StepVerifier.create(nodeOperations.sendAndReceiveNext(node, Mono.just(request), MessageType.UNKNOWN))
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    void sendAndReceiveNext_givenUnexpectedType_thenThrow() {
        final NettyInbound inbound = mockInbound(ByteBuffer.wrap(messageBytes(MessageType.UNKNOWN, new byte[0])));
        final NettyOutbound outbound = mockOutbound();
        mockConnection(inbound, outbound);

        QubicRequest request = mock();
        QubicMessage message = mock();
        when(request.toMessage()).thenReturn(message);
        when(message.toBytes()).thenReturn(new byte[0]);

        StepVerifier.create(nodeOperations.sendAndReceiveNext(node, Mono.just(request), MessageType.RESPOND_ENTITY))
                .expectError(InvalidResponseException.class)
                .verify();

    }

    @Test
    void sendAndReceiveNext_givenNoExpectedType_thenThrow() {
        final NettyInbound inbound = mockInbound();
        final NettyOutbound outbound = mockOutbound();
        mockConnection(inbound, outbound);

        QubicRequest request = mock();
        QubicMessage message = mock();
        when(request.toMessage()).thenReturn(message);
        when(message.toBytes()).thenReturn(new byte[0]);

        StepVerifier.create(nodeOperations.sendAndReceiveNext(node, Mono.just(request), MessageType.UNKNOWN))
                .expectError(InvalidResponseException.class)
                .verify();
    }

    @Test
    void sendAndReceive() {
        final NettyInbound inbound = mockInbound(
                ByteBuffer.wrap(messageBytes(MessageType.UNKNOWN, new byte[0])),
                ByteBuffer.wrap(messageBytes(MessageType.END_RESPONSE, new byte[0])));
        final NettyOutbound outbound = mockOutbound();
        mockConnection(inbound, outbound);

        QubicRequest request = mock();
        QubicMessage message = mock();
        when(request.toMessage()).thenReturn(message);
        when(message.toBytes()).thenReturn(new byte[0]);

        StepVerifier.create(nodeOperations.sendAndReceive(node, Mono.just(request), MessageType.UNKNOWN))
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    void sendAndReceive_givenRuntimeException_thenError() {
        when(tcpClient.connect()).then(x -> Mono.error(new RuntimeException("test")));

        QubicRequest request = mock();
        QubicMessage message = mock();
        when(request.toMessage()).thenReturn(message);
        when(message.toBytes()).thenReturn(new byte[0]);

        StepVerifier.create(nodeOperations.sendAndReceive(node, Mono.just(request), null))
                .expectError(RuntimeException.class)
                .verify();
    }

    @Test
    void sendAndReceive_givenTimeout_thenError() {
        when(tcpClient.connect()).then(x -> Mono.delay(Duration.ofSeconds(30)));

        QubicRequest request = mock();
        QubicMessage message = mock();
        when(request.toMessage()).thenReturn(message);
        when(message.toBytes()).thenReturn(new byte[0]);

        StepVerifier.withVirtualTime(() -> nodeOperations.sendAndReceive(node, Mono.just(request), null))
                .expectSubscription()
                .expectNoEvent(Duration.ofSeconds(2))
                .expectError(TimeoutException.class)
                .verify(Duration.ofSeconds(1));
    }

    @Test
    void send() {
        Connection connection = mock();
        when(tcpClient.handle(any())).then(args -> node.getClient());
        when(tcpClient.connect()).then(args -> Mono.just(connection));

        StepVerifier.create(nodeOperations.send(node, Mono.just(BroadcastTransaction.builder().build())))
                .assertNext(c -> assertThat(c).isEqualTo(connection))
                .verifyComplete();
    }

    @Test
    void send_givenConnectException_thenError() {
        when(tcpClient.handle(any())).then(args -> node.getClient());
        when(tcpClient.connect()).then(args -> Mono.error(new ConnectException("test")));

        StepVerifier.create(nodeOperations.send(node, Mono.just(BroadcastTransaction.builder().build())))
                .expectError(ConnectException.class)
                .verify();
    }

    private Node createNode() {
        TcpClientConfig config = mock();
        when(tcpClient.configuration()).thenReturn(config);
        when(config.remoteAddress()).then(args -> (Supplier<SocketAddress>) () -> new InetSocketAddress("localhost", 666));
        return new Node(tcpClient);
    }


    private void mockConnection(NettyInbound inbound, NettyOutbound outbound) {
        Connection connection = mock();
        when(connection.inbound()).thenReturn(inbound);
        when(connection.outbound()).thenReturn(outbound);
        when(tcpClient.connect()).then(x -> Mono.just(connection));
    }

    private static NettyOutbound mockOutbound() {
        NettyOutbound outbound = mock();
        when(outbound.sendByteArray(any())).thenReturn(outbound);
        when(outbound.then()).thenReturn(Mono.empty());
        return outbound;
    }

    private static NettyInbound mockInbound(ByteBuffer... payloads) {
        NettyInbound inbound = mock();
        ByteBufFlux byteBufFlux = mock();
        when(byteBufFlux.asByteBuffer()).thenReturn(Flux.just(payloads));
        when(inbound.receive()).thenReturn(byteBufFlux);
        return inbound;
    }

    private static byte[] messageBytes(MessageType messageType, byte[] payload) {
        return TestMessageFactory.createMessage(messageType, payload).toBytes();
    }

}