package at.qubic.api.network;

import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.Test;
import at.qubic.api.domain.std.response.TickInfo;
import at.qubic.api.properties.NetworkProperties;
import org.mockito.Mockito;
import reactor.core.publisher.Mono;
import reactor.test.StepVerifier;

import java.time.Duration;
import java.util.concurrent.atomic.AtomicInteger;

import static org.assertj.core.api.Assertions.assertThat;
import static org.mockito.Mockito.mock;

@Slf4j
class FixNodesStrategyTest {

    private final NetworkStatus networkStatus = new NoopNetworkStatus(); // noop because we do not want to test the base class too much
    private final NetworkProperties networkProperties = new NetworkProperties(666, 3, Duration.ofSeconds(10));
    private final FixNodesStrategy fixNodes = new FixNodesStrategy(networkStatus, networkProperties);

    @Test
    void getNodes_givenOneWorkingNode_thenUseIt() {

        fixNodes.addNode(mockNode("1", TickInfo.builder().tick(42).epoch((short) 43).build()));
        fixNodes.addNode(mockNode("2", TickInfo.builder().tick(42).build())); // invalid
        fixNodes.addNode(mockNode("3", TickInfo.builder().epoch((short) 43).build())); // invalid
        fixNodes.addNode(mockNode("4", TickInfo.builder().build())); // invalid

        for (int i = 0; i < 10; i++) {
            StepVerifier.create(fixNodes.getNodes(1))
                    .assertNext(n -> assertThat(n.getName()).isEqualTo("1"))
                    .verifyComplete();
        }

    }

    @Test
    void getNodes_givenTwoCurrentNodes_thenRandomCurrentNode() {
        fixNodes.addNode(mockNode("1", TickInfo.builder().tick(42).epoch((short) 43).build())); // most current
        fixNodes.addNode(mockNode("2", TickInfo.builder().tick(1).epoch((short) 1).build())); // old but valid
        fixNodes.addNode(mockNode("3", TickInfo.builder().build())); // invalid
        fixNodes.addNode(mockNode("4", TickInfo.builder().build())); // invalid

        final AtomicInteger first = new AtomicInteger(0);

        for (int i = 0; i < 100; i++) {

            StepVerifier.create(fixNodes.getNodes(1)
                            .doOnNext(n -> {
                                if (StringUtils.equals(n.getName(), "1")) {
                                    first.incrementAndGet();
                                }
                            }))
                    .expectNextCount(1)
                    .verifyComplete();

        }
        assertThat(first.get()).isBetween(20, 80);
    }

    @Test
    void getNodes_givenNoCurrentNode_thenRandomNode() {
        fixNodes.addNode("1"); // invalid tick info.
        fixNodes.addNode("2"); // invalid tick info.

        final AtomicInteger first = new AtomicInteger(0);

        for (int i = 0; i < 100; i++) {

            StepVerifier.create(fixNodes.getNodes(1)
                            .doOnNext(n -> {
                                if (StringUtils.equals(n.getName(), "1")) {
                                    first.incrementAndGet();
                                }
                            }))
                    .expectNextCount(1)
                    .verifyComplete();

        }
        assertThat(first.get()).isBetween(20, 80);
    }

    private static Node mockNode(String name, TickInfo tickInfo) {
        Node node = mock();
        Mockito.when(node.getName()).thenReturn(name);
        Mockito.when(node.getCurrentTickInfo()).thenReturn(Mono.just(tickInfo));
        Mockito.when(node.getTickInfo()).thenReturn(tickInfo);
        return node;
    }

}