package at.qubic.api.network;

import org.junit.jupiter.api.Test;
import at.qubic.api.domain.std.response.TickInfo;
import org.mockito.Mockito;
import reactor.core.publisher.Mono;
import reactor.test.StepVerifier;

import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.ArrayList;
import java.util.List;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.within;
import static org.mockito.Mockito.*;

class HealthCheckNetworkStatusTest {

    private static final TickStatus INITIAL_TICK_STATUS = TickStatus.builder()
            .tickInfo(TickInfo.builder().epoch((short) 1).tick(1).build())
            .latestUpdate(Instant.EPOCH)
            .build();

    private final HealthCheckNetworkStatus networkStatus = new HealthCheckNetworkStatus();

    @Test
    void getTickStatus() {
        StepVerifier.create(networkStatus.getTickStatus())
                .expectNext(INITIAL_TICK_STATUS)
                .verifyComplete();
    }

    @Test
    void update_givenNewTickInfo_ThenUpdate() {
        final TickInfo newTickInfo = TickInfo.builder().epoch((short) 2).tick(42).build();

        Node node = mockNode("1", newTickInfo);

        StepVerifier.create(networkStatus.update(List.of(node)))
                .expectNext(newTickInfo)
                .verifyComplete();

        StepVerifier.create(networkStatus.getTickStatus())
                .assertNext(tickStatus -> {
                    assertThat(tickStatus.getTickInfo()).isEqualTo(newTickInfo);
                    assertThat(tickStatus.getLatestUpdate()).isCloseTo(Instant.now(), within(1, ChronoUnit.SECONDS));
                })
                .verifyComplete();
    }

    @Test
    void update_givenTickInfos_ThenUpdateToLatest() {
        final TickInfo newTickInfo = TickInfo.builder().epoch((short) 2).tick(42).build();
        final TickInfo newestTickInfo = TickInfo.builder().epoch((short) 2).tick(43).build();

        Node node1 = mockNode("new", newTickInfo);
        Node node2 = mockNode("newest", newestTickInfo);
        Node node3 = mockNode("old", INITIAL_TICK_STATUS.getTickInfo());
        Node node4 = mockNode("empty", TickInfo.builder().build());

        StepVerifier.create(networkStatus.update(List.of(node1, node2, node3, node4)))
                .expectNext(newestTickInfo)
                .verifyComplete();

        StepVerifier.create(networkStatus.getTickStatus())
                .assertNext(tickStatus -> {
                    assertThat(tickStatus.getTickInfo()).isEqualTo(newestTickInfo);
                    assertThat(tickStatus.getLatestUpdate()).isCloseTo(Instant.now(), within(1, ChronoUnit.SECONDS));
                })
                .verifyComplete();
    }

    @Test
    void update_givenSameTickInfo_ThenUpdateTimestampOnly() {
        Node node = mockNode("1", INITIAL_TICK_STATUS.getTickInfo());
        List<Node> nodes = new ArrayList<>();
        nodes.add(node);

        StepVerifier.create(networkStatus.update(nodes))
                .expectNext(INITIAL_TICK_STATUS.getTickInfo())
                .verifyComplete();

        StepVerifier.create(networkStatus.getTickStatus())
                .assertNext(tickStatus -> {
                    assertThat(tickStatus.getTickInfo()).isEqualTo(INITIAL_TICK_STATUS.getTickInfo()); // not updated
                    assertThat(tickStatus.getLatestUpdate()).isCloseTo(Instant.now(), within(1, ChronoUnit.SECONDS)); // updated
                })
                .verifyComplete();
    }

    @Test
    void update_givenOlderTickInfo_ThenDoNotUpdate() {
        final TickInfo emptyTickInfo = TickInfo.builder().build();
        Node node = mockNode("1", emptyTickInfo);
        List<Node> nodes = new ArrayList<>();
        nodes.add(node);

        StepVerifier.create(networkStatus.update(nodes))
                .expectNext(emptyTickInfo)
                .verifyComplete();

        StepVerifier.create(networkStatus.getTickStatus())
                .expectNext(INITIAL_TICK_STATUS) // not updated including time stamp
                .verifyComplete();
    }

    private static Node mockNode(String name, TickInfo tickInfo) {
        Node node = mock();
        Mockito.when(node.getName()).thenReturn(name);
        Mockito.when(node.getCurrentTickInfo()).thenReturn(Mono.just(tickInfo));
        Mockito.when(node.getTickInfo()).thenReturn(tickInfo);
        return node;
    }

}