package at.qubic.api.network;

import org.apache.commons.lang3.RandomStringUtils;
import org.junit.jupiter.api.Test;
import at.qubic.api.domain.std.response.TickInfo;
import reactor.core.publisher.Mono;
import reactor.test.StepVerifier;

import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.List;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.within;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

class NetworkStatusTest {

    private final NetworkStatus networkStatus = new HealthCheckNetworkStatus();

    @Test
    void update() {

        final TickInfo defaultTick = TickInfo.builder().build();
        final TickInfo latestTick = TickInfo.builder().epoch((short) 3).tick(42).build();
        final TickInfo wrongEpoch = TickInfo.builder().epoch((short) 2).tick(43).build();

        final Node node1 = mockNode(defaultTick);
        final Node node2 = mockNode(latestTick);
        final Node node3 = mockNode(wrongEpoch);

        StepVerifier.create(networkStatus.update(List.of(node1, node2, node3)))
                .expectNext(latestTick)
                .verifyComplete();

        StepVerifier.create(networkStatus.getTickStatus())
                .assertNext(ts -> {
                    assertThat(ts.getTickInfo().getEpoch()).isEqualTo((short) 3);
                    assertThat(ts.getTickInfo().getTick()).isEqualTo(42);
                    assertThat(ts.getLatestUpdate()).isCloseTo(Instant.now(), within(1, ChronoUnit.SECONDS));
                })
                .verifyComplete();

    }

    @Test
    void updateLatestTick_givenEmpty_thenDoUpdateTo() {
        final TickInfo latestTick = TickInfo.builder().epoch((short) 3).tick(42).build();

        Node node1 = mockNode(TickInfo.builder().build());
        when(node1.getCurrentTickInfo()).thenReturn(Mono.empty());
        final Node node2 = mockNode(latestTick);

        StepVerifier.create(networkStatus.update(List.of(node1, node2)))
                .expectNext(latestTick)
                .verifyComplete();
    }

    @Test
    void updateLatestTick_givenError_thenDoUpdateTo() {
        final TickInfo latestTick = TickInfo.builder().epoch((short) 3).tick(42).build();

        Node node1 = mockNode(TickInfo.builder().build());
        when(node1.getCurrentTickInfo()).thenReturn(Mono.error(new RuntimeException("test")));
        final Node node2 = mockNode(latestTick);

        StepVerifier.create(networkStatus.update(List.of(node1, node2)))
                .expectNext(latestTick)
                .verifyComplete();
    }

    private static Node mockNode(TickInfo tickInfo) {
        Node node = mock();
        when(node.getName()).thenReturn(RandomStringUtils.insecure().nextAlphabetic(10));
        when(node.getCurrentTickInfo()).thenReturn(Mono.just(tickInfo));
        when(node.getTickInfo()).thenReturn(tickInfo);
        return node;
    }

}