package at.qubic.api.network;

import at.qubic.api.domain.std.request.RequestAssets.RequestType;
import at.qubic.api.domain.std.request.RequestAssetsByFilter;
import at.qubic.api.domain.std.response.*;
import at.qubic.api.exception.InvalidResponseException;
import at.qubic.api.util.AssetUtil;
import lombok.SneakyThrows;
import org.apache.commons.codec.DecoderException;
import org.apache.commons.codec.binary.Hex;
import org.awaitility.Awaitility;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.Test;
import at.qubic.api.domain.MessageType;
import at.qubic.api.domain.QubicHeader;
import at.qubic.api.domain.QubicMessage;
import at.qubic.api.domain.qx.request.*;
import at.qubic.api.domain.qx.response.AssetOrder;
import at.qubic.api.domain.qx.response.EntityOrder;
import at.qubic.api.domain.qx.response.QxFees;
import at.qubic.api.domain.std.SignedTransaction;
import at.qubic.api.domain.std.Transaction;
import at.qubic.api.tcp.TcpTestServer;
import at.qubic.api.util.TestFileUtil;
import reactor.core.publisher.Mono;
import reactor.test.StepVerifier;

import java.io.IOException;
import java.time.Duration;

import static org.assertj.core.api.Assertions.assertThat;
import static at.qubic.api.domain.MessageType.BROADCAST_FUTURE_TICK_DATA;
import static at.qubic.api.domain.MessageType.REQUEST_ENTITY;
import static at.qubic.api.util.QubicHeaderUtil.headerFromBytes;
import static at.qubic.api.util.TestMessageFactory.createMessage;

class NodeIT {

    private final TcpTestServer tcpServer = new TcpTestServer(6666).start();
    private final Node node = new Node("localhost", 6666);

    @AfterEach
    void tearDown() throws IOException {
        tcpServer.stop();
    }

    @Test
    void getEntity_ThenRetrieveEntity() throws Exception {

        final Entity entity = Entity.builder()
                .publicKey(Hex.decodeHex("30df4cb7d17334b6a84488d631a1f73a9bcd5420f00209cb91c90c251e4d77b4"))
                .incomingAmount(100_000_000_000L)
                .numberOfIncomingTransfers(1)
                .latestIncomingTransferTick(123456)
                .spectrum(new byte[24][32])
                .build();

        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        final QubicMessage respondEntity = createMessage(MessageType.RESPOND_ENTITY, entity.toByteBuffer().array());

        TcpTestServer.ExchangeThread server = tcpServer.respond(exchangePublicPeers.toBytes(), respondEntity.toBytes());
        StepVerifier.create(node.getEntity(Mono.just(new byte[32])))
                .expectNext(entity)
                .verifyComplete();

        assertThat(headerFromBytes(server.getHeader()).getType()).isEqualTo(REQUEST_ENTITY);
    }



    @Test
    void getCurrentTickInfo_thenRetrieveTickInfo() {

        TickInfo tickInfo = TickInfo.builder()
                .tickDuration((short) 1)
                .epoch((short) 2)
                .tick(3)
                .numberOfAlignedVotes((short) 4)
                .numberOfMisalignedVotes((short) 5)
                .initialTick(6)
                .build();
        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        final QubicMessage respondCurrentTickInfo = createMessage(MessageType.RESPOND_CURRENT_TICK_INFO, tickInfo.toByteBuffer().array());

        tcpServer.respond(exchangePublicPeers.toBytes(), respondCurrentTickInfo.toBytes());

        StepVerifier.create(node.getCurrentTickInfo().log())
                .assertNext(ti -> {
                    assertThat(ti.getTickDuration()).isEqualTo((short) 1);
                    assertThat(ti.getEpoch()).isEqualTo((short) 2);
                    assertThat(ti.getTick()).isEqualTo(3);
                    assertThat(ti.getNumberOfAlignedVotes()).isEqualTo((short) 4);
                    assertThat(ti.getNumberOfMisalignedVotes()).isEqualTo((short) 5);
                    assertThat(ti.getInitialTick()).isEqualTo(6);
                })
                .verifyComplete();
     }

    @Test
    void broadcastTransaction_ThenReturnTransactionHash() {
        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        final QubicMessage currentTickInfo = createMessage(MessageType.RESPOND_CURRENT_TICK_INFO, TickInfo.builder().tick(42).build().toByteBuffer().array());
        tcpServer.respond(exchangePublicPeers.toBytes(), currentTickInfo.toBytes());

        SignedTransaction signedTransaction = SignedTransaction.builder()
                .transaction(Transaction.builder().build())
                .build();

        StepVerifier.create(node.broadcastTransaction(Mono.just(signedTransaction).log()))
                .expectNext(signedTransaction)
                .verifyComplete();
    }

    @Test
    void getTickData() {
        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        final QubicMessage tickData = getBroadcastFutureTickData();
        tcpServer.respond(exchangePublicPeers.toBytes(), tickData.toBytes());

        StepVerifier.create(node.getTickData(42))
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    void getTickData_GivenEmpty_ThenThrow() {
        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        tcpServer.respond(exchangePublicPeers.toBytes(), new byte[0]); // empty response

        StepVerifier.create(node.getTickData(42))
                .expectError(InvalidResponseException.class)
                .verify();
    }

    @Test
    void getTickTransactions() {

        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        final QubicMessage endResponse = createMessage(MessageType.END_RESPONSE);
        byte[] tx1 = TestFileUtil.loadBinaryFile("/testdata/domain/broadcast-transactions-daomgxxs.bin");
        byte[] tx2 = TestFileUtil.loadBinaryFile("/testdata/domain/broadcast-transactions-cndabrkx.bin");
        byte[] tx3 = TestFileUtil.loadBinaryFile("/testdata/domain/broadcast-transactions-lgdvbraz.bin");
        tcpServer.respond(exchangePublicPeers.toBytes(), tx1, tx2, tx3, endResponse.toBytes());

        StepVerifier.create(node.getTickTransactions(42))
                .assertNext(tx -> assertThat(tx.getSignature()).isEqualTo(fromHex("8d4c2c037381731f4bcc753810d5563cbe499067aeefa29dfc357631ccb182738e951aa1c6e43091bd9dc4854e7bd64e968546d24c5e422277229a5e7d420f00")))
                .assertNext(tx -> assertThat(tx.getSignature()).isEqualTo(fromHex("83fef04f8b62e09e18ba1d97ae52221530dfd14f3dbcc4e527a79573be0e2c9662e58b153a694c4baf0ad08efd56b993f3237a79ee054b605b951a09b6922900")))
                .assertNext(tx -> assertThat(tx.getSignature()).isEqualTo(fromHex("d5a13558e589f0b1fbec4dc7ebcdeb24b62c86434c003e3375ee77b29e3d21dbe1187343965317a2eeca4cc5033983f33f8126d3aab19aefddd6363a46bb0b00")))
                .verifyComplete();
    }

    @Test
    void getTickTransactions_givenNoEndResponse_thenError() {
        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        byte[] tx1 = TestFileUtil.loadBinaryFile("/testdata/domain/broadcast-transactions-daomgxxs.bin");
        tcpServer.respond(exchangePublicPeers.toBytes(), tx1);

        StepVerifier.create(node.getTickTransactions(42))
                .expectError(InvalidResponseException.class)
                .verify();
    }

    @Test
    void getTickTransactions_givenNoElements_thenError() {
        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        tcpServer.respond(exchangePublicPeers.toBytes(), new byte[0]);

        StepVerifier.create(node.getTickTransactions(42))
                .expectError(InvalidResponseException.class)
                .verify();
    }

    @Test
    void getTickTransactions_givenNoTransactions_thenNoResult() {

        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        final QubicMessage endResponse = createMessage(MessageType.END_RESPONSE);
        tcpServer.respond(exchangePublicPeers.toBytes(), endResponse.toBytes());

        StepVerifier.create(node.getTickTransactions(42))
                .expectNextCount(0)
                .verifyComplete();

    }

    @Test
    void getOwnedAssets() {
        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        final QubicMessage endResponse = createMessage(MessageType.END_RESPONSE);
        byte[] qx = TestFileUtil.loadHexFile("/testdata/domain/respond-owned-asset-QX.hex");
        byte[] random = TestFileUtil.loadHexFile("/testdata/domain/respond-owned-asset-RANDOM.hex");
        byte[] mlm = TestFileUtil.loadHexFile("/testdata/domain/respond-owned-asset-MLM.hex");
        byte[] cfb = TestFileUtil.loadHexFile("/testdata/domain/respond-owned-asset-CFB.hex");
        tcpServer.respond(exchangePublicPeers.toBytes(), qx, random, mlm, cfb, endResponse.toBytes());

        StepVerifier.create(node.getOwnedAssets(Mono.just(new byte[32])))
                .assertNext(roa -> assertThat(new String(roa.getIssuanceAsset().getName()).trim()).isEqualTo("QX"))
                .assertNext(roa -> assertThat(new String(roa.getIssuanceAsset().getName()).trim()).isEqualTo("RANDOM"))
                .assertNext(roa -> assertThat(new String(roa.getIssuanceAsset().getName()).trim()).isEqualTo("MLM"))
                .assertNext(roa -> assertThat(new String(roa.getIssuanceAsset().getName()).trim()).isEqualTo("CFB"))
                .verifyComplete();
    }

    @Test
    void getPossessedAssets() {

        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        final QubicMessage endResponse = createMessage(MessageType.END_RESPONSE);
        byte[] qx = TestFileUtil.loadHexFile("/testdata/domain/respond-possessed-asset-QX.hex");
        byte[] cfb = TestFileUtil.loadHexFile("/testdata/domain/respond-possessed-asset-CFB.hex");
        tcpServer.respond(exchangePublicPeers.toBytes(), cfb, qx, endResponse.toBytes());

        StepVerifier.create(node.getPossessedAssets(Mono.just(new byte[32])))
                .assertNext(roa -> assertThat(new String(roa.getIssuanceAsset().getName()).trim()).isEqualTo("CFB"))
                .assertNext(roa -> assertThat(new String(roa.getIssuanceAsset().getName()).trim()).isEqualTo("QX"))
                .verifyComplete();
    }

    @Test
    void getQxFees() {
        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        byte[] qx = TestFileUtil.loadHexFile("/testdata/domain/response-qx-fees.hex");
        tcpServer.respond(exchangePublicPeers.toBytes(), qx);

        StepVerifier.create(node.getQxFees())
                .expectNext(QxFees.builder()
                        .assetIssuanceFee(1_000_000_000)
                        .transferFee(1_000_000)
                        .tradeFee(5_000_000)
                        .build())
                .verifyComplete();
    }

    @Test
    void getAssetAskOrders() {
        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        byte[] asks = TestFileUtil.loadHexFile("/testdata/domain/response-asset-ask-orders-cfb.hex");
        tcpServer.respond(exchangePublicPeers.toBytes(), asks);

        QxGetAssetOrders request = new QxGetAssetAskOrders(new byte[32], "FOO", 8);
        StepVerifier.create(node.getQxAssetOrders(request))
                .assertNext(list -> {
                    assertThat(list).hasSize(80);
                    assertThat(list.getFirst()).isEqualTo(AssetOrder.builder()
                            .entity(fromHex("c24a2290cfeca22a9a8db36288f23b5895b70748a2322e048b37fa8872278f8d"))
                            .price(5)
                            .numberOfShares(33482584)
                            .build());
                })
                .verifyComplete();
    }

    @Test
    void getAssetBidOrders() {
        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        byte[] bids = TestFileUtil.loadHexFile("/testdata/domain/response-asset-bid-orders-cfb.hex");
        tcpServer.respond(exchangePublicPeers.toBytes(), bids);

        QxGetAssetOrders request = new QxGetAssetBidOrders(new byte[32], "FOO", 0);
        StepVerifier.create(node.getQxAssetOrders(request))
                .assertNext(list -> {
                    assertThat(list).hasSize(33);
                    assertThat(list.getFirst()).isEqualTo(AssetOrder.builder()
                            .entity(fromHex("069593064b22d9996b656bd632fd9e100e6ac992dfb201204cd3f10fd66ce3ff"))
                            .price(3)
                            .numberOfShares(1570495966)
                            .build());
                })
                .verifyComplete();
    }

    @Test
    void getEntityAskOrders() {
        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        byte[] asks = TestFileUtil.loadHexFile("/testdata/domain/response-entity-ask-orders.hex");
        tcpServer.respond(exchangePublicPeers.toBytes(), asks);

        QxGetEntityOrders request = new QxGetEntityAskOrders(new byte[32], 0);
        StepVerifier.create(node.getQxEntityOrders(request))
                .assertNext(list -> {
                    assertThat(list).hasSize(18);
                    assertThat(list.get(2)).isEqualTo(EntityOrder.builder()
                                    .issuer(new byte[32])
                                    .assetName(new byte[] {'M', 'L', 'M', 0, 0, 0, 0, 0})
                                    .price(700_000_000)
                                    .numberOfShares(2)
                            .build());
                })
                .verifyComplete();
    }

    @Test
    void getEntityBidOrders() {

        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        byte[] bids = TestFileUtil.loadHexFile("/testdata/domain/response-entity-bid-orders.hex");
        tcpServer.respond(exchangePublicPeers.toBytes(), bids);

        QxGetEntityOrders request = new QxGetEntityBidOrders(new byte[32], 0);
        StepVerifier.create(node.getQxEntityOrders(request))
                .assertNext(list -> {
                    assertThat(list).hasSize(255);
                    assertThat(list.getFirst()).isEqualTo(EntityOrder.builder()
                            .issuer(new byte[32])
                            .assetName(new byte[] {'Q', 'X', 0, 0, 0, 0, 0, 0})
                            .price(5_000_000_000L)
                            .numberOfShares(1)
                            .build());
                })
                .verifyComplete();
    }

    @Test
    void getIssuedAssetByUniverseIndex() {
        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        byte[] response = TestFileUtil.loadHexFile("/testdata/domain/respond-issued-assets-by-universe-index.hex");
        tcpServer.respond(exchangePublicPeers.toBytes(), response);

        AssetIssuance expected = AssetIssuance.builder()
                .name(AssetUtil.nameTo7Bytes("RANDOM"))
                .publicKey(new byte[32])
                .type((byte) 1)
                .numberOfDecimalPlaces((byte) 0)
                .unitOfMeasurement(AssetUtil.unitOfMeasurementTo7Bytes("0000000"))
                .build();

        StepVerifier.create(node.getIssuedAssetByUniverseIndex(3))
                .assertNext(ria -> {
                    assertThat(ria.getAsset()).isEqualTo(expected);
                    assertThat(ria.getUniverseIndex()).isEqualTo(3);
                    assertThat(ria.getTick()).isEqualTo(20117803);
                })
                .verifyComplete();

    }

    @Test
    void getIssuedAssetsByFilter() {
        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        final QubicMessage endResponse = createMessage(MessageType.END_RESPONSE);
        byte[] response = TestFileUtil.loadHexFile("/testdata/domain/respond-issued-assets-by-filter.hex");
        tcpServer.respond(exchangePublicPeers.toBytes(), response, endResponse.toBytes());

        AssetIssuance expected = AssetIssuance.builder()
                .name(AssetUtil.nameTo7Bytes("RANDOM"))
                .publicKey(new byte[32])
                .type((byte) 1)
                .numberOfDecimalPlaces((byte) 0)
                .unitOfMeasurement(AssetUtil.unitOfMeasurementTo7Bytes("0000000"))
                .build();

        RequestAssetsByFilter request = RequestAssetsByFilter.builder()
                .assetRequestType(RequestType.REQUEST_TYPE_ISSUANCE_RECORDS)
                .build();

        StepVerifier.create(node.getIssuedAssetsByFilter(request))
                .assertNext(ria -> {
                    assertThat(ria.getAsset()).isEqualTo(expected);
                    assertThat(ria.getUniverseIndex()).isEqualTo(3);
                    assertThat(ria.getTick()).isEqualTo(20117803);
                })
                .verifyComplete();
    }

    @Test
    void getOwnedAssetsByFilter() throws DecoderException {
        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        final QubicMessage endResponse = createMessage(MessageType.END_RESPONSE);
        // returns 2 response asset headers and payload
        byte[] response = TestFileUtil.loadHexFile("/testdata/domain/respond-owned-assets-by-filter.hex");
        tcpServer.respond(exchangePublicPeers.toBytes(), response, endResponse.toBytes());

        AssetOwnership expected = AssetOwnership.builder()
                .publicKey(Hex.decodeHex("26843cfa419ce5ae9ed299757e57234eb50770a9bf711a103e8a068cf34ddffa"))
                .type((byte) 2)
                .numberOfUnits((byte) 1)
                .managingContractIndex((short) 1)
                .issuanceIndex(3)
                .build();

        RequestAssetsByFilter request = RequestAssetsByFilter.builder()
                .assetRequestType(RequestType.REQUEST_TYPE_OWNERSHIP_RECORDS)
                .issuer(new byte[32])
                .assetName("RANDOM")
                .build();

        StepVerifier.create(node.getOwnedAssetsByFilter(request))
                .assertNext(ria -> {
                    assertThat(ria.getAsset()).isEqualTo(expected);
                    assertThat(ria.getUniverseIndex()).isEqualTo(3965990);
                    assertThat(ria.getTick()).isEqualTo(20239283);
                })
                .expectNextCount(1) // one more, not validated here
                .verifyComplete();
    }

    @Test
    void getPossessedAssetsByFilter() throws DecoderException {
        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        final QubicMessage endResponse = createMessage(MessageType.END_RESPONSE);
        // returns 2 response asset headers and payload
        byte[] response = TestFileUtil.loadHexFile("/testdata/domain/respond-possessed-assets-by-filter.hex");
        tcpServer.respond(exchangePublicPeers.toBytes(), response, endResponse.toBytes());

        AssetPossession expected = AssetPossession.builder()
                .publicKey(Hex.decodeHex("26843cfa419ce5ae9ed299757e57234eb50770a9bf711a103e8a068cf34ddffa"))
                .type((byte) 3)
                .numberOfUnits((byte) 1)
                .managingContractIndex((short) 1)
                .ownershipIndex(3965990)
                .build();

        RequestAssetsByFilter request = RequestAssetsByFilter.builder()
                .assetRequestType(RequestType.REQUEST_TYPE_POSSESSION_RECORDS)
                .issuer(new byte[32])
                .assetName("RANDOM")
                .build();

        StepVerifier.create(node.getPossessedAssetsByFilter(request))
                .assertNext(ria -> {
                    assertThat(ria.getAsset()).isEqualTo(expected);
                    assertThat(ria.getUniverseIndex()).isEqualTo(3965991);
                    assertThat(ria.getTick()).isEqualTo(20239617);
                })
                .expectNextCount(1) // one more, not validated here
                .verifyComplete();
    }

    @Test
    void broadcastTransaction() {
        TcpTestServer.ExchangeThread exchange = this.tcpServer.respond(getExchangePublicPeers().toBytes()); // dummy we only send

        SignedTransaction stx = SignedTransaction.builder()
                .transaction(Transaction.builder().build())
                .build();

        StepVerifier.create(node.broadcastTransaction(Mono.just(stx)))
                .expectNext(stx)
                .verifyComplete();

        Awaitility.await().pollInterval(Duration.ofMillis(10)).until(exchange::isReceivedPayload);
        assertThat(headerFromBytes(exchange.getHeader()).getType()).isEqualTo(MessageType.BROADCAST_TRANSACTION);
    }

    @SneakyThrows
    private byte[] fromHex(String hex) {
        return Hex.decodeHex(hex);
    }

    private QubicMessage getBroadcastFutureTickData() {
        final byte[] payload = TestFileUtil.loadHexFile("/testdata/domain/broadcast-future-tick-data-payload.hex");
        return QubicMessage.builder()
                .header(QubicHeader.builder().size(41080).type(BROADCAST_FUTURE_TICK_DATA).build())
                .payload(payload)
                .build();
    }

    private QubicMessage getExchangePublicPeers() {
        return createMessage(
                MessageType.EXCHANGE_PUBLIC_PEERS,
                ExchangePublicPeers.builder()
                        .peers(new String[] { "1.2.3.4", "2.3.4.1", "3.4.1.2", "4.1.2.3" })
                        .build()
                        .toBytes()
                        .array());
    }

}