package at.qubic.api.network;

import at.qubic.api.domain.qearn.request.*;
import at.qubic.api.domain.std.request.RequestAssets;
import at.qubic.api.domain.std.request.RequestAssetsByFilter;
import at.qubic.api.domain.std.response.AssetIssuance;
import at.qubic.api.util.AssetUtil;
import lombok.extern.slf4j.Slf4j;
import org.junit.jupiter.api.Tag;
import org.junit.jupiter.api.Test;
import at.qubic.api.SpringBootTestConfiguration;
import at.qubic.api.crypto.IdentityUtil;
import at.qubic.api.domain.qx.request.*;
import at.qubic.api.domain.qx.response.QxFees;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import reactor.core.publisher.Mono;
import reactor.test.StepVerifier;

import static org.assertj.core.api.Assertions.assertThat;

@SuppressWarnings({"SpringBootApplicationProperties", "LoggingSimilarMessage"})
@Slf4j
@Tag("SIT")
@SpringBootTest(classes = SpringBootTestConfiguration.class, properties = { "use.public.peers=false", "check.health=true" })
class NodeSIT {

    @Autowired
    private IdentityUtil identityUtil;

    @Autowired
    private Nodes nodes;

    @Test
    void getTickInfo() {
        StepVerifier.create(nodes.getNode()
                        .flatMap(n -> n.getCurrentTickInfo())
                        .doOnNext(ti -> log.info("{}", ti)))
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    void getQxFees() {
        QxFees expected = QxFees.builder()
                .transferFee(100)
                .assetIssuanceFee(1_000_000_000)
                .tradeFee(3_000_000)
                .build();

        StepVerifier.create(nodes.getNode().flatMap(Node::getQxFees))
                .assertNext(qxFee -> assertThat(qxFee).isEqualTo(expected))
                .verifyComplete();
    }

    @Test
    void getEntity() {
        byte[] entityId = identityUtil.getPublicKeyFromIdentity("CFBMEMZOIDEXQAUXYYSZIURADQLAPWPMNJXQSNVQZAHYVOPYUKKJBJUCTVJL");
        StepVerifier.create(nodes.getNode()
                        .flatMap(node -> node.getEntity(Mono.just(entityId))))
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    void getOwnedAssets() {
        byte[] entityId = identityUtil.getPublicKeyFromIdentity("CFBMEMZOIDEXQAUXYYSZIURADQLAPWPMNJXQSNVQZAHYVOPYUKKJBJUCTVJL");
        StepVerifier.create(nodes.getNode()
                        .flatMapMany(node -> node.getOwnedAssets(Mono.just(entityId))))
                .expectNextCount(2)
                .verifyComplete();
    }

    @Test
    void getPossessedAssets() {
        byte[] entityId = identityUtil.getPublicKeyFromIdentity("CFBMEMZOIDEXQAUXYYSZIURADQLAPWPMNJXQSNVQZAHYVOPYUKKJBJUCTVJL");
        StepVerifier.create(nodes.getNode()
                        .flatMapMany(node -> node.getPossessedAssets(Mono.just(entityId))))
                .expectNextCount(2)
                .verifyComplete();
    }

    @Test
    void getQxAssetAskOrders() {
        byte[] issuer = identityUtil.getPublicKeyFromIdentity("ABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDVGRF");
        QxGetAssetOrders request = new QxGetAssetAskOrders(issuer, "CFB", 0);
        StepVerifier.create(nodes.getNode().flatMap(n -> n.getQxAssetOrders(request)
                        .doOnNext(orders -> orders.forEach(o -> log.info(o.toString())))))
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    void getQxAssetBidOrders() {
        byte[] issuer = identityUtil.getPublicKeyFromIdentity("ABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDVGRF");
        QxGetAssetOrders request = new QxGetAssetBidOrders(issuer, "CFB", 0);
        StepVerifier.create(nodes.getNode().flatMap(n -> n.getQxAssetOrders(request))
                        .doOnNext(orders -> orders.forEach(o -> log.info(o.toString()))))
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    void getQxEntityAskOrders() {
        byte[] publicKey = identityUtil.getPublicKeyFromIdentity("ABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDVGRF");
        QxGetEntityOrders request = new QxGetEntityAskOrders(publicKey, 0);
        StepVerifier.create(nodes.getNode().flatMap(n -> n.getQxEntityOrders(request))
                        .doOnNext(orders -> orders.forEach(o -> log.info(o.toString()))))
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    void getQxEntityBidOrders() {
        byte[] publicKey = identityUtil.getPublicKeyFromIdentity("ABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDVGRF");
        QxGetEntityOrders request = new QxGetEntityBidOrders(publicKey, 0);
        StepVerifier.create(nodes.getNode().flatMap(n -> n.getQxEntityOrders(request))
                        .doOnNext(orders -> orders.forEach(o -> log.info(o.toString()))))
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    void qearnGetLockInfoPerEpoch() {
        GetLockInfoForEpoch request = GetLockInfoForEpoch.builder()
                .epoch(146)
                .build();

        StepVerifier.create(nodes.getNode().flatMap(n -> n.getQearnLockInfoForEpoch(request))
                .doOnNext(lockInfo -> log.info(lockInfo.toString())))
                .expectNextCount(1)
                .verifyComplete();

    }

    @Test
    void qearnGetUserLockedInfo() {
        GetUserLockInfo request = GetUserLockInfo.builder()
                .publicKey(identityUtil.getPublicKeyFromIdentity("ABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDVGRF"))
                .epoch(140)
                .build();

        StepVerifier.create(nodes.getNode().flatMap(n -> n.getQearnUserLockInfo(request))
                        .doOnNext(lockInfo -> log.info(lockInfo.toString())))
                .expectNextCount(1)
                .verifyComplete();

    }

    @Test
    public void getQearnStateOfRound() {
        GetStateOfRound getStateOfRound = GetStateOfRound.builder().build();
        StepVerifier.create(nodes.getNode().flatMap(n -> n.getQearnStateOfRound(getStateOfRound))
                        .doOnNext(state -> log.info(state.toString())))
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    public void getQearnUserLockStatus() {
        GetUserLockStatus request = GetUserLockStatus.builder()
                .publicKey(identityUtil.getPublicKeyFromIdentity("ABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDVGRF"))
                .build();
        StepVerifier.create(nodes.getNode().flatMap(n -> n.getQearnUserLockStatus(request))
                        .doOnNext(state -> log.info("{} => {} => {}",
                                state.getStatus(),
                                state.getStatusAsBinaryString(),
                                state.getEpochOffsetsWithLocks())))
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    public void getQearnEndedStatus() {
        GetEndedStatus request = GetEndedStatus.builder()
                .publicKey(identityUtil.getPublicKeyFromIdentity("ABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDVGRF"))
                .build();
        StepVerifier.create(nodes.getNode().flatMap(n -> n.getQearnEndedStatus(request))
                        .doOnNext(state -> log.info(state.toString())))
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    public void getQearnStatsForEpoch() {
        GetStatsForEpoch request = GetStatsForEpoch.builder()
                .epoch(145)
                .build();
        StepVerifier.create(nodes.getNode().flatMap(n -> n.getQearnStatsForEpoch(request))
                        .doOnNext(state -> log.info(state.toString())))
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    public void getQearnBurnedAndBoostedStats() {
        GetBurnedAndBoostedStats request = new GetBurnedAndBoostedStats();
        StepVerifier.create(nodes.getNode().flatMap(n -> n.getQearnBurnedAndBoostedStats(request))
                        .doOnNext(stats -> log.info("Stats: {} {} {} {}", stats,
                                stats.getAverageBoostedPercentAsDecimal(),
                                stats.getAverageBurnedPercentAdDecimal(),
                                stats.getAverageRewardedPercentAsDecimal())))
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    void getQearnBurnedAndBoostedStatsForEpoch() {
        GetBurnedAndBoostedStatsForEpoch request = GetBurnedAndBoostedStatsForEpoch.builder().build();
        StepVerifier.create(nodes.getNode().flatMap(n -> n.getQearnBurnedAndBoostedStatsForEpoch(request))
                        .doOnNext(stats -> log.info("Stats: {} {} {} {}", stats,
                                stats.getBoostedPercentAsDecimal(),
                                stats.getBurnedPercentAsDecimal(),
                                stats.getRewardedPercentAsDecimal())))
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    public void getIssuedAssetByUniverseIndex() {
        AssetIssuance expectedAssetIssuance = AssetIssuance.builder()
                .name(AssetUtil.nameTo7Bytes("RANDOM"))
                .publicKey(new byte[32])
                .type((byte) 1)
                .numberOfDecimalPlaces((byte) 0)
                .unitOfMeasurement(AssetUtil.unitOfMeasurementTo7Bytes("0000000"))
                .build();

        StepVerifier.create(nodes.getNode().flatMap(n -> n.getIssuedAssetByUniverseIndex(3)))
                .assertNext(ria -> {
                    assertThat(ria.getAsset()).isEqualTo(expectedAssetIssuance);
                    assertThat(ria.getUniverseIndex()).isEqualTo(3);
                    assertThat(ria.getTick()).isPositive();
                })
                .verifyComplete();
    }

    @Test
    public void getIssuedAssetsByFilter() {

        StepVerifier.create(nodes.getNode()
                        .flatMapMany(n -> n.getIssuedAssetsByFilter(
                                RequestAssetsByFilter.builder()
                                        .assetRequestType(RequestAssets.RequestType.REQUEST_TYPE_ISSUANCE_RECORDS)
                                        .issuer(new byte[32])
                                        .assetName("RANDOM")
                                        .build())
                        ))
                .expectNextCount(1).verifyComplete();

    }

    @Test
    public void getOwnedAssetsByFilter() {
        StepVerifier.create(nodes.getNode()
                        .flatMapMany(n -> n.getOwnedAssetsByFilter(
                                RequestAssetsByFilter.builder()
                                        .assetRequestType(RequestAssets.RequestType.REQUEST_TYPE_OWNERSHIP_RECORDS)
                                        .assetName("RANDOM")
                                        .issuer(new byte[32])
                                        .build())
                        )
                        .take(2))
                .expectNextCount(2).verifyComplete();
    }

    @Test
    public void getPossessedAssetsByFilter() {

        StepVerifier.create(nodes.getNode()
                        .flatMapMany(n -> n.getPossessedAssetsByFilter(
                                RequestAssetsByFilter.builder()
                                        .assetRequestType(RequestAssets.RequestType.REQUEST_TYPE_POSSESSION_RECORDS)
                                        .assetName("RANDOM")
                                        .issuer(new byte[32])
                                        .build())
                        )
                        .take(2))
                .expectNextCount(2).verifyComplete();

    }

}