package at.qubic.api.network;

import at.qubic.api.domain.MessageType;
import at.qubic.api.domain.qx.request.QxGetAssetAskOrders;
import at.qubic.api.domain.qx.request.QxGetAssetOrders;
import at.qubic.api.domain.qx.request.QxGetEntityAskOrders;
import at.qubic.api.domain.qx.request.QxGetEntityOrders;
import at.qubic.api.domain.qx.response.AssetOrder;
import at.qubic.api.domain.qx.response.EntityOrder;
import at.qubic.api.domain.qx.response.QxFees;
import at.qubic.api.domain.std.SignedTransaction;
import at.qubic.api.domain.std.Transaction;
import at.qubic.api.domain.std.request.RequestAssets;
import at.qubic.api.domain.std.request.RequestAssetsByFilter;
import at.qubic.api.domain.std.response.*;
import at.qubic.api.exception.InvalidResponseException;
import at.qubic.api.util.AssetUtil;
import at.qubic.api.util.TestMessageFactory;
import org.apache.commons.codec.DecoderException;
import org.apache.commons.codec.binary.Hex;
import org.junit.jupiter.api.Test;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;
import reactor.netty.ByteBufFlux;
import reactor.netty.Connection;
import reactor.netty.NettyInbound;
import reactor.netty.NettyOutbound;
import reactor.netty.tcp.TcpClient;
import reactor.netty.tcp.TcpClientConfig;
import reactor.test.StepVerifier;

import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.Arrays;
import java.util.function.Supplier;

import static org.assertj.core.api.Assertions.assertThat;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

class NodeTest {

    private final TcpClient tcpClient = mock();
    private final Node node = createNode();

    @Test
    void getCurrentTickInfo() {
        TickInfo expected = TickInfo.builder().build();

        final NettyOutbound outbound = mockOutbound();
        final NettyInbound inbound = mockInbound(MessageType.RESPOND_CURRENT_TICK_INFO, expected.toByteBuffer());
        mockConnection(inbound, outbound);

        StepVerifier.create(node.getCurrentTickInfo())
                .expectNext(expected)
                .verifyComplete();
    }

    @Test
    void getTickData() {
        int sizeBytes = 41072; // 16 + 32 + 1024*32 + 1024*8 + 64;
        mockConnection(mockInbound(MessageType.BROADCAST_FUTURE_TICK_DATA, ByteBuffer.allocate(sizeBytes)), mockOutbound());
        StepVerifier.create(node.getTickData(42))
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    void getBalance() {
        Entity expected = Entity.builder()
                .publicKey(new byte[32]) // required
                .spectrum(new byte[24][32]) // required
                .build();

        final NettyOutbound outbound = mockOutbound();
        final NettyInbound inbound = mockInbound(MessageType.RESPOND_ENTITY, expected.toByteBuffer());
        mockConnection(inbound, outbound);

        StepVerifier.create(node.getEntity(Mono.just(new byte[32])))
                .expectNext(expected)
                .verifyComplete();
    }

    @Test
    void getTickTransactions() {

        SignedTransaction tx1 = SignedTransaction.builder()
                .transaction(Transaction.builder()
                        .sourcePublicKey(new byte[32])
                        .destinationPublicKey(new byte[32])
                        .build())
                .signature(new byte[64])
                .build();

        final byte[] tickTransactionMessage = messageBytes(MessageType.BROADCAST_TRANSACTION, tx1.toBytes());
        final NettyInbound inbound = mockInbound(
                ByteBuffer.wrap(tickTransactionMessage),
                ByteBuffer.wrap(tickTransactionMessage),
                ByteBuffer.wrap(messageBytes(MessageType.END_RESPONSE, new byte[0])));
        final NettyOutbound outbound = mockOutbound();
        mockConnection(inbound, outbound);

        StepVerifier.create(node.getTickTransactions(42))
                .expectNextCount(2)
                .verifyComplete();
    }

    @Test
    void broadcastTransaction() {
        mockSendOnly();
        SignedTransaction stx = SignedTransaction.builder().build();

        StepVerifier.create(node.broadcastTransaction(Mono.just(stx)))
                .expectNext(stx)
                .verifyComplete();
    }

    @Test
    void getOwnedAssets() {
        int ownedAssetBytes = (48 * 2) + 8 + (24 * 32);
        final byte[] ownedAsset1 = messageBytes(MessageType.RESPOND_OWNED_ASSETS, new byte[ownedAssetBytes]);
        final byte[] ownedAsset2 = messageBytes(MessageType.RESPOND_OWNED_ASSETS, new byte[ownedAssetBytes]);
        final ByteBuffer endResponse = ByteBuffer.wrap(messageBytes(MessageType.END_RESPONSE, new byte[0]));
        mockConnection(mockInbound(
                ByteBuffer.wrap(ownedAsset1),
                ByteBuffer.wrap(ownedAsset2),
                endResponse
        ), mockOutbound());

        StepVerifier.create(node.getOwnedAssets(Mono.just(new byte[32])))
                .expectNextCount(2)
                .verifyComplete();
    }

    @Test
    void getPossessedAssets() {
        int ownedAssetBytes = (48 * 3) + 8 + (24 * 32);
        final byte[] possessedAsset1 = messageBytes(MessageType.RESPOND_POSSESSED_ASSETS, new byte[ownedAssetBytes]);
        final byte[] possessedAsset2 = messageBytes(MessageType.RESPOND_POSSESSED_ASSETS, new byte[ownedAssetBytes]);
        final ByteBuffer endResponse = ByteBuffer.wrap(messageBytes(MessageType.END_RESPONSE, new byte[0]));
        mockConnection(mockInbound(
                ByteBuffer.wrap(possessedAsset1),
                ByteBuffer.wrap(possessedAsset2),
                endResponse
        ), mockOutbound());

        StepVerifier.create(node.getPossessedAssets(Mono.just(new byte[32])))
                .expectNextCount(2)
                .verifyComplete();
    }

    @Test
    void getQxFees() {
        ByteBuffer qxFees = ByteBuffer.allocate(12)
                .order(ByteOrder.LITTLE_ENDIAN)
                .putInt(1)
                .putInt(2)
                .putInt(3);
        mockConnection(mockInbound(MessageType.RESPOND_CONTRACT_FUNCTION, qxFees), mockOutbound());

        StepVerifier.create(node.getQxFees())
                .assertNext(qxf -> assertThat(qxf).isEqualTo(QxFees.builder().assetIssuanceFee(1).transferFee(2).tradeFee(3).build()))
                .verifyComplete();
    }

    @Test
    void getQxFees_givenSmallPayload_thenThrow() {
        mockConnection(mockInbound(MessageType.RESPOND_CONTRACT_FUNCTION, ByteBuffer.wrap(new byte[] { 1, 2, 3 })), mockOutbound());

        StepVerifier.create(node.getQxFees())
                .verifyError(InvalidResponseException.class);
    }

    @Test
    void getQxAssetOrders() {

        byte[] publicKey = new byte[32];
        Arrays.fill(publicKey, (byte) 1);
        ByteBuffer assetOrder = ByteBuffer.allocate(AssetOrder.ORDER_COUNT * AssetOrder.ORDER_SIZE_BYTES)
                .order(ByteOrder.LITTLE_ENDIAN)
                .put(publicKey)
                .putLong(123)
                .putLong(42)
                .put(new byte[(AssetOrder.ORDER_COUNT-1) * AssetOrder.ORDER_SIZE_BYTES]); // rest empty

        mockConnection(mockInbound(MessageType.RESPOND_CONTRACT_FUNCTION, assetOrder), mockOutbound());

        QxGetAssetOrders request = new QxGetAssetAskOrders(new byte[32], "foo", 0);
        StepVerifier.create(node.getQxAssetOrders(request))
                .assertNext(lao -> {
                    assertThat(lao).hasSize(1); // only non-empty are returned
                    assertThat(lao.getFirst()).isEqualTo(AssetOrder.builder()
                            .entity(publicKey)
                            .price(123)
                            .numberOfShares(42)
                            .build());
                })
                .verifyComplete();

    }

    @Test
    void getQxEntityOrders() {

        byte[] publicKey = new byte[32];
        byte[] assetName = {'F', 'O', 'O', 0, 0, 0, 0, 0};
        ByteBuffer entityOrder = ByteBuffer.allocate(EntityOrder.ORDER_COUNT * EntityOrder.ORDER_SIZE_BYTES)
                .order(ByteOrder.LITTLE_ENDIAN)
                .put(publicKey)
                .put(assetName)
                .putLong(123)
                .putLong(42)
                .put(new byte[(AssetOrder.ORDER_COUNT-1) * AssetOrder.ORDER_SIZE_BYTES]); // rest empty

        mockConnection(mockInbound(MessageType.RESPOND_CONTRACT_FUNCTION, entityOrder), mockOutbound());

        byte[] entityPublicKey = new byte[32];
        Arrays.fill(entityPublicKey, (byte) 'A');
        QxGetEntityOrders request = new QxGetEntityAskOrders(entityPublicKey,0);
        StepVerifier.create(node.getQxEntityOrders(request))
                .assertNext(lao -> {
                    assertThat(lao).hasSize(1); // only non-empty are returned
                    assertThat(lao.getFirst()).isEqualTo(EntityOrder.builder()
                            .issuer(publicKey)
                            .assetName(assetName)
                            .price(123)
                            .numberOfShares(42)
                            .build());
                })
                .verifyComplete();
    }

    @Test
    void getIssuedAssetByUniverseIndex() throws DecoderException {

        byte[] respondAssetsPayload = Hex.decodeHex("00000000000000000000000000000000000000000000000000000000000000000152414e444f4d000000000000000000f3fa320103000000");
        mockConnection(mockInbound(MessageType.RESPOND_ASSETS, ByteBuffer.wrap(respondAssetsPayload)), mockOutbound());

        AssetIssuance expectedAssetIssuance = AssetIssuance.builder()
                .name(AssetUtil.nameTo7Bytes("RANDOM"))
                .publicKey(new byte[32])
                .type((byte) 1)
                .numberOfDecimalPlaces((byte) 0)
                .unitOfMeasurement(AssetUtil.unitOfMeasurementTo7Bytes("0000000"))
                .build();

        StepVerifier.create(node.getIssuedAssetByUniverseIndex(3))
                .assertNext(ria -> {
                    assertThat(ria.getAsset()).isEqualTo(expectedAssetIssuance);
                    assertThat(ria.getUniverseIndex()).isEqualTo(3);
                    assertThat(ria.getTick()).isEqualTo(20118259);
                })
                .verifyComplete();

    }

    @Test
    void getIssuedAssetsByFilter() throws DecoderException {
        byte[] respondAssetsPayload = Hex.decodeHex("00000000000000000000000000000000000000000000000000000000000000000152414e444f4d000000000000000000f3fa320103000000");
        mockConnection(mockInbound(
                ByteBuffer.wrap((messageBytes(MessageType.RESPOND_ASSETS, respondAssetsPayload))),
                ByteBuffer.wrap(messageBytes(MessageType.END_RESPONSE, new byte[0]))
        ) , mockOutbound());

        AssetIssuance expected = AssetIssuance.builder()
                .name(AssetUtil.nameTo7Bytes("RANDOM"))
                .publicKey(new byte[32])
                .type((byte) 1)
                .numberOfDecimalPlaces((byte) 0)
                .unitOfMeasurement(AssetUtil.unitOfMeasurementTo7Bytes("0000000"))
                .build();

        RequestAssetsByFilter request = RequestAssetsByFilter.builder()
                .assetRequestType(RequestAssets.RequestType.REQUEST_TYPE_ISSUANCE_RECORDS)
                .build();

        StepVerifier.create(node.getIssuedAssetsByFilter(request))
                .assertNext(ria -> {
                    assertThat(ria.getAsset()).isEqualTo(expected);
                    assertThat(ria.getUniverseIndex()).isEqualTo(3);
                    assertThat(ria.getTick()).isEqualTo(20118259);
                })
                .verifyComplete();
    }

    @Test
    void getOwnedAssetsByFilter() throws DecoderException {
        byte[] respondAssetsPayload = Hex.decodeHex("26843cfa419ce5ae9ed299757e57234eb50770a9bf711a103e8a068cf34ddffa02000100030000000100000000000000b3d3340126843c00");
        mockConnection(mockInbound(
                ByteBuffer.wrap(messageBytes(MessageType.RESPOND_ASSETS, respondAssetsPayload)),
                ByteBuffer.wrap(messageBytes(MessageType.END_RESPONSE, new byte[0]))
        ) , mockOutbound());

        AssetOwnership expected = AssetOwnership.builder()
                .publicKey(Hex.decodeHex("26843cfa419ce5ae9ed299757e57234eb50770a9bf711a103e8a068cf34ddffa"))
                .type((byte) 2)
                .numberOfUnits((byte) 1)
                .managingContractIndex((short) 1)
                .issuanceIndex(3)
                .build();

        RequestAssetsByFilter request = RequestAssetsByFilter.builder()
                .assetRequestType(RequestAssets.RequestType.REQUEST_TYPE_OWNERSHIP_RECORDS)
                .issuer(new byte[32])
                .assetName("RANDOM")
                .build();

        StepVerifier.create(node.getOwnedAssetsByFilter(request))
                .assertNext(ria -> {
                    assertThat(ria.getAsset()).isEqualTo(expected);
                    assertThat(ria.getUniverseIndex()).isEqualTo(3965990);
                    assertThat(ria.getTick()).isEqualTo(20239283);
                })
                .verifyComplete();
    }

        @Test
    void getPossessedAssetsByFilter() throws DecoderException {
            byte[] respondAssetsPayload = Hex.decodeHex("26843cfa419ce5ae9ed299757e57234eb50770a9bf711a103e8a068cf34ddffa0300010026843c00010000000000000001d5340127843c00");
            mockConnection(mockInbound(
                    ByteBuffer.wrap(messageBytes(MessageType.RESPOND_ASSETS, respondAssetsPayload)),
                    ByteBuffer.wrap(messageBytes(MessageType.END_RESPONSE, new byte[0]))
            ) , mockOutbound());

        AssetPossession expected = AssetPossession.builder()
                .publicKey(Hex.decodeHex("26843cfa419ce5ae9ed299757e57234eb50770a9bf711a103e8a068cf34ddffa"))
                .type((byte) 3)
                .numberOfUnits((byte) 1)
                .managingContractIndex((short) 1)
                .ownershipIndex(3965990)
                .build();

        RequestAssetsByFilter request = RequestAssetsByFilter.builder()
                .assetRequestType(RequestAssets.RequestType.REQUEST_TYPE_POSSESSION_RECORDS)
                .issuer(new byte[32])
                .assetName("RANDOM")
                .build();

        StepVerifier.create(node.getPossessedAssetsByFilter(request))
                .assertNext(ria -> {
                    assertThat(ria.getAsset()).isEqualTo(expected);
                    assertThat(ria.getUniverseIndex()).isEqualTo(3965991);
                    assertThat(ria.getTick()).isEqualTo(20239617);
                })
                .verifyComplete();
    }

    private Node createNode() {
        TcpClientConfig config = mock();
        when(tcpClient.configuration()).thenReturn(config);
        when(config.remoteAddress()).then(args -> (Supplier<SocketAddress>) () -> new InetSocketAddress("localhost", 666));
        return new Node(tcpClient);
    }

    private void mockSendOnly() {
        when(tcpClient.handle(any())).then(x -> tcpClient);
        Connection connection = mock();
        when(tcpClient.connect()).then(x -> Mono.just(connection));
    }

    private void mockConnection(NettyInbound inbound, NettyOutbound outbound) {
        Connection connection = mock();
        when(connection.inbound()).thenReturn(inbound);
        when(connection.outbound()).thenReturn(outbound);
        when(tcpClient.connect()).then(x -> Mono.just(connection));
    }

    private static NettyOutbound mockOutbound() {
        NettyOutbound outbound = mock();
        when(outbound.sendByteArray(any())).thenReturn(outbound);
        when(outbound.then()).thenReturn(Mono.empty());
        return outbound;
    }

    private static NettyInbound mockInbound(MessageType messageType, ByteBuffer payload) {
        NettyInbound inbound = mock();

        final byte[] message = messageBytes(messageType, payload.array());
        ByteBufFlux byteBufFlux = mock();
        when(byteBufFlux.asByteBuffer()).thenReturn(Flux.just(ByteBuffer.wrap(message)));
        when(inbound.receive()).thenReturn(byteBufFlux);
        return inbound;
    }

    private static NettyInbound mockInbound(ByteBuffer... payloads) {
        NettyInbound inbound = mock();
        ByteBufFlux byteBufFlux = mock();
        when(byteBufFlux.asByteBuffer()).thenReturn(Flux.just(payloads));
        when(inbound.receive()).thenReturn(byteBufFlux);
        return inbound;
    }

    private static byte[] messageBytes(MessageType messageType, byte[] payload) {
        return TestMessageFactory.createMessage(messageType, payload).toBytes();
    }

}