package at.qubic.api.network;

import org.junit.jupiter.api.Test;
import at.qubic.api.domain.std.response.TickInfo;
import reactor.core.publisher.Mono;
import reactor.test.StepVerifier;

import java.time.Duration;
import java.time.Instant;

import static org.assertj.core.api.Assertions.assertThat;
import static org.mockito.Mockito.*;

class NodesManagementAdapterTest {

    private final NetworkStatus networkStatus = mock();
    private final NodesManagementAdapter adapter = new NodesManagementAdapter(networkStatus) {};

    @Test
    void getNodes_thenReturnCurrentNode() {
        TickInfo currentTick = TickInfo.builder()
                .tick(42)
                .epoch((short) 666)
                .build();

        when(networkStatus.getTickStatus()).thenReturn(Mono.just(TickStatus.builder()
                .tickInfo(currentTick)
                .latestUpdate(Instant.now())
                .build()));

        Node node1 = mockNode("1", currentTick);
        adapter.addNode(node1);

        StepVerifier.create(adapter.getNodes(1))
                .expectNext(node1)
                .verifyComplete();

        verify(networkStatus, never()).update(anyCollection());

    }

    @Test
    void getNodes_givenOutdated_thenReturnEmpty() {
        TickInfo currentTick = TickInfo.builder()
                .tick(42)
                .epoch((short) 666)
                .build();

        when(networkStatus.getTickStatus()).thenReturn(Mono.just(TickStatus.builder()
                .tickInfo(currentTick)
                .latestUpdate(Instant.now())
                .build()));

        when(networkStatus.update(anyCollection())).thenReturn(Mono.just(currentTick)); // tries update

        // outdated nodes
        adapter.addNode(mockNode("1", TickInfo.builder().epoch((short) 1).tick(42).build()));
        adapter.addNode(mockNode("2", TickInfo.builder().epoch((short) 666).tick(33).build()));
        adapter.addNode(mockNode("3", TickInfo.builder().build()));

        StepVerifier.create(adapter.getNodes(1))
                .verifyComplete();

    }

    @Test
    void getNodes_givenFewNodes_thenUpdateNetworkStatus() {
        TickInfo currentTick = TickInfo.builder()
                .tick(42)
                .epoch((short) 666)
                .build();

        when(networkStatus.getTickStatus()).thenReturn(Mono.just(TickStatus.builder()
                .tickInfo(currentTick)
                .latestUpdate(Instant.now())
                .build()));

        when(networkStatus.update(anyCollection())).thenReturn(Mono.just(currentTick));

        Node node1 = mockNode("1", currentTick);
        adapter.addNode(node1);

        StepVerifier.create(adapter.getNodes(100))
                .expectNext(node1)
                .verifyComplete();

        verify(networkStatus).update(anyCollection());

    }

    @Test
    void getNodes_givenOutOfDateStatus_thenUpdateNetworkStatus() {
        TickInfo currentTick = TickInfo.builder()
                .tick(42)
                .epoch((short) 666)
                .build();

        when(networkStatus.getTickStatus()).thenReturn(Mono.just(TickStatus.builder()
                .tickInfo(currentTick)
                .latestUpdate(Instant.now().minus(Duration.ofHours(1)))
                .build()));

        when(networkStatus.update(anyCollection())).thenReturn(Mono.just(currentTick)); // async

        Node node1 = mockNode("1", currentTick);
        adapter.addNode(node1);

        StepVerifier.create(adapter.getNodes(1))
                .expectNext(node1)
                .verifyComplete();

        verify(networkStatus).update(anyCollection());

    }


    @Test
    void addNode_thenCreateNewNode() {
        assertThat(adapter.nodes).isEmpty();

        adapter.addNode("localhost");
        assertThat(adapter.nodes.values().stream().map(Node::getName)).containsExactly("localhost");

        adapter.addNode("another");
        assertThat(adapter.nodes.values().stream().map(Node::getName)).containsExactlyInAnyOrder("localhost", "another");
    }

    @Test
    void addNode_givenSameNode_thenDoNotCreateNewNode() {
        adapter.addNode("localhost");
        adapter.addNode("localhost");
        assertThat(adapter.nodes).hasSize(1);
    }

    private static Node mockNode(String name, TickInfo tickInfo) {
        Node node = mock();
        when(node.getName()).thenReturn(name);
        when(node.getCurrentTickInfo()).thenReturn(Mono.just(tickInfo));
        when(node.getTickInfo()).thenReturn(tickInfo);
        return node;
    }

}