package at.qubic.api.service;

import lombok.extern.slf4j.Slf4j;
import org.apache.commons.codec.binary.Hex;
import org.awaitility.Awaitility;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.Test;
import at.qubic.api.SpringBootTestConfiguration;
import at.qubic.api.domain.MessageType;
import at.qubic.api.domain.QubicHeader;
import at.qubic.api.domain.QubicMessage;
import at.qubic.api.domain.std.SignedTransaction;
import at.qubic.api.domain.std.Transaction;
import at.qubic.api.domain.std.response.Entity;
import at.qubic.api.domain.std.response.ExchangePublicPeers;
import at.qubic.api.domain.std.response.TickInfo;
import at.qubic.api.tcp.TcpTestServer;
import at.qubic.api.util.TestFileUtil;
import at.qubic.api.util.TestMessageFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import reactor.test.StepVerifier;

import java.io.IOException;
import java.time.Duration;
import java.util.function.Function;

import static org.assertj.core.api.Assertions.assertThat;
import static at.qubic.api.domain.MessageType.BROADCAST_FUTURE_TICK_DATA;
import static at.qubic.api.util.QubicHeaderUtil.headerFromBytes;

@SuppressWarnings("SpringBootApplicationProperties")
@SpringBootTest(classes = SpringBootTestConfiguration.class, properties = { // for mock tcp server
        "computor.hosts=localhost",
        "network.port=6666",
        "use.public.peers=false",
        "check.health=false"
})
@Slf4j
class ComputorServiceIT {

    @Autowired
    private ComputorService computorService;

    private final TcpTestServer tcpServer = new TcpTestServer(6666).start();

    @AfterEach
    void tearDown() throws IOException {
        tcpServer.stop();
    }

    @Test
    void getEntity_ThenRetrieveEntity() throws Exception {

        final Entity entity = Entity.builder()
                .publicKey(Hex.decodeHex("30df4cb7d17334b6a84488d631a1f73a9bcd5420f00209cb91c90c251e4d77b4"))
                .incomingAmount(100_000_000_000L)
                .numberOfIncomingTransfers(1)
                .latestIncomingTransferTick(123456)
                .spectrum(new byte[24][32])
                .build();

        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        final QubicMessage respondEntity = TestMessageFactory.createMessage(MessageType.RESPOND_ENTITY, entity.toByteBuffer().array());

        tcpServer.respond(exchangePublicPeers.toBytes(), respondEntity.toBytes());
        StepVerifier.create(computorService.getEntity("MMPYLHEBCADPHFOXPTTSBRWJRNSBHEVFXCXEHRBIXFDTQCLJRFRKZGGFFQUM"))
                .expectNext(entity)
                .verifyComplete();
    }

    @Test
    void getCurrentTickInfo_thenRetrieveTickInfo() {

        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        final QubicMessage respondCurrentTickInfo = getCurrentTickInfo(3);

        tcpServer.respond(exchangePublicPeers.toBytes(), respondCurrentTickInfo.toBytes());

        StepVerifier.create(computorService.getCurrentTickInfo().log())
                .assertNext(ti -> {
                    assertThat(ti.getTickDuration()).isEqualTo((short) 1);
                    assertThat(ti.getEpoch()).isEqualTo((short) 2);
                    assertThat(ti.getTick()).isEqualTo(3);
                    assertThat(ti.getNumberOfAlignedVotes()).isEqualTo((short) 4);
                    assertThat(ti.getNumberOfMisalignedVotes()).isEqualTo((short) 5);
                    assertThat(ti.getInitialTick()).isEqualTo(6);
                })
                .verifyComplete();
     }

    @Test
    void getTickData() {
        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        final QubicMessage tickData = getBroadcastFutureTickData();
        tcpServer.respond(exchangePublicPeers.toBytes(), tickData.toBytes());

        StepVerifier.create(computorService.getTickData(42))
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    void getTickTransactions() {

        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        final QubicMessage endResponse = TestMessageFactory.createMessage(MessageType.END_RESPONSE);
        byte[] tx1 = TestFileUtil.loadBinaryFile("/testdata/domain/broadcast-transactions-daomgxxs.bin");
        byte[] tx2 = TestFileUtil.loadBinaryFile("/testdata/domain/broadcast-transactions-cndabrkx.bin");
        byte[] tx3 = TestFileUtil.loadBinaryFile("/testdata/domain/broadcast-transactions-lgdvbraz.bin");

        // make sure node is up to date first
        mockGetCurrentTickInfo();

        tcpServer.respond(exchangePublicPeers.toBytes(), tx1, tx2, tx3, endResponse.toBytes());

        StepVerifier.create(computorService.getTickTransactions(42))
                .assertNext(tx -> assertThat(tx.getTransactionHash()).isEqualTo("daomgxxsycpnxebpbqpzeurndmsacbejuuffwjfoahzqgwueqgzygdpdnzqf"))
                .assertNext(tx -> assertThat(tx.getTransactionHash()).isEqualTo("cndabrkxpusrvgdfgkohwccnhnwengbsbrskpsmiadqrowenerkunxkbdkhi"))
                .assertNext(tx -> assertThat(tx.getTransactionHash()).isEqualTo("lgdvbrazdowjidhbnqwyfxhupdfbcibcsxvlkagyybbgnfxqyclrsfubenih"))
                .verifyComplete();
    }

    @Test
    void getTickTransactions_givenNodeNotUpToDate_thenError() {
        final QubicMessage respondTickInfo = getCurrentTickInfo(42);
        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        tcpServer.respond(exchangePublicPeers.toBytes(), respondTickInfo.toBytes());

        // node is at tick 42. required tick: 43
        StepVerifier.create(computorService.getTickTransactions(43))
                .expectError(IllegalStateException.class)
                .verify();
    }


    @Test
    void getTickTransactions_givenNoTransactions_thenNoResult() {
        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        final QubicMessage endResponse = TestMessageFactory.createMessage(MessageType.END_RESPONSE);

        // make sure node is up to date first
        mockGetCurrentTickInfo();

        tcpServer.respond(exchangePublicPeers.toBytes(), endResponse.toBytes());

        StepVerifier.create(computorService.getTickTransactions(42))
                .expectNextCount(0)
                .verifyComplete();
    }

    @Test
    void sendTransferTransaction() {
        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        final QubicMessage currentTickInfo = TestMessageFactory.createMessage(MessageType.RESPOND_CURRENT_TICK_INFO, TickInfo.builder().tick(42).build().toByteBuffer().array());
        tcpServer.respond(exchangePublicPeers.toBytes(), currentTickInfo.toBytes());

        StepVerifier.create(computorService
                        .sendQuTransferTransaction("abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabc",
                                "TESTIOYEZSAUQBUNKJEGXOILIBICZYWAMVSWYCYFHDCSZOAXPTGSOKFCCUNK",
                                1, 5))
                .assertNext(tx -> assertThat(tx.getTransactionHash()).isNotEmpty())
                .verifyComplete();
    }

    @Test
    void sendSignedTransaction() {
        TcpTestServer.ExchangeThread exchange = tcpServer.respond(getExchangePublicPeers().toBytes()); // dummy

        SignedTransaction stx = SignedTransaction.builder()
                .transaction(Transaction.builder().build())
                .build();
        StepVerifier.create(computorService.sendTransaction(stx))
                .expectNext(stx)
                .verifyComplete();

        Awaitility.await().pollInterval(Duration.ofMillis(10)).until(exchange::isReceivedPayload);
        assertThat(headerFromBytes(exchange.getHeader()).getType()).isEqualTo(MessageType.BROADCAST_TRANSACTION);
    }

    @Test
    void sendTransaction_givenFunction_thenSendFunctionResult() {
        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        TickInfo tickInfo = TickInfo.builder().tick(42).build();
        final QubicMessage currentTickInfo = TestMessageFactory.createMessage(MessageType.RESPOND_CURRENT_TICK_INFO, tickInfo.toByteBuffer().array());
        tcpServer.respond(exchangePublicPeers.toBytes(), currentTickInfo.toBytes());

        Function<Integer, SignedTransaction> signedTransactionFunction = tick -> SignedTransaction.builder()
                .transaction(Transaction.builder().tick(tick + 10).build())
                .build();
        StepVerifier.create(computorService.sendTransaction(signedTransactionFunction))
                .assertNext(stx -> assertThat(stx.getTransaction().getTick()).isEqualTo(52))
                .verifyComplete();
    }

    private void mockGetCurrentTickInfo() {
        final QubicMessage respondTickInfo = getCurrentTickInfo(42);
        final QubicMessage exchangePublicPeers = getExchangePublicPeers();
        tcpServer.respond(exchangePublicPeers.toBytes(), respondTickInfo.toBytes());
        computorService.getCurrentTickInfo().block();
    }

    private QubicMessage getCurrentTickInfo(int tick) {
        TickInfo tickInfo = TickInfo.builder()
                .tickDuration((short) 1)
                .epoch((short) 2)
                .tick(tick)
                .numberOfAlignedVotes((short) 4)
                .numberOfMisalignedVotes((short) 5)
                .initialTick(6)
                .build();
        return TestMessageFactory.createMessage(MessageType.RESPOND_CURRENT_TICK_INFO, tickInfo.toByteBuffer().array());
    }

    private QubicMessage getBroadcastFutureTickData() {
        final byte[] payload = TestFileUtil.loadHexFile("/testdata/domain/broadcast-future-tick-data-payload.hex");
        return QubicMessage.builder()
                .header(QubicHeader.builder().size(41080).type(BROADCAST_FUTURE_TICK_DATA).build())
                .payload(payload)
                .build();
    }

    private QubicMessage getExchangePublicPeers() {
        return TestMessageFactory.createMessage(
                MessageType.EXCHANGE_PUBLIC_PEERS,
                ExchangePublicPeers.builder()
                        .peers(new String[] { "1.2.3.4", "2.3.4.1", "3.4.1.2", "4.1.2.3" })
                        .build()
                        .toBytes()
                        .array());
    }

}