package at.qubic.api.service;

import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.Tag;
import org.junit.jupiter.api.Test;
import at.qubic.api.SpringBootTestConfiguration;
import at.qubic.api.crypto.IdentityUtil;
import at.qubic.api.domain.std.SignedTransaction;
import at.qubic.api.domain.std.Transaction;
import at.qubic.api.domain.std.response.Entity;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;
import reactor.util.retry.Retry;

import java.time.Duration;
import java.util.NoSuchElementException;

@Slf4j
@Tag("OneTimeTest")
@SpringBootTest(classes = SpringBootTestConfiguration.class)
public class ComputorServiceManualSIT {

    @Autowired
    private ComputorService computorService;

    @Autowired
    private IdentityUtil idUtil;

    @Test
    void getTickTransactions() {
        final Flux<SignedTransaction> result = computorService.getCurrentTickInfo()
                .flatMapMany(ti -> computorService.getTickTransactions(ti.getTick() - 1));

        result.filter(tx -> tx.getTransaction().getAmount() > 0)
                .doOnNext(this::logTransaction)
                .blockLast();

    }

    @Test
    void sendTransactionAndWait() {
        // initialize public peers (hacky)
        computorService.getCurrentTickInfo().block();
        Mono.delay(Duration.ofSeconds(5)).block();
        // now transact
        computorService.getCurrentTickInfo()
                .then(sendTransaction("nfbrqtymxoexbvbioyaiyalffrwsowtyegnlrohrqvterrpvtfiuwqo", "ITXUCKTOAXGLKFPOQMBDZJDSHUKCKZNNCACNONPWZESXDCILUKUMLJRAPPPD", 2))
                .zipWith(sendTransaction("hcqofilpsqqnnpdfoanwqcvsxhvvchkdpqlewodlihrkdjcsiaywirk", "GNRDTDTIEAAATAUPJZEQEEZHXPFAOQRAWIRJWGLFDANVSMQAWURDTHHGXGHC", 1))
                .block();
                //.repeat(2)
                //.blockLast();
    }

    private Mono<?> sendTransaction(String seed, String destination, long amount) {
        return getEntity(destination, idUtil.getIdentityFromPublicKey(idUtil.getPublicKeyFromSeed(seed)))
                .then(computorService.sendQuTransferTransaction(seed, destination, amount, 4)
                        .flatMap(this::checkIfTxIsIncluded)
                        .doOnNext(this::logTransaction)
                        .retryWhen(Retry.backoff(3, Duration.ofSeconds(1)))
                        .doOnError(t -> log.error("Unexpected error.", t))
                ).then(getEntity(destination, idUtil.getIdentityFromPublicKey(idUtil.getPublicKeyFromSeed(seed))));
    }

    private Mono<Entity> getEntity(String destination, String source) {
        return computorService
                .getEntity(destination)
                .doOnNext(b -> log.info("Destination: balance [{}], txs in: [{}@{}].", b.getIncomingAmount() - b.getOutgoingAmount(), b.getNumberOfIncomingTransfers(), b.getLatestIncomingTransferTick()))
                .then(computorService.getEntity(source)
                        .doOnNext(b -> log.info("Source balance [{}], txs out: [{}@{}]", b.getIncomingAmount() - b.getOutgoingAmount(), b.getNumberOfOutgoingTransfers(), b.getLatestOutgoingTransferTick())));
    }

    private Mono<SignedTransaction> checkIfTxIsIncluded(SignedTransaction tx) {
        return Flux.interval(Duration.ofSeconds(30), Duration.ofSeconds(5))
                .flatMap(count -> computorService.getCurrentTickInfo().retry(3))
                .doOnNext(ti -> log.info("Current tick: [{}]", ti.getTick()))
                .skipUntil(ti -> ti.getTick() > tx.getTransaction().getTick())
                .next()
                .flatMapMany(ti -> computorService.getTickTransactions(tx.getTransaction().getTick()).retry(3))
                .filter(transaction -> StringUtils.equals(tx.getTransactionHash(), transaction.getTransactionHash()))
                .single()
                .doOnError(NoSuchElementException.class, t -> log.warn("Transaction [{}] not included in tick [{}].", tx.getTransactionHash(), tx.getTransaction().getTick()))
                ;
    }

    private void logTransaction(SignedTransaction stx) {
        if (stx == null) {
            log.warn("Transaction status unclear.");
        } else {
            Transaction tx = stx.getTransaction();
            log.info("Confirmed. [{}] qu from [{}] to [{}] / hash: [{}].",
                    idUtil.getIdentityFromPublicKey(tx.getSourcePublicKey()),
                    tx.getAmount(),
                    idUtil.getIdentityFromPublicKey(tx.getDestinationPublicKey()),
                    stx.getTransactionHash()
            );
        }
    }

}
