package at.qubic.api.service;

import lombok.extern.slf4j.Slf4j;
import org.junit.jupiter.api.Tag;
import org.junit.jupiter.api.Test;
import at.qubic.api.SpringBootTestConfiguration;
import at.qubic.api.crypto.IdentityUtil;
import at.qubic.api.domain.std.SignedTransaction;
import at.qubic.api.domain.qx.request.*;
import at.qubic.api.domain.qx.response.AssetOrder;
import at.qubic.api.domain.qx.response.EntityOrder;
import at.qubic.api.domain.qx.response.QxFees;
import at.qubic.api.domain.std.response.RespondOwnedAssets;
import at.qubic.api.domain.std.response.RespondPossessedAssets;
import at.qubic.api.domain.std.response.TickData;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;
import reactor.test.StepVerifier;

import java.util.List;
import java.util.function.Function;

import static org.assertj.core.api.Assertions.assertThat;

@SuppressWarnings("SpringBootApplicationProperties")
@Slf4j
@Tag("SIT")
@SpringBootTest(classes = SpringBootTestConfiguration.class, properties = { "use.public.peers=false", "check.health=true" })
class ComputorServiceSIT {

    @Autowired
    private QxTransactionService qxTransactionService;

    @Autowired
    private ComputorService computorService;

    @Autowired
    private IdentityUtil idUtil;

    @Test
    void getEntity_ThenRetrieveEntity() {
        StepVerifier.create(computorService.getEntity("CFBMEMZOIDEXQAUXYYSZIURADQLAPWPMNJXQSNVQZAHYVOPYUKKJBJUCTVJL"))
                .assertNext(entity -> {
                    assertThat(idUtil.getIdentityFromPublicKey(entity.getPublicKey()))
                            .isEqualTo("CFBMEMZOIDEXQAUXYYSZIURADQLAPWPMNJXQSNVQZAHYVOPYUKKJBJUCTVJL");
                    assertThat(entity.getIncomingAmount()).isPositive();
                    assertThat(entity.getOutgoingAmount()).isPositive();
                    assertThat(entity.getNumberOfIncomingTransfers()).isPositive();
                    assertThat(entity.getNumberOfOutgoingTransfers()).isPositive();
                    assertThat(entity.getLatestIncomingTransferTick()).isPositive();
                    assertThat(entity.getLatestOutgoingTransferTick()).isPositive();
                    assertThat(entity.getTick()).isPositive();
                    assertThat(entity.getSpectrumIndex()).isPositive();
                    assertThat(entity.getSpectrum()).isNotEmpty();
                    assertThat(entity.getSpectrum()).hasDimensions(24, 32);
                })
                .verifyComplete();
    }

    @Test
    void getEntity_GivenEmpty_ThenRetrieveEntity() {
            StepVerifier.create(computorService.getEntity("TESTJFDGICVGFHLRAZCKJKQYVEUBVYWTMRYYTPFXRESOQJGTKHCIHDJGFQZJ"))
                    .assertNext(entity -> {
                        assertThat(entity.getPublicKey()).isEqualTo(idUtil.getPublicKeyFromIdentity("TESTJFDGICVGFHLRAZCKJKQYVEUBVYWTMRYYTPFXRESOQJGTKHCIHDJGFQZJ"));
                        assertThat(entity.getIncomingAmount()).isZero();
                        assertThat(entity.getOutgoingAmount()).isZero();
                        assertThat(entity.getNumberOfIncomingTransfers()).isZero();
                        assertThat(entity.getNumberOfOutgoingTransfers()).isZero();
                        assertThat(entity.getLatestIncomingTransferTick()).isZero();
                        assertThat(entity.getTick()).isPositive();
                        assertThat(entity.getSpectrumIndex()).isEqualTo(-1);
                        assertThat(entity.getSpectrum()).hasDimensions(24, 32);
                    })
                    .verifyComplete();
    }

    @Test
    void sendTransaction_ThenReturnTransactionHash() {
        StepVerifier.create(computorService
                        .sendQuTransferTransaction("abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabc",
                                "TESTIOYEZSAUQBUNKJEGXOILIBICZYWAMVSWYCYFHDCSZOAXPTGSOKFCCUNK",
                                1, 5))
                .assertNext(tx -> assertThat(tx.getTransactionHash()).isNotEmpty())
                .verifyComplete();
    }

    @Test
    void getCurrentTickInfo_thenRetrieveTickInfo() {
        StepVerifier.create(computorService.getCurrentTickInfo().log())
                .assertNext(ti -> {
                    assertThat(ti.getTickDuration()).isNotNegative();
                    assertThat(ti.getTick()).isPositive();
                    assertThat(ti.getEpoch()).isPositive();
                    assertThat(ti.getNumberOfAlignedVotes()).isNotNegative();
                    assertThat(ti.getNumberOfMisalignedVotes()).isNotNegative();
                })
                .verifyComplete();
    }

    @Test
    void getLatestTickData() {

        final Mono<TickData> getTickData = computorService.getCurrentTickInfo()
                .doOnNext(td -> log.info(td.toString()))
                .flatMap(ti -> computorService.getTickData(ti.getTick() - 1));

        // node might return no 'next' element. for example if tick is empty.
        StepVerifier.create(getTickData)
                .assertNext(td -> {
                    assertThat(td.getComputorIndex()).isPositive();
                    assertThat(td.getEpoch()).isPositive();
                    assertThat(td.getTick()).isPositive();
                    assertThat(td.getYear()).isPositive();
                    assertThat(td.getMonth()).isPositive();
                    assertThat(td.getDay()).isPositive();
                    assertThat(td.getHour()).isNotNegative();
                    assertThat(td.getMinute()).isNotNegative();
                    assertThat(td.getSecond()).isNotNegative();
                    assertThat(td.getMillisecond()).isNotNegative();
                    assertThat(td.getTimelock()).isNotEmpty();
                    assertThat(td.getTransactionDigests()).isNotEmpty();
                    assertThat(td.getContractFees()).isNotEmpty();
                    assertThat(td.getSignature()).isNotEmpty();
                })
                .verifyComplete();
    }

    @Test
    void getTickData_empty() {
        StepVerifier.create(computorService.getCurrentTickInfo()
                        .flatMap(ti -> computorService.getTickData(ti.getInitialTick() - 1)))
                //.expectNextCount(1)
                .verifyComplete();
    }

    @Test
    void getTickTransactions() {
        final Flux<SignedTransaction> result = computorService.getCurrentTickInfo()
                .flatMapMany(ti -> computorService.getTickTransactions(ti.getTick() - 1));

        StepVerifier.create(result.collectList()) // collect to list because we don't care if empty or not
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    void getOwnedAssets() {
        final Flux<RespondOwnedAssets> assets = computorService.getOwnedAssets("MMPYLHEBCADPHFOXPTTSBRWJRNSBHEVFXCXEHRBIXFDTQCLJRFRKZGGFFQUM");
        StepVerifier.create(assets)
                // .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    void getPossessedAssets() {
        final Flux<RespondPossessedAssets> assets = computorService.getPossessedAssets("MMPYLHEBCADPHFOXPTTSBRWJRNSBHEVFXCXEHRBIXFDTQCLJRFRKZGGFFQUM");
        StepVerifier.create(assets)
                // .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    void getQxFees() {
        final Mono<QxFees> fees = computorService.getQxFees();
        StepVerifier.create(fees)
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    void getQxAssetAskOrders() {
        String issuer = "CFBMEMZOIDEXQAUXYYSZIURADQLAPWPMNJXQSNVQZAHYVOPYUKKJBJUCTVJL";
        Mono<List<AssetOrder>> orders = computorService.getQxAskAssetOrders(issuer, "CFB", 0);
        StepVerifier.create(orders)
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    void getQxAssetBidOrders() {
        String issuer = "CFBMEMZOIDEXQAUXYYSZIURADQLAPWPMNJXQSNVQZAHYVOPYUKKJBJUCTVJL";
        Mono<List<AssetOrder>> orders = computorService.getQxBidAssetOrders(issuer, "CFB", 0);
        StepVerifier.create(orders)
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    void getQxEntityAskOrders() {
        String entity = "EOKXREPZIQRTTDHZUTVDGBUBIWVATNGUGCTOUWAJWBAVQWQZESDHQSTAVGQN";
        Mono<List<EntityOrder>> orders = computorService.getQxAskEntityOrders(entity, 0);
        StepVerifier.create(orders)
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    void getQxEntityBidOrders() {
        String entity = "EOKXREPZIQRTTDHZUTVDGBUBIWVATNGUGCTOUWAJWBAVQWQZESDHQSTAVGQN";
        Mono<List<EntityOrder>> orders = computorService.getQxBidEntityOrders(entity, 0);
        StepVerifier.create(orders)
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    void addQxBidOrder() {
        QxAddBidOrder order = QxAddBidOrder.builder()
                .orderData(QxAssetOrderData.builder()
                        .assetIssuer(new byte[32])
                        .price(1000)
                        .numberOfShares(1)
                        .assetName("RANDOM")
                        .build())
                .build();
        Function<Integer, SignedTransaction> createTransaction = qxTransactionService.createTransactionFunction(
                "hcqofilpsqqnnpdfoanwqcvsxhvvchkdpqlewodlihrkdjcsiaywirk", order, 5);
        StepVerifier.create(computorService.sendTransaction(createTransaction))
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    void removeQxBidOrder() {
        QxRemoveBidOrder order = QxRemoveBidOrder.builder()
                .orderData(QxAssetOrderData.builder()
                        .assetIssuer(new byte[32])
                        .price(350)
                        .numberOfShares(1)
                        .assetName("RANDOM")
                        .build())
                .build();
        Function<Integer, SignedTransaction> createTransaction = qxTransactionService.createTransactionFunction(
                "hcqofilpsqqnnpdfoanwqcvsxhvvchkdpqlewodlihrkdjcsiaywirk", order, 5);
        StepVerifier.create(computorService.sendTransaction(createTransaction))
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    void addQxAskOrder() {
        QxAddAskOrder order = QxAddAskOrder.builder()
                .orderData(QxAssetOrderData.builder()
                        .assetIssuer(new byte[32])
                        .price(825)
                        .numberOfShares(3)
                        .assetName("RANDOM")
                        .build())
                .build();
        Function<Integer, SignedTransaction> createTransaction = qxTransactionService.createTransactionFunction(
                "hcqofilpsqqnnpdfoanwqcvsxhvvchkdpqlewodlihrkdjcsiaywirk", order, 5);
        StepVerifier.create(computorService.sendTransaction(createTransaction))
                .expectNextCount(1)
                .verifyComplete();
    }

    @Test
    void removeQxAskOrder() {
        QxRemoveAskOrder order = QxRemoveAskOrder.builder()
                .orderData(QxAssetOrderData.builder()
                        .assetIssuer(new byte[32])
                        .price(900)
                        .numberOfShares(1)
                        .assetName("RANDOM")
                        .build())
                .build();
        Function<Integer, SignedTransaction> createTransaction = qxTransactionService.createTransactionFunction(
                "hcqofilpsqqnnpdfoanwqcvsxhvvchkdpqlewodlihrkdjcsiaywirk", order, 5);
        StepVerifier.create(computorService.sendTransaction(createTransaction))
                .expectNextCount(1)
                .verifyComplete();
    }

}