package at.qubic.api.service;

import lombok.extern.slf4j.Slf4j;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import at.qubic.api.crypto.IdentityUtil;
import at.qubic.api.domain.std.SignedTransaction;
import at.qubic.api.domain.std.Transaction;
import at.qubic.api.domain.qx.response.AssetOrder;
import at.qubic.api.domain.qx.response.EntityOrder;
import at.qubic.api.domain.qx.response.QxFees;
import at.qubic.api.domain.std.response.Entity;
import at.qubic.api.domain.std.response.TickInfo;
import at.qubic.api.network.Node;
import at.qubic.api.network.Nodes;
import at.qubic.api.util.AssetUtil;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;
import reactor.test.StepVerifier;

import java.util.List;
import java.util.function.Function;

import static org.mockito.Mockito.*;

@Slf4j
class ComputorServiceTest {

    private final Nodes nodes = mock();
    private final IdentityUtil identityUtil = mock();
    private final TransactionService transactionService = mock();
    private final ComputorService computorService = new ComputorService(nodes, identityUtil, transactionService);
    private final Node node = mock();

    @BeforeEach
    void setUp() {
        when(nodes.getNode()).thenReturn(Mono.just(node));
    }

    @Test
    void getCurrentTickInfo() {
        TickInfo ti = TickInfo.builder().build();
        when(node.getCurrentTickInfo()).thenReturn(Mono.just(ti));

        StepVerifier.create(computorService.getCurrentTickInfo())
                .expectNext(ti)
                .verifyComplete();
    }

    @Test
    void getEntity() {
        Entity entity = Entity.builder().build();
        when(node.getEntity(any())).thenReturn(Mono.just(entity));
        when(identityUtil.getPublicKeyFromIdentity("foo")).thenReturn(new byte[0]);

        StepVerifier.create(computorService.getEntity("foo").log())
                .expectNext(entity)
                .verifyComplete();
    }

    @Test
    void sendTransaction_givenSignedTransaction() {
        Transaction tx = mock();
        SignedTransaction stx = mock();
        when(stx.getTransaction()).thenReturn(tx);
        when(node.broadcastTransaction(any())).thenReturn(Mono.just(stx));

        StepVerifier.create(computorService.sendTransaction(stx))
                .expectNext(stx)
                .verifyComplete();
    }

    @Test
    void sendTransaction_givenValuesToCreateAndSignTransaction() {
        Transaction tx = Transaction.builder().build();
        SignedTransaction stx = SignedTransaction.builder().transaction(tx).build();

        when(node.getCurrentTickInfo()).thenReturn(Mono.just(TickInfo.builder().tick(42).build()));
        when(transactionService.createTransaction(45, "seed", "destination", 42)).thenReturn(stx);
        when(node.broadcastTransaction(any())).then(invocation -> invocation.getArgument(0));

        StepVerifier.create(computorService.sendQuTransferTransaction("seed", "destination", 42, 3).log())
                .expectNext(stx)
                .verifyComplete();
    }

    @Test
    void sendTransaction_givenTickCreationFunction() {
        Transaction tx = Transaction.builder().build();
        SignedTransaction stx = SignedTransaction.builder().transaction(tx).build();

        when(node.getCurrentTickInfo()).thenReturn(Mono.just(TickInfo.builder().build()));
        Function<Integer, SignedTransaction> signTransaction = tick -> stx;
        when(node.broadcastTransaction(any())).then(invocation -> invocation.getArgument(0));

        StepVerifier.create(computorService.sendTransaction(signTransaction).log())
                .expectNext(stx)
                .verifyComplete();
    }

    @Test
    void getTickTransactions() {
        Transaction tx = Transaction.builder().build();
        SignedTransaction stx = SignedTransaction.builder().transaction(tx).build();
        when(node.getTickInfo()).thenReturn(TickInfo.builder().tick(42).build());
        when(node.getTickTransactions(42)).thenReturn(Flux.just(stx));

        StepVerifier.create(computorService.getTickTransactions(42).log())
                .expectNext(stx)
                .verifyComplete();
    }

    @Test
    void getTickTransactions_givenNodeNotAtRequiredTick_thenCallGetCurrentTickInfo() {
        Transaction tx = Transaction.builder().build();
        SignedTransaction stx = SignedTransaction.builder().transaction(tx).build();
        when(node.getTickInfo()).thenReturn(TickInfo.builder().tick(1).build());
        when(node.getCurrentTickInfo()).thenReturn(Mono.just(TickInfo.builder().tick(42).build()));
        when(node.getTickTransactions(42)).thenReturn(Flux.just(stx));

        StepVerifier.create(computorService.getTickTransactions(42).log())
                .expectNext(stx)
                .verifyComplete();
    }

    @Test
    void getQxFees() {
        QxFees qxFees = QxFees.builder().build();
        when(node.getQxFees()).thenReturn(Mono.just(qxFees));
        StepVerifier.create(computorService.getQxFees())
                .expectNext(qxFees)
                .verifyComplete();
    }

    @Test
    void getQxAssetOrders() {
        when(identityUtil.getPublicKeyFromIdentity("foo")).thenReturn(new byte[] { 'a', 'b', 'c' });
        List<AssetOrder> expected = List.of(AssetOrder.builder()
                .entity(new byte[] { 1, 2, 3})
                .price(4)
                .numberOfShares(5)
                .build());
        when(node.getQxAssetOrders(any())).thenReturn(Mono.just(expected));

        StepVerifier.create(computorService.getQxAskAssetOrders("foo", "bar", 0))
                .expectNext(expected)
                .verifyComplete();

        StepVerifier.create(computorService.getQxBidAssetOrders("foo", "bar", 0))
                .expectNext(expected)
                .verifyComplete();

    }

    @Test
    void getQxEntityOrders() {
        byte[] entityPublicKey = new byte[0];
        when(identityUtil.getPublicKeyFromIdentity("foo")).thenReturn(entityPublicKey);
        List<EntityOrder> expected = List.of(EntityOrder.builder()
                .issuer(new byte[32])
                .assetName(AssetUtil.nameTo8Bytes("FOO"))
                .price(1)
                .numberOfShares(2)
                .build());

        when(node.getQxEntityOrders(any())).thenReturn(Mono.just(expected));

        StepVerifier.create(computorService.getQxAskEntityOrders("foo", 0))
                .expectNext(expected)
                .verifyComplete();

        StepVerifier.create(computorService.getQxBidEntityOrders("foo", 0))
                .expectNext(expected)
                .verifyComplete();

    }

}