package at.qubic.api.tcp;

import lombok.Getter;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.codec.binary.Hex;
import at.qubic.api.domain.MessageType;

import java.io.*;
import java.net.*;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

import static at.qubic.api.domain.QubicHeader.HEADER_LENGTH;

@Getter
@Slf4j
public class TcpTestServer {

    private final int port;
    private ServerSocket serverSocket;

    public TcpTestServer(int port) {
        this.port = port;
    }

    @SneakyThrows
    public TcpTestServer start() {
        this.serverSocket = new ServerSocket(port);
        log.info("Server is listening on port [{}].", port);
        return this;
    }

    public void stop() throws IOException {
        this.serverSocket.close();
    }

    public ExchangeThread respond(byte[]... data) {
        ExchangeThread serverThread = new ExchangeThread(data);
        serverThread.start();
        return serverThread;
    }

    public class ExchangeThread extends Thread {

        private final byte[][] data;

        @Getter
        private byte[] header = new byte[0];

        @Getter
        private byte[] payload = new byte[0];

        @Getter
        boolean receivedPayload = false;

        public ExchangeThread(byte[]... data) {
            this.data = data;
        }

        public void run() {
            try (Socket socket = serverSocket.accept();
                 InputStream input = socket.getInputStream();
                 OutputStream output = socket.getOutputStream()) {
                log.info("New client connected.");
                processPayload(input);
                sendResponse(output);
            } catch (IOException ioe) {
                throw new RuntimeException("Error responding with test response.", ioe);
            }
        }

        private void processPayload(InputStream input) throws IOException {

            header = input.readNBytes(HEADER_LENGTH);
            int size = getHeaderSize(header) - HEADER_LENGTH;
            log.debug("Payload size [{}].", size);
            payload = input.readNBytes(size);
            receivedPayload = true;
            log.debug("Payload: [{}]", Hex.encodeHexString(payload));
        }

        private void sendResponse(OutputStream output) throws IOException {
            for (byte[] bytes : data) {
                output.write(bytes);
                output.flush();
            }
            log.info("Wrote response.");
        }

        private int getHeaderSize(byte[] header) {
            int size = ByteBuffer.allocate(4)
                    .order(ByteOrder.LITTLE_ENDIAN)
                    .put(header[0])
                    .put(header[1])
                    .put(header[2])
                    .rewind() // go to beginning
                    .getInt();
            log.debug("Message size: [{}].", size);
            log.info("Received [{}].", MessageType.fromCode(header[3]));
            return size;
        }
    }

}
