package at.qubic.shell.command;

import at.qubic.shell.component.QearnLockComponent;
import at.qubic.shell.exception.InputValidationException;
import at.qubic.shell.exception.SendTransactionException;
import jakarta.validation.constraints.NotBlank;
import jakarta.validation.constraints.Positive;
import jakarta.validation.constraints.PositiveOrZero;
import lombok.extern.slf4j.Slf4j;
import org.springframework.shell.command.annotation.Command;
import org.springframework.shell.command.annotation.Option;
import org.springframework.shell.command.annotation.OptionValues;

import static at.qubic.shell.command.SharedCommandConstants.OPTION_DESCRIPTION_IDENTITY;
import static at.qubic.shell.command.SharedCommandConstants.OPTION_VALUE_PROVIDER_IDENTITY;
import static org.springframework.shell.command.CommandRegistration.OptionArity.EXACTLY_ONE;

// TODO test
@Slf4j
@Command(group = "qearn")
public class QearnProcedureCommands {

    private static final String DEFAULT_TICK_OFFSET = "5";
    private static final String DEFAULT_RETRIES = "0";

    private final QearnLockComponent qearnLockComponent;

    public QearnProcedureCommands(QearnLockComponent qearnLockComponent) {
        this.qearnLockComponent = qearnLockComponent;
    }

    @Command(command = "qearn unlock", description = "Unlock previously locked amount from qearn.")
    public String unlock(
            @OptionValues(provider = OPTION_VALUE_PROVIDER_IDENTITY)
            @Option(description = OPTION_DESCRIPTION_IDENTITY, arity = EXACTLY_ONE, required = true)
            @NotBlank
            String identity,
            @Option(description = "Epoch number.", arity = EXACTLY_ONE, required = true)
            @Positive
            int epoch,
            @Option(description = "Amount", arity = EXACTLY_ONE, required = true)
            @Positive
            long amount,
            @Option(description = "Tick offset.", defaultValue = DEFAULT_TICK_OFFSET, arity = EXACTLY_ONE)
            @Positive
            int offset,
            @Option(description = "Number of retries.",  defaultValue = DEFAULT_RETRIES, arity = EXACTLY_ONE)
            @PositiveOrZero
            int retries,
            @Option(description = "Ignore validations in case of 'false'.", defaultValue = "true", arity = EXACTLY_ONE)
            boolean validate
    ) {

        try {
            qearnLockComponent.unlock(identity, epoch, amount, offset, retries, validate).block();
            return "";
        } catch (SendTransactionException bse) {
            return handleException(bse);
        } catch (InputValidationException e) {
            return handleException(e);
        } catch (Exception e) {
            return handleException(e);
        }

    }

    @Command(command = "qearn lock", description = "Lock amount in qearn in current epoch.")
    public String lock(
            @OptionValues(provider = OPTION_VALUE_PROVIDER_IDENTITY)
            @Option(description = OPTION_DESCRIPTION_IDENTITY, arity = EXACTLY_ONE, required = true)
            @NotBlank
            String identity,
            @Option(description = "Amount", arity = EXACTLY_ONE, required = true)
            @Positive
            long amount,
            @Option(description = "Tick offset.", defaultValue = DEFAULT_TICK_OFFSET, arity = EXACTLY_ONE)
            @Positive
            int offset,
            @Option(description = "Number of retries.",  defaultValue = DEFAULT_RETRIES, arity = EXACTLY_ONE)
            @PositiveOrZero
            int retries,
            @Option(description = "Ignore validations in case of 'false'.", defaultValue = "true", arity = EXACTLY_ONE)
            boolean validate
    ) {

        try {
            qearnLockComponent.lock(identity, amount, offset, retries, validate).block();
            return "";
        } catch (SendTransactionException bse) {
            return handleException(bse);
        } catch (InputValidationException e) {
            return handleException(e);
        } catch (Exception e) {
            return handleException(e);
        }

    }

    private static String handleException(InputValidationException e) {
        log.warn("Input validation failed: {}", e.getMessage());
        return e.getMessage();
    }

    private static String handleException(SendTransactionException bse) {
        log.warn("Error trying to send transaction: {}", bse.getMessage());
        return "ERROR: " + bse.getMessage();
    }

    private static String handleException(Exception e) {
        log.error("Unexpected error", e);
        return "ERROR: " + e.getMessage();
    }

}
