package at.qubic.shell.command;

import jakarta.validation.constraints.NotEmpty;
import jakarta.validation.constraints.Positive;
import jakarta.validation.constraints.PositiveOrZero;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.collections4.ListUtils;
import org.apache.commons.lang3.StringUtils;
import at.qubic.api.crypto.IdentityUtil;
import at.qubic.api.domain.qx.response.AssetOrder;
import at.qubic.api.domain.qx.response.EntityOrder;
import at.qubic.api.domain.qx.response.QxFees;
import at.qubic.shell.command.service.QxService;
import at.qubic.shell.util.FormatUtil;
import at.qubic.shell.wallet.Assets;
import org.springframework.shell.command.annotation.Command;
import org.springframework.shell.command.annotation.Option;
import org.springframework.shell.command.annotation.OptionValues;
import reactor.core.publisher.Mono;
import reactor.util.function.Tuple2;

import java.util.*;

import static at.qubic.shell.command.SharedCommandConstants.OPTION_DESCRIPTION_IDENTITY;
import static at.qubic.shell.command.SharedCommandConstants.OPTION_VALUE_PROVIDER_IDENTITY;
import static org.springframework.shell.command.CommandRegistration.OptionArity.EXACTLY_ONE;

@Slf4j
@Command(group = "qx")
public class QxFunctionsCommands {

    private static final String OPTION_VALUE_PROVIDER_ASSET_NAME = "assetNameCompletionProvider";
    private static final String OPTION_VALUE_PROVIDER_ASSET_ISSUER = "assetIssuerCompletionProvider";
    private static final String STRING_NO_ORDERS = "No %s orders.\n";
    private static final String DEFAULT_ORDER_BOOK_LIMIT = "25";
    private static final String ASK = "ASK";
    private static final String BID = "BID";

    private final FormatUtil formatUtil = new FormatUtil();
    private final QxService qxService;
    private final IdentityUtil identityUtil;
    private final Assets assets;

    public QxFunctionsCommands(QxService qxService, IdentityUtil identityUtil, Assets assets) {
        this.identityUtil = identityUtil;
        this.qxService = qxService;
        this.assets = assets;
    }

    @Command(command = "qx fees", description = "Show qx fees.")
    public String getFees() {
        QxFees fees = qxService.getQxFees().block();
        return fees == null ? "Error getting qx fees" : """
                Issuance Fee: %s qu
                Transfer Fee: %s qu
                Trading Fee:  %s %%""".formatted(formatUtil.formatUnsigned(fees.getAssetIssuanceFee()),
                formatUtil.formatUnsigned(fees.getTransferFee()),
                fees.getTradeFeePercentage());
    }

    @Command(command = "qx entity-orders", description = "Shows orders of an entity.")
    public String getEntityOrders(@OptionValues(provider = OPTION_VALUE_PROVIDER_IDENTITY)
                                      @Option(description = OPTION_DESCRIPTION_IDENTITY, arity = EXACTLY_ONE, required = true)
                                      @NotEmpty String identity,
                                  @Option(description = "Show ASK orders.", arity = EXACTLY_ONE, defaultValue = "true")
                                  boolean ask,
                                  @Option(description = "Show BID orders.", arity = EXACTLY_ONE, defaultValue = "true")
                                      boolean bid,
                                  @Option(description = "Limit number of ask orders.", arity = EXACTLY_ONE, defaultValue = "256")
                                      @Positive int limitAsk,
                                  @Option(description = "Limit number of bid orders.", arity = EXACTLY_ONE, defaultValue = "256")
                                      @Positive int limitBid,
                                  @Option(description = "Skip number of ask orders.", arity = EXACTLY_ONE, defaultValue = "0")
                                      @PositiveOrZero int offsetAsk,
                                  @Option(description = "Skip number of bid orders.", arity = EXACTLY_ONE, defaultValue = "0")
                                      @PositiveOrZero int offsetBid) {

        if (!identityUtil.isValidIdentity(identity)) {
            return invalidEntityMessage(identity);
        } else {
            Mono<String> askOrders = Mono.defer(() -> qxService.getEntityAskOrders(identity, offsetAsk))
                    .map(orders -> printEntityOrders(orders, limitAsk, ASK));
            Mono<String> bidOrders = Mono.defer(() -> qxService.getEntityBidOrders(identity, offsetBid))
                    .map(orders -> printEntityOrders(orders, limitBid, BID));
            if (ask && bid) {
                Tuple2<String, String> orders = Objects.requireNonNull(Mono.zip(askOrders, bidOrders)
                        .block());
                return String.format("Orders for [%s]:\n%s%s", identity, orders.getT1(), orders.getT2());
            } else if (ask) {
                return String.format("Orders for [%s]:\n%s", identity, askOrders.block());
            } else if (bid) {
                return String.format("Orders for [%s]:\n%s", identity, bidOrders.block());
            } else {
                return "Please select to show either ask or bid orders.";
            }
        }
    }

    @Command(command = "qx order-book", description = "Shows orders for an asset.")
    public String getAssetOrders(
            @OptionValues(provider = OPTION_VALUE_PROVIDER_ASSET_NAME)
            @Option(description = "Name of the asset", arity = EXACTLY_ONE, required = true)
            @NotEmpty String assetName,
            @OptionValues(provider = OPTION_VALUE_PROVIDER_ASSET_ISSUER)
            @Option (description = "Issuer of the asset.", arity = EXACTLY_ONE)
            String issuer,
            @Option(description = "Show ASK orders.", arity = EXACTLY_ONE, defaultValue = "true") boolean ask,
            @Option(description = "Show BID orders.", arity = EXACTLY_ONE, defaultValue = "true") boolean bid,
            @Option(description = "Limit number of ask orders.", arity = EXACTLY_ONE, defaultValue = DEFAULT_ORDER_BOOK_LIMIT) @Positive int limitAsk,
            @Option(description = "Limit number of bid orders.", arity = EXACTLY_ONE, defaultValue = DEFAULT_ORDER_BOOK_LIMIT) @Positive int limitBid,
            @Option(description = "Skip number of ask orders.", arity = EXACTLY_ONE, defaultValue = "0") @PositiveOrZero int offsetAsk,
            @Option(description = "Skip number of bid orders.", arity = EXACTLY_ONE, defaultValue = "0") @PositiveOrZero int offsetBid
    ) {
        String responseMessage;
        final String assetIssuer = StringUtils.isBlank(issuer) ? getIssuerForAssetName(assetName) : issuer;
        if (!identityUtil.isValidIdentity(assetIssuer)) {
            responseMessage = invalidEntityMessage(assetIssuer);
        } else {
            Mono<String> askOrders = Mono.defer(() -> qxService.getAssetAskOrders(assetIssuer, assetName, offsetAsk))
                    .map(orders -> printAssetOrders(orders, limitAsk, Operation.ASK));
            Mono<String> bidOrders = Mono.defer(() -> qxService.getAssetBidOrders(assetIssuer, assetName, offsetBid))
                    .map(orders -> printAssetOrders(orders, limitBid, Operation.BID));
            if (ask && bid) {
                Tuple2<String, String> orders = Objects.requireNonNull(Mono.zip(bidOrders, askOrders)
                        .block());
                responseMessage = String.format("Orders for [%s] of issuer [%s]:\n\n%s-\n%s", assetName, assetIssuer, orders.getT1(), orders.getT2());
            } else if (ask) {
                responseMessage = String.format("Orders for [%s] of issuer[%s]:\n\n%s", assetName, assetIssuer, askOrders.block());
            } else if (bid) {
                responseMessage = String.format("Orders for [%s] of issuer[%s]:\n\n%s", assetName, assetIssuer, bidOrders.block());
            } else {
                responseMessage = "Please select to show either ask or bid orders.";
            }
        }
        return responseMessage;
    }

    private String printEntityOrders(List<EntityOrder> orders, int limit, String operation) {
        StringBuilder sb = new StringBuilder();
        orders.stream()
                .limit(limit)
                .forEach(eo -> sb.append("""
                                %s %6s Shares: %s Price: %s Issuer: %s
                                """.formatted(
                        operation,
                        eo.getAssetNameString(),
                        formatUtil.formatUnsigned(eo.getNumberOfShares()),
                        formatUtil.formatUnsigned(eo.getPrice()),
                        identityUtil.getIdentityFromPublicKey(eo.getIssuer()))));
        return sb.isEmpty() ? String.format(STRING_NO_ORDERS, operation) : sb.toString();
    }

    private String printAssetOrders(List<AssetOrder> orders, int limit, Operation operation) {
        StringBuilder sb = new StringBuilder();
        if (CollectionUtils.size(orders) > 0) {
            List<AssetOrder> limited = new ArrayList<>(ListUtils.partition(orders, limit).getFirst());
            if (operation == Operation.BID) { // show bids in ascending order
                Collections.reverse(limited);
            }
            limited.forEach(ao -> sb.append("""
                                %s from %s Price: %s Shares: %s
                                """.formatted(
                            operation,
                            identityUtil.getIdentityFromPublicKey(ao.getEntity()),
                            formatUtil.formatUnsigned(ao.getPrice()),
                            formatUtil.formatUnsigned(ao.getNumberOfShares()))));
        }
        return sb.isEmpty() ? String.format(STRING_NO_ORDERS, operation) : sb.toString();
    }

    private String getIssuerForAssetName(String name) {
        return assets.getUniqueAssetForName(name).getIssuer();
    }

    private static String invalidEntityMessage(String identity) {
        return String.format("Identity [%s] is invalid.", identity);
    }

    private enum Operation {
        BID,
        ASK
    }

}
