package at.qubic.shell.command;

import at.qubic.api.crypto.IdentityUtil;
import at.qubic.api.domain.qx.request.*;
import at.qubic.shell.component.QxOrderComponent;
import at.qubic.shell.exception.InputValidationException;
import at.qubic.shell.exception.SendTransactionException;
import at.qubic.shell.exception.UnknownAssetException;
import at.qubic.shell.wallet.Asset;
import at.qubic.shell.wallet.Assets;
import jakarta.validation.constraints.NotEmpty;
import jakarta.validation.constraints.Positive;
import jakarta.validation.constraints.PositiveOrZero;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.shell.command.annotation.Command;
import org.springframework.shell.command.annotation.Option;
import org.springframework.shell.command.annotation.OptionValues;

import static at.qubic.shell.command.SharedCommandConstants.OPTION_VALUE_PROVIDER_IDENTITY;
import static org.springframework.shell.command.CommandRegistration.OptionArity.EXACTLY_ONE;

@Slf4j
@Command(group = "qx")
public class QxProcedureCommands {

    private static final String OPTION_VALUE_PROVIDER_ASSET_NAME = "assetNameCompletionProvider";
    private static final String OPTION_VALUE_PROVIDER_ASSET_ISSUER = "assetIssuerCompletionProvider";
    private static final String DEFAULT_TICK_OFFSET = "5";
    private static final String DEFAULT_RETRIES = "0";

    private final IdentityUtil identityUtil;
    private final QxOrderComponent orderComponent;
    private final Assets assets;

    public QxProcedureCommands(IdentityUtil identityUtil, QxOrderComponent orderComponent, Assets assets) {
        this.identityUtil = identityUtil;
        this.orderComponent = orderComponent;
        this.assets = assets;
    }

    @Command(command = "qx add-bid", description = "Adds bid (buy) order for an asset.")
    public String addAssetBidOrder(
            @OptionValues(provider = OPTION_VALUE_PROVIDER_IDENTITY)
            @Option(description = "Identity (source).", required = true)
            @NotEmpty
            String from,
            @OptionValues(provider = OPTION_VALUE_PROVIDER_ASSET_NAME)
            @Option(description = "Name of the asset", arity = EXACTLY_ONE, required = true)
            @NotEmpty String assetName,
            @OptionValues(provider = OPTION_VALUE_PROVIDER_ASSET_ISSUER)
            @Option (description = "Issuer of the asset.", arity = EXACTLY_ONE)
            String issuer,
            @Option(description = "Limit price to pay per share.", arity = EXACTLY_ONE, required = true)
            @Positive
            long pricePerShare,
            @Option(description = "Number of desired shares.", arity = EXACTLY_ONE, required = true)
            @Positive
            long numberOfShares,
            @Option(description = "Tick offset.",
                    defaultValue = DEFAULT_TICK_OFFSET, arity = EXACTLY_ONE)
            @Positive
            int offset,
            @Option(description = "Number of retries.",
                    defaultValue = DEFAULT_RETRIES, arity = EXACTLY_ONE)
            @PositiveOrZero
            int retries,
            @Option(description = "Validate transaction before sending.",
                    defaultValue = "true", arity = EXACTLY_ONE)
            boolean validate
    ) {
        try {
            final Asset asset = calculateAsset(assetName, issuer);
            QxAssetOrder order = createAddBidOrder(pricePerShare, numberOfShares, asset);
            orderComponent.sendQxOrder(from, order, offset, retries, validate)
                    .block();
            return "";
        } catch (InputValidationException ive) {
            return handleException(ive);
        } catch (SendTransactionException ste) {
            return handleException(ste);
        } catch (UnknownAssetException uae) {
            return returnUnknownAssetMessage(uae);
        }
    }

    @Command(command = "qx add-ask", description = "Adds ask (sell) order for an asset.")
    public String addAssetAskOrder(
                                    @OptionValues(provider = OPTION_VALUE_PROVIDER_IDENTITY)
                                    @Option(description = "Identity (source).",
                                            required = true)
                                    @NotEmpty
                                    String from,
                                    @OptionValues(provider = OPTION_VALUE_PROVIDER_ASSET_NAME)
                                    @Option(description = "Name of the asset",
                                            arity = EXACTLY_ONE,
                                            required = true)
                                    @NotEmpty String assetName,
                                    @OptionValues(provider = OPTION_VALUE_PROVIDER_ASSET_ISSUER)
                                    @Option (description = "Issuer of the asset.",
                                            arity = EXACTLY_ONE)
                                    String issuer,
                                    @Option(description = "Price per share.",
                                            arity = EXACTLY_ONE,
                                            required = true)
                                    @Positive
                                    long pricePerShare,
                                    @Option(description = "Number of shares to sell.",
                                            arity = EXACTLY_ONE,
                                            required = true)
                                    @Positive
                                    long numberOfShares,
                                    @Option(description = "Tick offset.",
                                            defaultValue = DEFAULT_TICK_OFFSET,
                                            arity = EXACTLY_ONE)
                                    @Positive
                                    int offset,
                                    @Option(description = "Number of retries.",
                                            defaultValue = DEFAULT_RETRIES,
                                            arity = EXACTLY_ONE)
                                    @PositiveOrZero
                                    int retries,
                                    @Option(description = "Validate transaction before sending.",
                                            defaultValue = "true",
                                            arity = EXACTLY_ONE)
                                    boolean validate
    ) {
        try {
            final Asset asset = calculateAsset(assetName, issuer);
            QxAssetOrder order = createAddAskOrder(pricePerShare, numberOfShares, asset);
            orderComponent.sendQxOrder(from, order, offset, retries, validate)
                    .block();
            return "";
        } catch (InputValidationException ive) {
            return handleException(ive);
        } catch (SendTransactionException ste) {
            return handleException(ste);
        } catch (UnknownAssetException uae) {
            return returnUnknownAssetMessage(uae);
        }
    }

    @Command(command = "qx remove-bid", description = "Removes bid (buy) order for an asset.")
    public String removeAssetBidOrder(@OptionValues(provider = OPTION_VALUE_PROVIDER_IDENTITY)
                                    @Option(description = "Identity (source).",
                                            required = true)
                                    @NotEmpty
                                    String from,
                                    @OptionValues(provider = OPTION_VALUE_PROVIDER_ASSET_NAME)
                                    @Option(description = "Name of the asset",
                                            arity = EXACTLY_ONE,
                                            required = true)
                                    @NotEmpty String assetName,
                                    @OptionValues(provider = OPTION_VALUE_PROVIDER_ASSET_ISSUER)
                                    @Option (description = "Issuer of the asset.",
                                            arity = EXACTLY_ONE)
                                    String issuer,
                                    @Option(description = "Price per share (from original order).",
                                            arity = EXACTLY_ONE,
                                            required = true)
                                    @Positive
                                    long pricePerShare,
                                    @Option(description = "Number of shares (to be removed from order).",
                                            arity = EXACTLY_ONE,
                                            required = true)
                                    @Positive
                                    long numberOfShares,
                                    @Option(description = "Tick offset.",
                                            defaultValue = DEFAULT_TICK_OFFSET,
                                            arity = EXACTLY_ONE)
                                    @Positive
                                    int offset,
                                    @Option(description = "Number of retries.",
                                            defaultValue = DEFAULT_RETRIES, arity = EXACTLY_ONE)
                                    @PositiveOrZero
                                    int retries,
                                    @Option(description = "Validate transaction before sending.",
                                            defaultValue = "true", arity = EXACTLY_ONE)
                                    boolean validate
    ) {
        try {
            final Asset asset = calculateAsset(assetName, issuer);
            QxAssetOrder order = createRemoveBidOrder(pricePerShare, numberOfShares, asset);
            orderComponent.sendQxOrder(from, order, offset, retries, validate)
                    .block();
            return "";
        } catch (InputValidationException ive) {
            return handleException(ive);
        } catch (SendTransactionException ste) {
            return handleException(ste);
        } catch (UnknownAssetException uae) {
            return returnUnknownAssetMessage(uae);
        }
    }

    @Command(command = "qx remove-ask", description = "Removes ask (sell) order for an asset.")
    public String removeAssetAskOrder(@OptionValues(provider = OPTION_VALUE_PROVIDER_IDENTITY)
                                       @Option(description = "Identity (source).", required = true)
                                       @NotEmpty
                                       String from,
                                       @OptionValues(provider = OPTION_VALUE_PROVIDER_ASSET_NAME)
                                       @Option(description = "Name of the asset", arity = EXACTLY_ONE, required = true)
                                       @NotEmpty String assetName,
                                       @OptionValues(provider = OPTION_VALUE_PROVIDER_ASSET_ISSUER)
                                       @Option (description = "Issuer of the asset.", arity = EXACTLY_ONE)
                                       String issuer,
                                       @Option(description = "Price per share (from original order).",
                                               arity = EXACTLY_ONE, required = true)
                                       @Positive
                                       long pricePerShare,
                                       @Option(description = "Number of shares (to be removed from order).",
                                               arity = EXACTLY_ONE, required = true)
                                       @Positive
                                       long numberOfShares,
                                       @Option(description = "Tick offset.",
                                               defaultValue = DEFAULT_TICK_OFFSET, arity = EXACTLY_ONE)
                                       @Positive
                                       int offset,
                                       @Option(description = "Number of retries.",
                                               defaultValue = DEFAULT_RETRIES, arity = EXACTLY_ONE)
                                       @PositiveOrZero
                                       int retries,
                                       @Option(description = "Validate transaction before sending.",
                                               defaultValue = "true", arity = EXACTLY_ONE)
                                       boolean validate
    ) {
        try {
            final Asset asset = calculateAsset(assetName, issuer);
            QxAssetOrder order = createRemoveAskOrder(pricePerShare, numberOfShares, asset);
            orderComponent.sendQxOrder(from, order, offset, retries, validate)
                    .block();
            return "";
        } catch (InputValidationException ive) {
            return handleException(ive);
        } catch (SendTransactionException ste) {
            return handleException(ste);
        } catch (UnknownAssetException uae) {
            return returnUnknownAssetMessage(uae);
        }
    }

    @Command(command = "qx transfer-asset", description = "Sends assets from one entity to another.")
    public String transferAsset(
         @OptionValues(provider = OPTION_VALUE_PROVIDER_IDENTITY)
         @Option(description = "Identity (source).", required = true)
         @NotEmpty
         String from,
         @OptionValues(provider = OPTION_VALUE_PROVIDER_IDENTITY)
         @Option(description = "Identity (destination).", required = true)
         @NotEmpty
         String to,
         @OptionValues(provider = OPTION_VALUE_PROVIDER_ASSET_NAME)
         @Option(description = "Name of the asset", arity = EXACTLY_ONE, required = true)
         @NotEmpty String assetName,
         @OptionValues(provider = OPTION_VALUE_PROVIDER_ASSET_ISSUER)
         @Option (description = "Issuer of the asset.", arity = EXACTLY_ONE)
         String issuer,
         @Option(description = "Number of shares (to be removed from order).",
                 arity = EXACTLY_ONE, required = true)
         @Positive
         long numberOfShares,
         @Option(description = "Amount of qus sent with the transaction. Should be equal to QX transfer fee.",
                 defaultValue = "100", arity = EXACTLY_ONE)
         @PositiveOrZero
         long amount,
         @Option(description = "Tick offset.",
                 defaultValue = DEFAULT_TICK_OFFSET, arity = EXACTLY_ONE)
         @Positive
         int offset,
         @Option(description = "Number of retries.",
                 defaultValue = DEFAULT_RETRIES, arity = EXACTLY_ONE)
         @PositiveOrZero
         int retries,
         @Option(description = "Validate transaction before sending.",
                 defaultValue = "true", arity = EXACTLY_ONE)
         boolean validate
    ) {

        try {
            orderComponent.validateIdentity(to);
            final Asset asset = calculateAsset(assetName, issuer);
            QxTransferAssetOrder order = createTransferAssetOrder(to, numberOfShares, asset, amount);
            orderComponent.sendQxTransfer(from, order, offset, retries, validate)
                    .block();
            return "";
        } catch (InputValidationException ive) {
            return handleException(ive);
        } catch (SendTransactionException ste) {
            return handleException(ste);
        } catch (UnknownAssetException uae) {
            return returnUnknownAssetMessage(uae);
        }

    }

    private QxAssetOrder createAddBidOrder(long pricePerShare, long numberOfShares, Asset asset) {
        return QxAddBidOrder.builder()
                .orderData(buildOrderData(pricePerShare, numberOfShares, asset))
                .build();
    }

    private QxAssetOrder createAddAskOrder(long pricePerShare, long numberOfShares, Asset asset) {
        return QxAddAskOrder.builder()
                .orderData(buildOrderData(pricePerShare, numberOfShares, asset))
                .build();
    }

    private QxAssetOrder createRemoveBidOrder(long pricePerShare, long numberOfShares, Asset asset) {
        return QxRemoveBidOrder.builder()
                .orderData(buildOrderData(pricePerShare, numberOfShares, asset))
                .build();
    }

    private QxAssetOrder createRemoveAskOrder(long pricePerShare, long numberOfShares, Asset asset) {
        return QxRemoveAskOrder.builder()
                .orderData(buildOrderData(pricePerShare, numberOfShares, asset))
                .build();
    }

    private QxTransferAssetOrder createTransferAssetOrder(String destinationIdentity, long numberOfShares, Asset asset,
                                                          long amount) {
        return QxTransferAssetOrder.builder()
                .newOwnerAndPossessor(identityUtil.getPublicKeyFromIdentity(destinationIdentity))
                .assetName(asset.getName())
                .assetIssuer(identityUtil.getPublicKeyFromIdentity(asset.getIssuer()))
                .amount(amount) // TODO set dynamically
                .numberOfUnits(numberOfShares)
                .build();
    }

    private QxAssetOrderData buildOrderData(long pricePerShare, long numberOfShares, Asset asset) {
        return QxAssetOrderData.builder()
                .assetName(asset.getName())
                .assetIssuer(identityUtil.getPublicKeyFromIdentity(asset.getIssuer()))
                .price(pricePerShare)
                .numberOfShares(numberOfShares)
                .build();
    }

    private Asset calculateAsset(String assetName, String issuer) throws InputValidationException {
        Asset asset = StringUtils.isBlank(issuer)
                ? assets.getUniqueAssetForName(assetName)
                : assets.getAsset(assetName, issuer).orElse(
                    Asset.builder().name(assetName).issuer(issuer).build());
        orderComponent.validateIdentity(asset.getIssuer());
        return asset;
    }

    private static String returnUnknownAssetMessage(UnknownAssetException uae) {
        log.warn("Error resolving asset: {}", uae.getMessage());
        return uae.getMessage();
    }

    private static String handleException(InputValidationException e) {
        log.warn("Input validation failed: {}", e.getMessage());
        return e.getMessage();
    }

    private static String handleException(SendTransactionException bse) {
        log.warn("Error trying to send transaction: {}", bse.getMessage());
        return "ERROR: " + bse.getMessage();
    }

}
