package at.qubic.shell.command;

import at.qubic.shell.component.QuTransferComponent;
import at.qubic.shell.exception.InputValidationException;
import at.qubic.shell.exception.SendTransactionException;
import jakarta.validation.constraints.NotBlank;
import jakarta.validation.constraints.Positive;
import jakarta.validation.constraints.PositiveOrZero;
import lombok.extern.slf4j.Slf4j;
import org.springframework.shell.command.annotation.Command;
import org.springframework.shell.command.annotation.Option;
import org.springframework.shell.command.annotation.OptionValues;

import static at.qubic.shell.command.SharedCommandConstants.OPTION_VALUE_PROVIDER_IDENTITY;
import static org.springframework.shell.command.CommandRegistration.OptionArity.EXACTLY_ONE;

@Slf4j
@Command(group = "entity")
public class SendTransactionCommand {

    private static final String DEFAULT_TICK_OFFSET = "5";
    private static final String DEFAULT_RETRIES = "0";
    private final QuTransferComponent component;

    public SendTransactionCommand(QuTransferComponent component) {
        this.component = component;
    }

    @Command(command = "send", description = "Transfer qubics from one identity to another.")
    public String send(@Option(description = "Amount in qu.", required = true)
                           @PositiveOrZero
                           long amount,
                           @OptionValues(provider = OPTION_VALUE_PROVIDER_IDENTITY)
                           @Option(description = "Identity (source).", required = true)
                           @NotBlank
                           String from,
                       @OptionValues(provider = OPTION_VALUE_PROVIDER_IDENTITY)
                           @Option(description = "Identity (destination).", required = true)
                           @NotBlank
                           String to,
                       @Option(description = "Tick offset.",
                               defaultValue = DEFAULT_TICK_OFFSET, arity = EXACTLY_ONE)
                           @Positive
                           int offset,
                       @Option(description = "Number of retries.",
                               defaultValue = DEFAULT_RETRIES, arity = EXACTLY_ONE)
                           @PositiveOrZero
                           int retries,
                       @Option(description = "Validate transaction before sending.",
                               defaultValue = "true", arity = EXACTLY_ONE)
                           boolean validate
    ) {
        String output;
        try {
            log.info("Send [{}] from [{}] to [{}].", amount, from, to);
            component.transfer(amount, from, to, offset, retries, validate)
                    .block();
            log.info("Sent [{}] to [{}] successfully.", amount, from);
            output = "";
        } catch (SendTransactionException e) {
            output = handleException(e);
        } catch (InputValidationException e) {
            output = handleException(e);
        }
        return output;
    }

    private static String handleException(InputValidationException e) {
        log.warn("Input validation failed: {}", e.getMessage());
        return e.getMessage();
    }

    private static String handleException(SendTransactionException bse) {
        log.warn("Error trying to send transaction: {}", bse.getMessage());
        return "ERROR: " + bse.getMessage();
    }

}
