package at.qubic.shell.command.service;

import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import at.qubic.api.crypto.IdentityUtil;
import at.qubic.api.service.ComputorService;
import at.qubic.api.util.AssetUtil;
import at.qubic.shell.util.FormatUtil;

import java.util.List;

@Slf4j
public class EntityCommandService {

    private final FormatUtil formatUtil = new FormatUtil();
    private final IdentityUtil identityUtil;
    private final ComputorService computorService;

    public EntityCommandService(ComputorService computorService, IdentityUtil identityUtil) {
        this.identityUtil = identityUtil;
        this.computorService = computorService;
    }

    public String getEntity(String identity) {
        log.debug("Query entity: [{}].", identity);
        if (!identityUtil.isValidIdentity(identity)) {
            return invalidEntityMessage(identity);
        }

        String result = computorService.getEntity(identity)
                .retry(2)
                .map(entity ->
                    """
                    
                    Identity: %s
                    Balance: %s
                    Incoming amount: %s
                    Outgoing amount: %s
                    Incoming transfers: %s
                    Outgoing transfers: %s
                    Latest incoming tick: %s
                    Latest outgoing tick: %s
                    Tick: %s
                    """.formatted(identity,
                            formatUtil.formatUnsigned(entity.getIncomingAmount() - entity.getOutgoingAmount()),
                            formatUtil.formatUnsigned(entity.getIncomingAmount()),
                            formatUtil.formatUnsigned(entity.getOutgoingAmount()),
                            entity.getNumberOfIncomingTransfers(),
                            entity.getNumberOfOutgoingTransfers(),
                            entity.getLatestIncomingTransferTick(),
                            entity.getLatestOutgoingTransferTick(),
                            entity.getTick()))
                .onErrorComplete()
                .block();

        // result is null on error or empty
        return StringUtils.defaultIfBlank(result, String.format("Could not retrieve entity [%s].", identity));
    }

    public String getOwnedAssets(String identity) {
        log.debug("Query owned assets: [{}].", identity);
        return identityUtil.isValidIdentity(identity)
                ? ownedAssetsMessage(identity)
                : invalidEntityMessage(identity);

    }

    public String getPossessedAssets(String identity) {
        log.debug("Query possessed assets: [{}].", identity);
        return identityUtil.isValidIdentity(identity)
                ? possessedAssetsMessage(identity)
                : invalidEntityMessage(identity);
    }

    private String ownedAssetsMessage(String identity) {
        return computorService.getOwnedAssets(identity)
                .retry(2)
                .map(asset ->
                        """
                        
                        Issuer: %s
                        Name: %s
                        Units: %s
                        Managing contract index: %d
                        Issuance index: %d
                        Tick: %d
                        """.formatted(
                                identityFromPublicKey(asset.getIssuanceAsset().getPublicKey()),
                                AssetUtil.nameToString(asset.getIssuanceAsset().getName()),
                                formatUtil.formatUnsigned(asset.getAsset().getNumberOfUnits()),
                                asset.getAsset().getManagingContractIndex(),
                                asset.getAsset().getIssuanceIndex(),
                                asset.getTick()
                        )
                )
                .collectList() // returns empty list on empty mono
                .filter(list -> !list.isEmpty())
                .map(list -> formatOutput("Owned", identity, list))
                .defaultIfEmpty(String.format("[%s] owns no assets.", identity))
                .onErrorReturn(String.format("Could not retrieve owned assets for [%s].", identity))
                .block();
    }

    private String possessedAssetsMessage(String identity) {
        return computorService.getPossessedAssets(identity)
                .retry(2)
                .map(asset ->
                        """
                        
                        Issuer: %s
                        Name: %s
                        Units: %s
                        Managing contract index: %d
                        Owner: %s
                        Owner index: %d
                        Tick: %d
                        """.formatted(
                                identityFromPublicKey(asset.getIssuanceAsset().getPublicKey()),
                                AssetUtil.nameToString(asset.getIssuanceAsset().getName()),
                                formatUtil.formatUnsigned(asset.getAsset().getNumberOfUnits()),
                                asset.getAsset().getManagingContractIndex(),
                                identityFromPublicKey(asset.getOwnershipAsset().getPublicKey()),
                                asset.getAsset().getOwnershipIndex(),
                                asset.getTick()
                        )
                )
                .collectList() // returns empty list on empty mono
                .filter(list -> !list.isEmpty())
                .map(list -> formatOutput("Possessed", identity, list))
                .defaultIfEmpty(String.format("[%s] possesses no assets.", identity))
                .onErrorReturn(String.format("Could not retrieve possessed assets for [%s].", identity))
                .block();
    }

    private static String invalidEntityMessage(String identity) {
        return String.format("Identity [%s] is invalid.", identity);
    }

    private static String formatOutput(String prefix, String identity, List<String> list) {
        return String.format("%s assets of identity %s:\n%s", prefix, identity, String.join("\n", list));
    }

    private String identityFromPublicKey(byte[] publicKey) {
        return ArrayUtils.isNotEmpty(publicKey)
                ? identityUtil.getIdentityFromPublicKey(publicKey)
                : "N/A";
    }

}
