package at.qubic.shell.command.service;

import at.qubic.api.domain.std.SignedTransaction;
import at.qubic.api.domain.std.response.TickInfo;
import at.qubic.api.service.ComputorService;
import at.qubic.shell.exception.NotIncludedException;
import lombok.Builder;
import lombok.EqualsAndHashCode;
import lombok.Getter;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;
import reactor.util.retry.Retry;
import reactor.util.retry.RetrySpec;

import java.time.Duration;
import java.util.NoSuchElementException;

// TODO move out of here. Should not be in command service package
@Slf4j
public class InclusionCheck {

    private static final int NUM_RETRIES = 3;
    private final ComputorService computorService;
    private final Duration checkInterval;

    public InclusionCheck(ComputorService computorService, Duration checkInterval) {
        this.computorService = computorService;
        this.checkInterval = checkInterval;
    }

    public Flux<InclusionStatus> checkForInclusion(int targetTick, String transactionHash) {
        return Flux.interval(checkInterval, checkInterval)
                .flatMap(count -> getCurrentTickInfo())
                .takeUntil(ti -> ti.getTick() > targetTick)
                .flatMap(ti -> checkForInclusion(ti.getTick(), targetTick, transactionHash));
    }

    // FIXME it is possible to get positive inclusion status for empty ticks (seems that single nodes sometimes return
    //       tick transactions for empty ticks. How to validate that?)
    private Mono<InclusionStatus> checkForInclusion(int currentTick, int targetTick, String transactionHash) {
        if (currentTick > targetTick) {
            return getTickTransactions(targetTick)
                    .switchIfEmpty(tryAgain(targetTick))
                    .filter(tx -> StringUtils.equals(transactionHash, tx.getTransactionHash()))
                    .single()
                    .doOnNext(tx -> log.info("Transaction [{}] is included in tick [{}]. Current tick: [{}].", tx.getTransactionHash(), targetTick, currentTick))
                    .map(tx -> InclusionStatus.builder()
                            .checkedTick(targetTick)
                            .signedTransaction(tx)
                            .build())
                    .onErrorMap(NoSuchElementException.class, e -> notIncludedInTick(targetTick, transactionHash));
        } else {
            return Mono.just(InclusionStatus.builder()
                    .checkedTick(currentTick-1)
                    .build());
        }
    }

    // This is probably not true anymore as getTickTransactions checks the nodes tick first
    // FIXME in rare cases we get a node that is not up to date. In this case get tick transactions does not return
    //      transactions. We cannot infer that the transaction is not in the tick in this case. As a quick and dirty
    //      workaround we wait and try again.
    private Flux<SignedTransaction> tryAgain(int targetTick) {
        return Mono.delay(Duration.ofMillis(100))
                .doOnNext(x -> log.info("No transactions in target tick. Asking again..."))
                .thenMany(getTickTransactions(targetTick));
    }

    private Mono<TickInfo> getCurrentTickInfo() {
        return Mono.defer(computorService::getCurrentTickInfo)
                .retryWhen(getRetrySpec("getting tick info"));
    }

    private Flux<SignedTransaction> getTickTransactions(int tick) {
        return Flux.defer(() -> computorService.getTickTransactions(tick))
                .retryWhen(getRetrySpec(String.format("getting transactions for tick [%s]", tick)));
    }

    private static RetrySpec getRetrySpec(String message) {
        return Retry.max(NUM_RETRIES)
                .doBeforeRetry(retrySignal -> log.warn("Error {}. Retries: [{}]. Error: {}", message, retrySignal.totalRetries(), retrySignal.failure().getMessage()))
                .doAfterRetry(retrySignal -> log.info("Retry {}.", message));
    }

    private static NotIncludedException notIncludedInTick(int targetTick, String transactionHash) {
        String message = String.format(
                "Transaction [%s] is not included in target tick [%s].",
                transactionHash,
                targetTick);
        log.info(message);
        return new NotIncludedException(message);
    }

    @Getter
    @Builder
    @EqualsAndHashCode
    public static class InclusionStatus {

        private final int checkedTick;
        private final SignedTransaction signedTransaction;

    }

}
