package at.qubic.shell.command.service;

import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.exception.ExceptionUtils;
import at.qubic.api.crypto.IdentityUtil;
import at.qubic.api.crypto.KangarooTwelve;
import at.qubic.shell.exception.CommandFailureException;
import at.qubic.shell.exception.KeyManagementException;
import at.qubic.shell.wallet.KeyManager;
import at.qubic.shell.wallet.Wallet;
import at.qubic.shell.wallet.Wallets;

import java.security.UnrecoverableKeyException;

@Slf4j
public class KeyService {

    private final KeyManager keyManager;
    private final Wallets wallets;
    private final IdentityUtil identityUtil;

    public KeyService(KeyManager keyManager, Wallets wallets, IdentityUtil identityUtil) {
        this.keyManager = keyManager;
        this.wallets = wallets;
        this.identityUtil = identityUtil;
    }

    public String tryToUpdatePassword(String newPassword) {
        String response;
        try {
            keyManager.updatePassword(newPassword.toCharArray());
            response = "Password successfully updated.";
        } catch (Exception e) {
            String message = "Password update failed.";
            log.warn("{}: {}", message, e.getCause() == null ? e.getMessage() : e.getCause().getMessage());
            throw new CommandFailureException(message, e);
        }
        return response;
    }

    public String tryToUnlockKeyStore(String password) {
        try {
            keyManager.unlock(password.toCharArray());
            // maybe wallets were removed when key store was locked.
            // remove the keys for these wallets now.
            deleteUnknownWalletKeys();
            return "Unlocked wallet.";
        } catch (KeyManagementException kme) {
            String message = "Cannot unlock wallet.";
            log.warn("{} {}", message, kme.getCause() == null ? kme.getMessage() : kme.getCause().getMessage());
            if (ExceptionUtils.getRootCause(kme) instanceof UnrecoverableKeyException) {
                return message + " Wrong password.";
            } else {
                throw new CommandFailureException(message, kme);
            }
        }
    }

    public void deleteUnknownWalletKeys() {
        keyManager.getKeyAliases().stream()
                // our keystore implementation stores aliases lower case
                .map(StringUtils::upperCase) // careful: alias might be lower case
                .filter(a -> !wallets.containsWalletIdentity(a))
                .forEach(keyManager::deleteKeyEntry);
    }

    public boolean isValidSeed(String seed) {
        return identityUtil.isValidSeed(seed);
    }

    public String addSeed(String seed) {
        if (isValidSeed(seed)) {
            byte[] subSeed = identityUtil.getSubSeedFromSeed(seed);
            String identity = identityUtil.getIdentityFromPublicKey(identityUtil.getPublicKeyFromSubSeed(subSeed));
            byte[] nonceK = KangarooTwelve.hash(subSeed, 64);
            keyManager.setPrivateKey(identity, nonceK);
            if (!wallets.containsWalletIdentity(identity)) {
                wallets.putWallet(new Wallet(identity));
            }
            return identity;
        } else {
            throw new IllegalArgumentException("Invalid seed.");
        }
    }

    public boolean containsKeyForIdentity(String identity) {
        return keyManager.containsKeyAlias(identity);
    }

    public boolean verifyIdentity(String identity) {
        boolean matches = false;
        if (containsKeyForIdentity(identity)) {
            byte[] nonceK = keyManager.getPrivateKey(identity);
            byte[] privateKey = ArrayUtils.subarray(nonceK, 0, 32);
            byte[] publicKey = identityUtil.getPublicKeyFromPrivateKey(privateKey);
            String loadedIdentity = identityUtil.getIdentityFromPublicKey(publicKey);
            matches = StringUtils.equals(identity, loadedIdentity);
            if (!matches) {
                log.warn("Stored key does not match. Expected [{}] but got [{}].", identity, loadedIdentity);
            }
        } else {
            log.warn("There is no key stored for identity [{}].", identity);
        }
        return matches;
    }

    public void lock() {
        keyManager.lock();
    }

    public boolean isLocked() {
        return keyManager.isLocked();
    }

    public void deletePrivateKey(String identity) {
        keyManager.deleteKeyEntry(identity);
    }

    public boolean isValidIdentity(String identity) {
        return identityUtil.isValidIdentity(identity);
    }

}
