package at.qubic.shell.command.service;

import at.qubic.api.domain.qearn.response.UserLockInfo;
import at.qubic.api.domain.qearn.response.UserLockStatus;
import at.qubic.api.domain.std.response.TickInfo;
import at.qubic.api.service.ComputorService;
import at.qubic.shell.util.FormatUtil;
import jakarta.validation.constraints.NotBlank;
import jakarta.validation.constraints.NotEmpty;
import org.apache.commons.lang3.StringUtils;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

// TODO test
public class QearnService {

    private static final int NUM_RETRIES = 3;
    private final FormatUtil formatUtil = new FormatUtil();
    private final ComputorService computorService;

    public QearnService(ComputorService computorService) {
        this.computorService = computorService;
    }

    public String getEpochLockInfo(int epoch) {
        return Mono.zip(
                        computorService.getQearnBurnedAndBoostedStatsForEpoch(epoch).retry(NUM_RETRIES),
                        computorService.getQearnLockInfoForEpoch(epoch).retry(NUM_RETRIES),
                        computorService.getQearnStateOfRound(epoch).retry(NUM_RETRIES))
                .map(tuple -> """
                        
                        Qearn lock info for epoch: %d
                        Currently locked: %s
                        Current bonus: %s
                        Yield: %.3f%%
                        
                        Initially locked: %s
                        Initial bonus: %s
                        Early unlocked: %s (%.3f%%)
                        Burned amount: %s (%.3f%%)
                        Boosted amount: %s (%.3f%%)
                        Rewarded amount: %s (%.3f%%)
                        Status: %s
                        """.formatted(epoch,
                        formatUtil.formatUnsigned(tuple.getT2().getCurrentLockedAmount()),
                        formatUtil.formatUnsigned(tuple.getT2().getCurrentBonusAmount()),
                        tuple.getT2().getYieldPercentage(),
                        formatUtil.formatUnsigned(tuple.getT2().getLockedAmount()),
                        formatUtil.formatUnsigned(tuple.getT2().getBonusAmount()),
                        formatUtil.formatUnsigned(tuple.getT2().getLockedAmount() - tuple.getT2().getCurrentLockedAmount()),
                        ((tuple.getT2().getLockedAmount() - tuple.getT2().getCurrentLockedAmount()) / (double) tuple.getT2().getLockedAmount()) * 100,
                        formatUtil.formatUnsigned(tuple.getT1().getBurnedAmount()), tuple.getT1().getBurnedPercentAsDecimal(),
                        formatUtil.formatUnsigned(tuple.getT1().getBoostedAmount()), tuple.getT1().getBoostedPercentAsDecimal(),
                        formatUtil.formatUnsigned(tuple.getT1().getRewardedAmount()), tuple.getT1().getRewardedPercentAsDecimal(),
                        tuple.getT3().getStatus()

                )).block();
    }

    public String getUserLockInfo(String identity, int epoch) {
        return computorService.getQearnUserLockInfo(identity, epoch)
                .retry(NUM_RETRIES)
                .map(info -> """
                        
                        Qearn lock info for identity: %s
                        Epoch: %d
                        Locked amount: %s
                        """.formatted(identity, epoch,
                        formatUtil.formatUnsigned(info.getLockedAmount())))
                .block();
    }

    public String getUserLockInfo(@NotEmpty String identity) {
        return computorService.getQearnUserLockStatus(identity)
                .retry(NUM_RETRIES)
                .map(UserLockStatus::getEpochOffsetsWithLocks)
                .flatMap(offsets -> computorService.getCurrentTickInfo()
                        .retry(NUM_RETRIES)
                        .map(TickInfo::getEpoch)
                        .map(epoch -> offsets.stream().map(offset -> epoch - offset).toList())
                        .flatMapMany(epochs -> Flux.fromIterable(epochs)
                                .flatMap(epoch -> computorService.getQearnUserLockInfo(identity, epoch)
                                        .retry(NUM_RETRIES)
                                        .map(uli -> lockInfoForEpoch(uli, epoch))))
                        .sort()
                        .collectList())
                .map(infos -> """
                        
                        Qearn lock info for identity: %s
                        %s
                        """.formatted(identity, StringUtils.join(infos.toArray(), '\n')))
                .block();
    }

    public String getOverallStats() {
        return Mono.zip(computorService.getQearnStatsForEpoch(0).retry(NUM_RETRIES),
                computorService.getQearnBurnedAndBoostedStats().retry(NUM_RETRIES))
                .map(tuple -> """
                        
                        Qearn overall information
                        Total locked amount: %s
                        Rewarded amount: %s (%.3f%% avg/epoch)
                        Boosted amount: %s (%.3f%% avg/epoch)
                        Burned amount: %s (%.3f%% avg/epoch)
                        APY: %.3f%% avg/epoch
                        """.formatted(
                        formatUtil.formatUnsigned(tuple.getT1().getTotalLockedAmount()),
                        formatUtil.formatUnsigned(tuple.getT2().getRewardedAmount()), tuple.getT2().getAverageRewardedPercentAsDecimal(),
                        formatUtil.formatUnsigned(tuple.getT2().getBoostedAmount()), tuple.getT2().getAverageBoostedPercentAsDecimal(),
                        formatUtil.formatUnsigned(tuple.getT2().getBurnedAmount()), tuple.getT2().getAverageBurnedPercentAdDecimal(),
                        tuple.getT1().getAverageApyAsDecimal()
                        )
                )
                .block();

    }

    public String getEndedStatus(@NotBlank String identity) {
        return computorService.getQearnEndedStatus(identity)
                .retry(NUM_RETRIES)
                .map(status -> """
                        
                        Identity: %s
                        Early rewarded amount (current epoch): %s
                        Early unlocked amount (current epoch): %s
                        Fully rewarded amount (previous epoch): %s
                        Fully unlocked amount (previous epoch): %s
                        """.formatted(identity,
                        formatUtil.formatUnsigned(status.getEarlyRewardedAmount()),
                        formatUtil.formatUnsigned(status.getEarlyUnlockedAmount()),
                        formatUtil.formatUnsigned(status.getFullyRewardedAmount()),
                        formatUtil.formatUnsigned(status.getFullyUnlockedAmount())
                        )
                )
                .block();
    }

    private String lockInfoForEpoch(UserLockInfo lockInfo, int epoch) {
        return """
                Locked amount in epoch %d: %s""".formatted(epoch, formatUtil.formatUnsigned(lockInfo.getLockedAmount()));
    }
}
