package at.qubic.shell.component;

import at.qubic.api.crypto.IdentityUtil;
import at.qubic.api.domain.qearn.request.LockRequest;
import at.qubic.api.domain.qearn.request.UnlockRequest;
import at.qubic.api.domain.std.SignedTransaction;
import at.qubic.api.service.QearnTransactionService;
import at.qubic.shell.service.SendTransactionService;
import at.qubic.shell.component.terminal.QearnLockTerminalOutput;
import at.qubic.shell.component.terminal.QearnUnlockTerminalOutput;
import at.qubic.shell.component.terminal.TransactionTerminal;
import at.qubic.shell.exception.InputValidationException;
import at.qubic.shell.service.KeyManagementService;
import org.springframework.shell.standard.AbstractShellComponent;
import reactor.core.publisher.Mono;

import java.util.function.Function;

public class QearnLockComponent extends AbstractShellComponent {

    private final KeyManagementService keyManagementService;
    private final QearnTransactionService qearnTransactionService;
    private final SendTransactionService sendTransactionService;
    private final IdentityUtil identityUtil;

    public QearnLockComponent(KeyManagementService keyManagementService, QearnTransactionService qearnTransactionService, SendTransactionService sendTransactionService, IdentityUtil identityUtil) {
        this.keyManagementService = keyManagementService;
        this.qearnTransactionService = qearnTransactionService;
        this.sendTransactionService = sendTransactionService;
        this.identityUtil = identityUtil;
    }

    public Mono<?> unlock(String identity, int epoch, long amount, int offset, int retries, boolean validate) throws InputValidationException {
        validateInput(identity, amount, validate);
        byte[] nonce = keyManagementService.getPrivateKey(identity);
        UnlockRequest unlockRequest = UnlockRequest.builder().amount(amount).epoch(epoch).build();
        Function<Integer, SignedTransaction> transactionFunction = qearnTransactionService.createTransactionWithNonceK(nonce, identity, unlockRequest, offset);
        TransactionTerminal terminal = new QearnUnlockTerminalOutput(identityUtil, this::getTerminal, identity, epoch, amount);
        return sendTransactionService.sendTransaction(identity, transactionFunction, retries, validate, terminal);
    }

    public Mono<?> lock(String identity, long amount, int offset, int retries, boolean validate) throws InputValidationException {
        validateInput(identity, amount, validate);
        byte[] nonce = keyManagementService.getPrivateKey(identity);
        LockRequest lockRequest = LockRequest.builder().amount(amount).build();
        Function<Integer, SignedTransaction> transactionFunction = qearnTransactionService.createTransactionWithNonceK(nonce, identity, lockRequest, offset);
        TransactionTerminal terminal = new QearnLockTerminalOutput(identityUtil, this::getTerminal, identity, amount);
        return sendTransactionService.sendTransaction(identity, transactionFunction, retries, validate, terminal);
    }

    private void validateInput(String identity, long amount, boolean validate) throws InputValidationException {
        validateIdentity(identity);
        validateSenderKeys(identity);
        validateAmount(amount, validate);
    }

    private void validateIdentity(String identity) throws InputValidationException{
        if (!identityUtil.isValidIdentity(identity)) {
            throw new InputValidationException(String.format("Identity [%s] is invalid.", identity));
        }
    }

    private void validateSenderKeys(String identity) throws InputValidationException{
        if (keyManagementService.isLocked()) {
            throw new InputValidationException("Wallet needs to be unlocked.");
        } else if (!keyManagementService.containsKey(identity)) {
            throw new InputValidationException(String.format("There is no private key stored for identity [%s].", identity));
        }
    }

    private void validateAmount(long amount, boolean validate) throws InputValidationException {
        if (validate && amount < 10_000_000) {
            throw new InputValidationException("Amount must be at least 10'000'000 qu.");
        }
    }
}
