package at.qubic.shell.component;

import at.qubic.api.crypto.IdentityUtil;
import at.qubic.api.domain.std.SignedTransaction;
import at.qubic.api.service.TransactionService;
import at.qubic.shell.component.terminal.QuTransferTerminalOutput;
import at.qubic.shell.exception.InputValidationException;
import at.qubic.shell.service.KeyManagementService;
import at.qubic.shell.service.SendQuTransferService;
import org.springframework.shell.standard.AbstractShellComponent;
import reactor.core.publisher.Mono;

import java.util.function.Function;

public class QuTransferComponent extends AbstractShellComponent {

    private final KeyManagementService keyManagementService;
    private final TransactionService transactionService;
    private final SendQuTransferService sendQuTransferService;
    private final IdentityUtil identityUtil;

    public QuTransferComponent(KeyManagementService keyManagementService, TransactionService transactionService, SendQuTransferService sendQuTransferService, IdentityUtil identityUtil) {
        this.keyManagementService = keyManagementService;
        this.transactionService = transactionService;
        this.sendQuTransferService = sendQuTransferService;
        this.identityUtil = identityUtil;
    }

    public Mono<?> transfer(long amount, String from, String to, int offset, int retries, boolean validate) throws InputValidationException {
        validateIdentity(to);
        validateIdentity(from);
        validateSenderKeys(from);
        byte[] nonce = keyManagementService.getPrivateKey(from);
        Function<Integer, SignedTransaction> createTransactionFunction = tick -> transactionService.createTransactionWithNonceK(tick + offset, nonce, from, to, amount);
        QuTransferTerminalOutput terminal = new QuTransferTerminalOutput(identityUtil, this::getTerminal, from, to, amount);
        return sendQuTransferService.sendTransaction(from, createTransactionFunction, retries, validate, terminal);
    }

    private void validateIdentity(String identity) throws InputValidationException {
        if (!identityUtil.isValidIdentity(identity)) {
            throw new InputValidationException(String.format("Identity [%s] is invalid.", identity));
        }
    }

    private void validateSenderKeys(String identity) throws InputValidationException{
        if (keyManagementService.isLocked()) {
            throw new InputValidationException("Wallet needs to be unlocked.");
        } else if (!keyManagementService.containsKey(identity)) {
            throw new InputValidationException(String.format("There is no private key stored for identity [%s].", identity));
        }
    }

}
