package at.qubic.shell.component;

import at.qubic.api.crypto.IdentityUtil;
import at.qubic.api.domain.TransactionExtraData;
import at.qubic.api.domain.qx.request.QxAssetOrder;
import at.qubic.api.domain.qx.request.QxTransferAssetOrder;
import at.qubic.api.domain.std.SignedTransaction;
import at.qubic.api.service.QxTransactionService;
import at.qubic.shell.service.SendTransactionService;
import at.qubic.shell.component.terminal.QxOrderTerminalOutput;
import at.qubic.shell.component.terminal.QxTransferTerminalOutput;
import at.qubic.shell.component.terminal.TransactionTerminal;
import at.qubic.shell.exception.InputValidationException;
import at.qubic.shell.service.KeyManagementService;
import org.springframework.shell.standard.AbstractShellComponent;
import reactor.core.publisher.Mono;

import java.util.function.Function;

public class QxOrderComponent extends AbstractShellComponent {

    private final KeyManagementService keyManagementService;
    private final QxTransactionService transactionService;
    private final SendTransactionService sendTransactionService;
    private final IdentityUtil identityUtil;

    public QxOrderComponent(KeyManagementService keyManagementService, QxTransactionService transactionService, SendTransactionService sendTransactionService, IdentityUtil identityUtil) {
        this.keyManagementService = keyManagementService;
        this.transactionService = transactionService;
        this.sendTransactionService = sendTransactionService;
        this.identityUtil = identityUtil;
    }

    public Mono<?> sendQxOrder(String sender, QxAssetOrder qxOrder, int offset, int retries, boolean validate) throws InputValidationException {
        TransactionTerminal terminal = new QxOrderTerminalOutput(identityUtil, this::getTerminal, sender, qxOrder);
        return sendTransaction(sender, qxOrder, offset, retries, validate, terminal);
    }

    public Mono<?> sendQxTransfer(String sender, QxTransferAssetOrder transfer, int offset, int retries, boolean validate) throws InputValidationException {
        TransactionTerminal terminal = new QxTransferTerminalOutput(identityUtil, this::getTerminal, sender, transfer);
        return sendTransaction(sender, transfer, offset, retries, validate, terminal);
    }

    private Mono<TransactionContext> sendTransaction(String sender, TransactionExtraData qxOrder, int offset, int retries, boolean validate, TransactionTerminal terminal) throws InputValidationException {
        validateInput(sender);
        byte[] nonce = keyManagementService.getPrivateKey(sender);
        Function<Integer, SignedTransaction> transactionFunction = transactionService.createTransactionFunctionWithNonceK(nonce, sender, qxOrder, offset);
        return sendTransactionService.sendTransaction(sender, transactionFunction, retries, validate, terminal);
    }

    private void validateInput(String identity) throws InputValidationException {
        validateIdentity(identity);
        validateSenderKeys(identity);
    }

    public void validateIdentity(String identity) throws InputValidationException {
        if (!identityUtil.isValidIdentity(identity)) {
            throw new InputValidationException(String.format("Identity [%s] is invalid.", identity));
        }
    }

    private void validateSenderKeys(String identity) throws InputValidationException{
        if (keyManagementService.isLocked()) {
            throw new InputValidationException("Wallet needs to be unlocked.");
        } else if (!keyManagementService.containsKey(identity)) {
            throw new InputValidationException(String.format("There is no private key stored for identity [%s].", identity));
        }
    }


}
