package at.qubic.shell.component.terminal;

import at.qubic.api.crypto.IdentityUtil;
import at.qubic.api.domain.std.SignedTransaction;
import at.qubic.api.domain.std.Transaction;
import at.qubic.shell.component.TransactionContext;
import at.qubic.shell.exception.PostProcessingException;
import at.qubic.shell.util.FormatUtil;
import lombok.extern.slf4j.Slf4j;
import reactor.core.publisher.Mono;

@Slf4j
public abstract class AbstractTransactionTerminal<T extends TransactionContext> implements TransactionTerminal<T> {

    public static final String CHA = "\u001B[0G"; // ESC [ <n> G 	CHA

    protected final FormatUtil formatUtil = new FormatUtil();
    protected final TerminalWriter terminalWriter;
    protected final IdentityUtil identityUtil;

    protected AbstractTransactionTerminal(IdentityUtil identityUtil, TerminalWriter terminalWriter) {
        this.terminalWriter = terminalWriter;
        this.identityUtil = identityUtil;
    }

    @Override
    public TerminalWriter getTerminalWriter() {
        return terminalWriter;
    }

    @Override
    public void writeWaitForTickMessage(SignedTransaction signedTransaction) {
        terminalWriter.writeAsync("""
            Sent transaction [%s]. Please wait for tick [%d].
            """.formatted(signedTransaction.getTransactionHash(), signedTransaction.getTransaction().getTick()));
    }

    @Override
    public Mono<?> writeTransactionProgress(int startTick, int targetTick, int currentTick) {
        return terminalWriter.writePublisher(String.format("Current tick: [%s]%s", currentTick, CHA));
    }

    @Override
    public void writeErrorMessage(PostProcessingException e) {
        TransactionContext context = e.getContext();
        String msg = String.format("Transaction post processing failed. Context: %s.", context);
        log.error(msg, e);
        String txInformation = getTransactionInformationString(context.getSignedTransaction());
        String output = """
            Transaction included but could not verify success because post processing failed.
            Please check the transaction on the explorer.
            
            %s
            """.formatted(txInformation);
        terminalWriter.writeAsync(output);
    }

    protected String getTransactionInformationString(SignedTransaction signedTransaction) {
        Transaction transaction = signedTransaction.getTransaction();
        return """
                Source: %s
                Destination: %s
                Hash: %s
                Amount: %s
                Type: %d
                Tick: %d
                
                https://explorer.qubic.org/network/tx/%s""".formatted(
                identityUtil.getIdentityFromPublicKey(transaction.getSourcePublicKey()),
                identityUtil.getIdentityFromPublicKey(transaction.getDestinationPublicKey()),
                signedTransaction.getTransactionHash(),
                formatUtil.formatUnsigned(transaction.getAmount()),
                transaction.getInputType(),
                transaction.getTick(),
                signedTransaction.getTransactionHash()
        );
    }



}
