package at.qubic.shell.component.terminal;

import lombok.extern.slf4j.Slf4j;
import at.qubic.shell.exception.*;
import reactor.util.retry.Retry;
import reactor.util.retry.RetrySpec;

@Slf4j
public class DefaultTerminalErrorHandler {

    private final TerminalWriter terminal;

    public DefaultTerminalErrorHandler(TerminalWriter terminal) {
        this.terminal = terminal;
    }

    public RetrySpec retryIfRetryable(int maxRetries) {
        return Retry.max(maxRetries)
                .filter(t -> t instanceof SendTransactionException && ((SendTransactionException) t).isRetryable())
                .doBeforeRetryAsync(rs -> terminal.writePublisher(String.format("\nRetry: attempt [%d] of [%d].\n", (rs.totalRetries() + 1), maxRetries)).then())
                .doBeforeRetry(rs -> log.info("Retry [{}] of [{}].", rs.totalRetries(), maxRetries))
                .onRetryExhaustedThrow((retrySpec, retrySignal) -> new SendTransactionException("Transaction failed. Retries exhausted.", retrySignal.failure()));
    }

    public void outputErrorMessage(BeforSendingException e) {
        String message = String.format("Error before sending transaction. Transaction was not sent. Error message: %s", e.getMessage());
        log.error(message, e);
        terminal.writeAsync(message);
    }

    public void outputErrorMessage(RetryableSendTransactionException e) {
        String message = String.format("Error trying to send transaction. Transaction was not sent. Error message: %s", e.getMessage());
        log.error(message, e);
        terminal.writeAsync(message);
    }

    public void outputErrorMessage(NotIncludedException e) {
        log.warn(e.getMessage());
        terminal.writeAsync(e.getMessage());
    }

    public void outputErrorMessage(AfterSendingException ase) {
        String message = String.format("Error sending transaction. Transaction status unclear. Please check on explorer. Error message: %s", ase.getCause() == null ? ase.getMessage() : ase.getCause().getMessage());
        log.error(message, ase);
        terminal.writeAsync(message);
    }

    public RetryableSendTransactionException wrapIntoRetryableException(Throwable throwable) {
        log.error("Error before sending transaction: [{}].", throwable.getMessage());
        return new RetryableSendTransactionException(throwable.getMessage(), throwable);
    }

    public AfterSendingException wrapIntoAfterSendingException(Throwable e) {
        String message = "Error after sending transaction.";
        log.error(message);
        return new AfterSendingException(message, e);
    }

}