package at.qubic.shell.component.terminal;

import at.qubic.api.crypto.IdentityUtil;
import at.qubic.api.domain.std.SignedTransaction;
import at.qubic.api.domain.std.response.Entity;
import at.qubic.shell.component.QuTransferContext;
import lombok.extern.slf4j.Slf4j;
import org.jline.terminal.Terminal;

import java.util.function.Supplier;

@Slf4j
public class QuTransferTerminalOutput extends AbstractTransactionTerminal<QuTransferContext> {

    private final String sourceIdentity;
    private final String destinationIdentity;
    private final Long amount;

    public QuTransferTerminalOutput(IdentityUtil identityUtil, Supplier<Terminal> terminalSupplier, String sourceIdentity, String destinationIdentity, Long amount) {
        super(identityUtil, new TerminalWriter(terminalSupplier));
        this.sourceIdentity = sourceIdentity;
        this.destinationIdentity = destinationIdentity;
        this.amount = amount;
    }

    @Override
    public void writeOperationDescription() {
        terminalWriter.writeAsync(String.format("Send [%s] from [%s] to [%s].\n", amount, sourceIdentity, destinationIdentity));
    }

    @Override
    public void writeSuccessMessage(QuTransferContext tc) {
        String message = getResultMessage(tc);
        terminalWriter.writeAsync(message);
    }

    private String getResultMessage(QuTransferContext tc) {
        Entity sourceBefore = tc.getSourceBefore();
        Entity destinationBefore = tc.getDestinationBefore();
        SignedTransaction stx = tc.getSignedTransaction();
        Entity sourceAfter = tc.getSourceAfter();
        Entity destinationAfter = tc.getDestinationAfter();
        long outgoingAmount = sourceAfter.getOutgoingAmount() - sourceBefore.getOutgoingAmount();
        long incomingAmount = destinationAfter.getIncomingAmount() - destinationBefore.getIncomingAmount();

        String transactionInformation = getTransactionInformationString(stx);

        String balanceInformation = """
                    Outgoing amount (source at tick %d): %s
                    Latest outgoing tick: %d
                    Incoming amount (destination at tick %d): %s
                    Latest incoming tick: %d""".formatted(
                sourceAfter.getTick(),
                formatUtil.formatUnsigned(outgoingAmount),
                sourceAfter.getLatestOutgoingTransferTick(),
                destinationAfter.getTick(),
                formatUtil.formatUnsigned(incomingAmount),
                destinationAfter.getLatestIncomingTransferTick());

        long expectedAmount = stx.getTransaction().getAmount();
        if (outgoingAmount >= expectedAmount && incomingAmount >= expectedAmount) {
            log.info("Transaction [{}] successful.", stx.getTransactionHash());
            // \s to clear previous text
            String message = """
                Transaction successful. \s
                
                %s
                """.formatted(transactionInformation);

            if (outgoingAmount > expectedAmount || incomingAmount > expectedAmount) {
                message = """
                        %s
                        
                        Unexpected source and/or destination amount after transaction:
                        %s
                        """.formatted(message, balanceInformation);
            }
            return message;

        } else if (outgoingAmount == 0) {
            return """
                Transaction included but NOT successful as no funds were moved.
                
                %s
                """.formatted(transactionInformation);
        } else {
            // Balances did change but not enough. Transaction failed. This should never happen.
            String message = """
            Transaction included but NOT successful as balances didn't change as expected.
            
            %s
            
            Unexpected source and/or destination amount after transaction:
            %s
            """.formatted(transactionInformation, balanceInformation);
            log.error(message);
            return message;
        }
    }


}
