package at.qubic.shell.configuration;

import lombok.extern.slf4j.Slf4j;
import at.qubic.api.crypto.IdentityUtil;
import at.qubic.api.crypto.QubicFourQ;
import at.qubic.api.crypto.SchnorrQubic;
import at.qubic.api.network.*;
import at.qubic.api.properties.ComputorProperties;
import at.qubic.api.properties.NetworkProperties;
import at.qubic.api.service.ComputorService;
import at.qubic.api.service.QxTransactionService;
import at.qubic.api.service.TransactionService;
import org.springframework.boot.context.properties.ConfigurationProperties;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.core.env.Environment;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;
import reactor.core.scheduler.Schedulers;

import java.util.Arrays;
import java.util.List;

@Slf4j
@Configuration
public class QubicApiConfiguration {

    @Bean
    @ConfigurationProperties("computor")
    ComputorProperties computorProperties() {
        return new ComputorProperties();
    }

    @Bean
    @ConfigurationProperties(value = "network", ignoreUnknownFields = false)
    NetworkProperties networkProperties() {
        return new NetworkProperties();
    }

    @Bean
    NetworkStatus networkStatus() {
        return new HealthCheckNetworkStatus();
    }

    @Bean
    NodesManagementStrategy nodesManagementStrategy(NetworkProperties properties, NetworkStatus networkStatus, Environment environment) {
        boolean usePublicPeers = environment.getRequiredProperty("use-public-peers", Boolean.class);
        log.info("Using port [{}].", properties.getPort());
        return usePublicPeers
                ? new PublicNodesStrategy(networkStatus, properties)
                : new FixNodesStrategy(networkStatus, properties);
    }

    @Bean
    Nodes nodes(NodesManagementStrategy nodesManagementStrategy, ComputorProperties properties) {
        Nodes nodes = new Nodes(nodesManagementStrategy);
        List<String> hosts = Arrays.asList(properties.getHosts());

        log.info("Using computors {}.", hosts);
        Flux.fromIterable(hosts)
                .parallel()
                .runOn(Schedulers.parallel())
                .flatMap(h -> Mono.just(nodes.addNode(h)))
                .sequential()
                .subscribe(n -> log.debug("Added configured node: {}", n.getName()),
                        t -> log.error("Error while adding node: {}", t.getMessage()),
                        () -> log.info("Completed adding configured nodes.")
                );
        return nodes;
    }

    @Bean
    QubicFourQ crypto(Environment env) {
        return new QubicFourQ(env.getRequiredProperty("fourq.lib.name"));
    }

    @Bean
    IdentityUtil identityUtil(QubicFourQ fourQ) {
        return new IdentityUtil(true, fourQ);
    }

    @Bean
    TransactionService transactionService(IdentityUtil identityUtil, SchnorrQubic signing) {
        return new TransactionService(identityUtil, signing);
    }

    @Bean
    QxTransactionService qxTransactionService(TransactionService transactionService) {
        return new QxTransactionService(transactionService);
    }

    @Bean
    ComputorService qubicComputorService(Nodes nodes, IdentityUtil identityUtil, TransactionService transactionService) {
        return new ComputorService(nodes, identityUtil, transactionService);
    }

}
