package at.qubic.shell.file;

import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.StringUtils;

import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.attribute.PosixFilePermission;
import java.nio.file.attribute.PosixFilePermissions;
import java.util.List;
import java.util.Set;

@Slf4j
public class FileUtility {

    public List<String> load(Path filePath) throws IOException {
        if (Files.exists(filePath)) {
            return FileUtils.readLines(filePath.toFile(), StandardCharsets.UTF_8)
                    .stream()
                    .filter(StringUtils::isNotBlank)
                    .filter(l -> !StringUtils.startsWithAny(l, "//", "#"))
                    .toList();
        } else {
            return List.of();
        }
    }

    public byte[] loadBinary(Path filePath) throws IOException {
        if (Files.exists(filePath)) {
            return FileUtils.readFileToByteArray(filePath.toFile());
        } else {
            return new byte[0];
        }
    }

    public void save(Path filePath, byte[] binary) throws IOException {
        createDirectory(filePath.getParent());
        FileUtils.writeByteArrayToFile(filePath.toFile(), binary);
        tryToSetPermissions(filePath);
    }

    public void save(Path filePath, List<String> lines) throws IOException {
        createDirectory(filePath.getParent());
        FileUtils.writeLines(filePath.toFile(), StandardCharsets.UTF_8.name(), lines);
        tryToSetPermissions(filePath);
    }

    public void move(Path source, Path target) throws IOException {
        FileUtils.moveFile(source.toFile(), target.toFile());
    }

    public void delete(Path path) throws IOException {
        FileUtils.delete(path.toFile());
    }

    @SneakyThrows
    public void createDirectory(Path directory) {
        FileUtils.forceMkdir(new File(directory.toString()));
    }

    public void tryToSetPermissions(Path filePath) {
        try {
            Set<PosixFilePermission> permissions = PosixFilePermissions.fromString("rw-------");
            Files.setPosixFilePermissions(filePath, permissions);
        } catch (Exception e) { // missing permissions or unsupported by os
            log.warn("Error setting permissions for [{}]. Ignoring.", filePath.toAbsolutePath(), e);
        }
    }

}
