package at.qubic.shell.repository;

import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.ArrayUtils;
import at.qubic.shell.exception.KeyManagementException;
import at.qubic.shell.file.FileUtility;
import org.springframework.security.crypto.keygen.KeyGenerators;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.CertificateException;

@Slf4j
public class KeyStoreRepository {

    public static final String BACKUP_EXTENSION = ".bak";
    private final FileUtility fileUtility;
    private final String applicationDirectoryPath;
    private final String keystoreFile;
    private final String saltFile;

    public KeyStoreRepository(FileUtility fileUtility, String applicationDirectoryPath, String keystoreFile, String saltFile) {
        this.fileUtility = fileUtility;
        this.applicationDirectoryPath = applicationDirectoryPath;
        this.keystoreFile = keystoreFile;
        this.saltFile = saltFile;
    }

    public KeyStore load(KeyStore keyStore, char[] password) {
        final Path path = Paths.get(applicationDirectoryPath, keystoreFile);
        log.info("Loading keystore from [{}].", path);
        if (Files.exists(path)) {
            loadKeystore(keyStore, password, path);
        } else {
            createNewKeystore(keyStore, password);
        }
        return keyStore;
    }

    public byte[] loadPasswordSalt() {
        final Path path = Paths.get(applicationDirectoryPath, saltFile);
        byte[] initialSalt;
        try {
            initialSalt = fileUtility.loadBinary(path);
            if (ArrayUtils.isEmpty(initialSalt)) {
                log.info("Creating initial salt...");
                initialSalt = KeyGenerators.secureRandom(16).generateKey();
                fileUtility.save(path, initialSalt);
            }
        } catch (IOException e) {
            String message = "Error loading password salt.";
            log.error("{} {}", message, e.getMessage());
            throw new KeyManagementException(message, e);
        }
        return initialSalt;
    }

    public Path moveSaltFile() {
        final Path source = Paths.get(applicationDirectoryPath, saltFile);
        final Path target = Paths.get(applicationDirectoryPath, saltFile + BACKUP_EXTENSION);
        return move(source, target);
    }

    public Path moveKeystore() {
        final Path source = Paths.get(applicationDirectoryPath, keystoreFile);
        final Path target = Paths.get(applicationDirectoryPath, keystoreFile + BACKUP_EXTENSION);
        return move(source, target);
    }

    public void delete(Path path) {
        try {
            log.info("Delete file [{}].", path);
            fileUtility.delete(path);
        } catch (IOException e) {
            String message = String.format("Error deleting [%s].", path);
            log.error("{} {}", message, e.getMessage());
            throw new KeyManagementException(message, e);
        }
    }

    private Path move(Path source, Path target) {
        try {
            log.info("Move file from [{}] to [{}].", source, target);
            fileUtility.move(source, target);
            return target;
        } catch (IOException e) {
            String message = "Error backing up salt file.";
            log.error("{} {}", message, e.getMessage());
            throw new KeyManagementException(message, e);
        }
    }

    public void save(KeyStore keystore, char[] password) {
        final Path path = Paths.get(applicationDirectoryPath, keystoreFile);
        fileUtility.createDirectory(path.getParent());
        log.info("Saving keystore to [{}].", path);
        saveKeystore(keystore, password, path);
        fileUtility.tryToSetPermissions(path);
    }

    private static void loadKeystore(KeyStore keyStore, char[] password, Path path) {
        try (FileInputStream fis = new FileInputStream(path.toFile())) {
            keyStore.load(fis, password);
        } catch (IOException | CertificateException | NoSuchAlgorithmException e) {
            String msg = String.format("Error loading keystore: %s", e.getMessage());
            log.error(msg);
            throw new KeyManagementException(msg, e);
        }
    }

    private static void createNewKeystore(KeyStore keyStore, char[] password) {
        try {
            log.info("Creating new keystore.");
            keyStore.load(null, password);
        } catch (IOException | CertificateException | NoSuchAlgorithmException e) {
            String msg = String.format("Error creating new keystore: %s", e.getMessage());
            log.error(msg);
            throw new KeyManagementException(msg, e);
        }
    }

    private static void saveKeystore(KeyStore keystore, char[] password, Path path) {
        try (FileOutputStream fos = new FileOutputStream(path.toFile())) {
            keystore.store(fos, password);
        } catch (IOException | CertificateException | KeyStoreException | NoSuchAlgorithmException e) {
            String msg = String.format("Error trying to save keystore: %s", e.getMessage());
            log.warn(msg);
            throw new KeyManagementException(msg, e);
        }
    }

}
