package at.qubic.shell.service;

import at.qubic.api.domain.std.SignedTransaction;
import at.qubic.shell.component.InclusionCheckComponent;
import at.qubic.shell.component.TransactionContext;
import at.qubic.shell.component.terminal.DefaultTerminalErrorHandler;
import at.qubic.shell.component.terminal.TransactionTerminal;
import at.qubic.shell.exception.*;
import lombok.extern.slf4j.Slf4j;
import reactor.core.publisher.Mono;

import java.util.function.Function;

@Slf4j
public abstract class AbstractSendTransactionService<T extends TransactionContext> {

    protected final ShellComputorService computorService;
    protected final InclusionCheckComponent inclusionCheck;

    public AbstractSendTransactionService(ShellComputorService computorService,
                                          InclusionCheckComponent inclusionCheck) {
        this.computorService = computorService;
        this.inclusionCheck = inclusionCheck;
    }

    public Mono<T> sendTransaction(String senderIdentity,
                                                    Function<Integer, SignedTransaction> transactionFunction,
                                                    int retries, boolean validate,
                                                    TransactionTerminal<T> terminal) {

        DefaultTerminalErrorHandler errorHandler = new DefaultTerminalErrorHandler(terminal.getTerminalWriter());

        return Mono.just(senderIdentity)
                .doOnNext(id -> terminal.writeOperationDescription())
                .then(computorService.updateNetworkIfNecessary())
                .then(preProcessTransaction(senderIdentity, transactionFunction, validate)
                        .onErrorMap(e -> !(e instanceof SendTransactionException), errorHandler::wrapIntoRetryableException) // retry by default
                        .flatMap(tc -> computorService.sendTransaction(tc.getSignedTransaction()).map(x -> tc)) // send transaction
                        .doOnNext(tc -> terminal.writeWaitForTickMessage(tc.getSignedTransaction()))
                        .flatMap(tc -> checkTxIncluded(terminal, tc)) // wait for inclusion
                        .flatMap(this::postProcessTransaction) // get entity information
                        .doOnNext(terminal::writeSuccessMessage)

                        .doOnError(RetryableSendTransactionException.class, errorHandler::outputErrorMessage) // retry
                        .doOnError(NotIncludedException.class, errorHandler::outputErrorMessage) // retry
                        .retryWhen(errorHandler.retryIfRetryable(retries))

                        .doOnError(BeforSendingException.class, errorHandler::outputErrorMessage) // unrecoverable, abort
                        .doOnError(PostProcessingException.class, terminal::writeErrorMessage)
                        .onErrorMap(e -> !(e instanceof SendTransactionException), errorHandler::wrapIntoAfterSendingException)
                        .doOnError(AfterSendingException.class, errorHandler::outputErrorMessage) // unrecoverable and unexpected

                );
    }

    protected abstract Mono<T> preProcessTransaction(String senderIdentity,
                                                           Function<Integer, SignedTransaction> createTransactionFunction,
                                                           boolean validate);

    protected abstract  Mono<T> postProcessTransaction(T transactionContext);

    private Mono<T> checkTxIncluded(TransactionTerminal<T> terminal, T context) {
        return inclusionCheck.waitForInclusion(terminal,
                        context.getSourceBefore().getTick(),
                        context.getSignedTransaction().getTransaction().getTick(),
                        context.getSignedTransaction().getTransactionHash())
                .map(s -> context);
    }

}
