package at.qubic.shell.service;

import at.qubic.api.crypto.IdentityUtil;
import at.qubic.api.domain.std.SignedTransaction;
import at.qubic.api.domain.std.response.Entity;
import at.qubic.shell.component.InclusionCheckComponent;
import at.qubic.shell.component.QuTransferContext;
import at.qubic.shell.exception.BeforSendingException;
import at.qubic.shell.exception.PostProcessingException;
import lombok.extern.slf4j.Slf4j;
import reactor.core.publisher.Mono;

import java.util.function.Function;

@Slf4j
public class SendQuTransferService extends AbstractSendTransactionService<QuTransferContext> {

    private final IdentityUtil identityUtil;

    public SendQuTransferService(ShellComputorService computorService,
                                 InclusionCheckComponent inclusionCheck, IdentityUtil identityUtil) {
        super(computorService, inclusionCheck);
        this.identityUtil = identityUtil;
    }

    protected Mono<QuTransferContext> preProcessTransaction(String senderIdentity,
                                                           Function<Integer, SignedTransaction> createTransactionFunction,
                                                           boolean validate) {

        return computorService.getEntitiesSameTick(identityUtil.getPublicKeyFromIdentity(senderIdentity),
                        createTransactionFunction.apply(0).getTransaction().getDestinationPublicKey()) // get from transaction function
                .map(tuple -> {
                    Entity source = tuple.getT1();
                    Entity destination = tuple.getT2();
                    SignedTransaction stx = createTransactionFunction.apply(source.getTick());
                    if (validate) {
                        verifyBalance(senderIdentity, source, stx.getTransaction().getAmount());
                    }
                    return QuTransferContext.builder()
                            .sourceBefore(source)
                            .destinationBefore(destination)
                            .signedTransaction(stx)
                            .build();

                });
    }

    protected Mono<QuTransferContext> postProcessTransaction(QuTransferContext context) {
        return computorService.getEntitiesMinimumTick(
                        context.getSourceBefore().getPublicKey(),
                        context.getDestinationBefore().getPublicKey(),
                        context.getSignedTransaction().getTransaction().getTick())
                .map(t2 -> {
                    context.setSourceAfter(t2.getT1());
                    context.setDestinationAfter(t2.getT2());
                    return context;
                })
                .onErrorMap(e -> new PostProcessingException(context, "Error post processing transaction.", e)); // not fatal. Transaction included but couldn't fetch post transaction entity information
    }

    private void verifyBalance(String identity, Entity entity, long requiredBalance) {
        long balance = entity.getIncomingAmount() - entity.getOutgoingAmount();
        if (balance < requiredBalance) {
            String message = String.format("Insufficient balance on [%s]: [%d]. Required: [%d].",
                    identity, balance, requiredBalance);
            log.warn(message);
            throw new BeforSendingException(message);
        }
    }

}
