package at.qubic.shell.service;

import at.qubic.api.crypto.IdentityUtil;
import at.qubic.api.domain.std.SignedTransaction;
import at.qubic.api.domain.std.response.Entity;
import at.qubic.api.service.ComputorService;
import at.qubic.shell.exception.SendTransactionException;
import lombok.extern.slf4j.Slf4j;
import reactor.core.publisher.Mono;
import reactor.util.function.Tuple2;

import java.time.Instant;

@Slf4j
public class ShellComputorService {

    private static final int ALLOWED_NETWORK_DELAY_SECONDS = 60;
    private static final int NUM_RETRIES = 3;

    private final ComputorService computorService;
    private final IdentityUtil identityUtil;

    public ShellComputorService(ComputorService computorService, IdentityUtil identityUtil) {
        this.computorService = computorService;
        this.identityUtil = identityUtil;
    }

    public Mono<Entity> getEntity(String identity) {
        return computorService.getEntity(identity)
                .retry(NUM_RETRIES);
    }

    public Mono<Entity> getEntity(byte[] publicKey) {
        return getEntity(publicKey, 0);
    }

    public Mono<Entity> getEntity(byte[] publicKey, int minTick) {
        return computorService.getEntity(publicKey)
                .filter(entity -> entity.getTick() >= minTick)
                .switchIfEmpty(Mono.error(new IllegalStateException("Could not retrieve entity in required tick " + minTick)))
                .retry(NUM_RETRIES);
    }

    public Mono<Tuple2<Entity, Entity>> getEntitiesSameTick(byte[] one, byte[] two) {
        return getEntities(one, two, true, 0);
    }

    public Mono<Tuple2<Entity, Entity>> getEntitiesMinimumTick(byte[] one, byte[] two, int minTick) {
        return getEntities(one, two, false, minTick);
    }

    private Mono<Tuple2<Entity, Entity>> getEntities(byte[] one, byte[] two, boolean sameTick, int minTick) {
        return Mono.defer(() -> Mono.zip(getEntity(one), getEntity(two)))
                .doOnNext(t2 -> { if (minTick > 0) verifyMinimumTick(t2.getT1(), minTick); verifyMinimumTick(t2.getT2(), minTick); })
                .doOnNext(t2 -> { if (sameTick) verifyInSameTick(t2); })
                .doOnError(e -> log.warn("Error getting entities: [{}]", e.getMessage()))
                .retry(NUM_RETRIES);
    }

    private void verifyMinimumTick(Entity entity, int minTick) {
        if (entity.getTick() < minTick) {
            String identity = identityUtil.getIdentityFromPublicKey(entity.getPublicKey());
            log.warn("Entity [{}] is in old tick [{}]. Minimum required: [{}].", identity, minTick, entity.getTick());
            throw new IllegalStateException("Entity [" + identity + "] information is outdated.");
        }
    }

    private void verifyInSameTick(Tuple2<Entity, Entity> twoEntities) {
        if (twoEntities.getT1().getTick() == twoEntities.getT2().getTick()) {
            log.debug("Both entities are in same tick [{}].",
                    twoEntities.getT1().getTick());
        } else {
            String source = identityUtil.getIdentityFromPublicKey(twoEntities.getT1().getPublicKey());
            long sourceTick = Integer.toUnsignedLong(twoEntities.getT1().getTick());
            String destination = identityUtil.getIdentityFromPublicKey(twoEntities.getT2().getPublicKey());
            long destinationTick = Integer.toUnsignedLong(twoEntities.getT2().getTick());
            log.warn("Entities in different ticks. [{}]: [{}]. [{}]: [{}].", source, sourceTick, destination, destinationTick);
            throw new IllegalStateException("Entities are not in same tick.");
        }
    }

    public Mono<Instant> updateNetworkIfNecessary() {
        return Mono.defer(() -> computorService
                .forceNetworkUpdateIfBefore(Instant.now().minusSeconds(ALLOWED_NETWORK_DELAY_SECONDS))
                .doOnNext(updatedAt -> log.debug("Network last updated at {}", updatedAt))
        );
    }

    public Mono<SignedTransaction> sendTransaction(SignedTransaction signedTransaction) {
        return Mono.zipDelayError(sendTransactionReturnStatus(signedTransaction),
                        sendTransactionReturnStatus(signedTransaction),
                        sendTransactionReturnStatus(signedTransaction))
                .handle((tuples, sink) -> {
                    if (tuples.getT1() + tuples.getT2() + tuples.getT3() > 0) {
                        sink.next(signedTransaction);
                    } else {
                        sink.error(new SendTransactionException("Send transaction failed"));
                    }
                });
    }

    private Mono<Integer> sendTransactionReturnStatus(SignedTransaction signedTransaction) {
        return sendTransactionViaComputorService(signedTransaction)
                .map(tx -> 1)
                .switchIfEmpty(Mono.just(0))
                .onErrorResume(t -> Mono.just(0));
    }

    private Mono<SignedTransaction> sendTransactionViaComputorService(SignedTransaction transaction) {
        return Mono.defer(() -> computorService.sendTransaction(transaction)) // TODO we should send to several nodes
                .retry(NUM_RETRIES);
    }

}
