package at.qubic.shell.wallet;

import jakarta.annotation.PreDestroy;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Pair;
import at.qubic.shell.exception.KeyManagementException;
import at.qubic.shell.repository.KeyStoreRepository;
import at.qubic.shell.util.ConversionUtil;
import at.qubic.shell.util.EncryptionUtility;
import org.springframework.security.crypto.keygen.KeyGenerators;
import org.springframework.util.Assert;

import javax.crypto.SecretKey;
import javax.crypto.spec.PBEParameterSpec;
import javax.crypto.spec.SecretKeySpec;
import java.nio.CharBuffer;
import java.nio.file.Path;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableEntryException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.List;

@Slf4j
public class KeyManager {

    static final String ALIAS_SALT = "salt";
    static final String INFO_KEYSTORE = "keystore";
    static final String INFO_SALT = "salt";
    static final String INFO_ENTRY = "entry";
    static final String INFO_ENCRYPTION = "encryption";
    public static final int ITERATION_COUNT_PBE = 1_000_000; // good enough because we encrypt data
    private final KeyStoreRepository keyStoreRepository;
    private final EncryptionUtility encryptionUtility;

    private char[] storePassword; // password for keystore
    private char[] saltPassword; // password for salt retrieval
    private char[] entryPassword; // password for entries
    private char[] encryptionPassword; // password for encryption
    private byte[] salt; // salt for entries and encryption

    private KeyStore keyStore;

    public KeyManager(KeyStoreRepository keyStoreRepository) {
        this(keyStoreRepository, new EncryptionUtility());
    }

    KeyManager(KeyStoreRepository keyStoreRepository, EncryptionUtility encryptionUtility) {
        this.keyStoreRepository = keyStoreRepository;
        this.encryptionUtility = encryptionUtility;
    }

    @PreDestroy
    void shutdown() {
        if (!isLocked()) {
            lock();
        }
    }

    public void unlock(char[] pass) {
        Assert.isTrue(!ArrayUtils.isEmpty(pass) && StringUtils.isNotBlank(CharBuffer.wrap(pass)),
                "Password cannot be null or blank.");
        if (isLocked()) {
            // initial salt for checking keystore integrity and storing salt for entries and encryption
            byte[] initialSalt = keyStoreRepository.loadPasswordSalt();
            // hash password once and use for deriving keys
            byte[] hashedPassword = encryptionUtility.hashPassword(ConversionUtil.charsToBytes(pass), initialSalt);
            this.storePassword = deriveKey(hashedPassword, initialSalt, INFO_KEYSTORE);
            this.saltPassword = deriveKey(hashedPassword, initialSalt, INFO_SALT);
            try {
                KeyStore ks = KeyStore.getInstance(KeyStore.getDefaultType());
                this.keyStore = keyStoreRepository.load(ks, storePassword);
                this.salt = getSalt(keyStore);
                this.entryPassword = deriveKey(hashedPassword, salt, INFO_ENTRY);
                this.encryptionPassword = deriveKey(hashedPassword, salt, INFO_ENCRYPTION);
            } catch (KeyStoreException e) {
                String message = String.format("Error creating keystore instance: %s", e.getMessage());
                log.error(message);
                throw new KeyManagementException(message, e);
            }
        } else {
            throw new IllegalStateException("Already unlocked.");
        }
    }

    public void updatePassword(char[] newPass) {
        Assert.state(!isLocked(), "KeyManager is locked");
        List<Pair<String, byte[]>> allKeyEntries = getAllKeys();
        this.lock(); // save because we would like to have a complete backup
        // move away old files
        Path oldKeystoreFile = keyStoreRepository.moveKeystore();
        Path oldSaltFile = keyStoreRepository.moveSaltFile();
        this.unlock(newPass);
        // restore entries
        for (Pair<String, byte[]> keyEntry : allKeyEntries) {
            setPrivateKey(keyEntry.getLeft(), keyEntry.getRight());
        }
        this.saveKeyStore(); // save new keystore
        // all good, clean up
        keyStoreRepository.delete(oldKeystoreFile);
        keyStoreRepository.delete(oldSaltFile);
    }

    public List<String> getKeyAliases() {
        try {
            List<String> aliases = new ArrayList<>();
            for (Enumeration<String> aliasEnum = keyStore.aliases(); aliasEnum.hasMoreElements();) {
                String alias = aliasEnum.nextElement();
                if (!StringUtils.equalsIgnoreCase(ALIAS_SALT, alias)) { // leave salt alone
                    aliases.add(alias);
                }
            }
            return aliases;
        } catch (KeyStoreException e) {
            String message = String.format("Error accessing keystore: %s", e.getMessage());
            log.error(message);
            throw new KeyManagementException(message, e);
        }
    }

    private List<Pair<String, byte[]>> getAllKeys() {
            return getKeyAliases().stream().map(a -> Pair.of(a, getPrivateKey(a))).toList();
    }

    private byte[] getSalt(KeyStore keyStore) throws KeyStoreException {
        // salt is not encrypted because encryption password needs this salt first
        if (!keyStore.containsAlias(ALIAS_SALT)) {
            byte[] newSalt = encryptionUtility.generateSalt();
            putKey(keyStore, ALIAS_SALT, createSecretKey(newSalt), saltPassword);
        }
        return getPrivateKey(keyStore, ALIAS_SALT, saltPassword);
    }

    public void setPrivateKey(String alias, byte[] data) {
        if (isLocked()) {
            throw new IllegalStateException("Keystore is locked.");
        } else {
            if (containsKeyAlias(alias)) {
                log.warn("Keystore already contains alias [{}].", alias);
            }
            log.info("Adding private key for [{}]", alias);
            byte[] encrypted = encryptionUtility.encryptData(encryptionPassword, salt, data);
            SecretKey secretKey = createSecretKey(encrypted);
            putKey(keyStore, alias, secretKey, entryPassword);
        }
    }

    public byte[] getPrivateKey(String alias) {
        byte[] data = null;
        if (containsKeyAlias(alias)) {
            log.info("Getting private key for [{}]", alias);
            byte[] storedKey = getPrivateKey(keyStore, alias, entryPassword);
            data = encryptionUtility.decryptData(encryptionPassword, salt, storedKey);
        }
        return data;
    }

    public void lock() {
        saveKeyStore();
        this.keyStore = null;
        storePassword = clearAndNullify(storePassword);
        saltPassword = clearAndNullify(saltPassword);
        entryPassword = clearAndNullify(entryPassword);
        encryptionPassword = clearAndNullify(encryptionPassword);
        salt = clearAndNullify(salt);
        log.info("Locked keystore.");
    }

    public boolean isLocked() {
        return keyStore == null
                || storePassword == null
                || salt == null
                || encryptionPassword == null
                || entryPassword == null
                || saltPassword == null;
    }

    public void deleteKeyEntry(String alias) {
        if (containsKeyAlias(alias)) {
            try {
                log.info("Deleting entry for alias [{}] from key store.", alias);
                keyStore.deleteEntry(alias);
            } catch (KeyStoreException e) {
                String msg = String.format("Error deleting alias [%s] from keystore: %s", alias, e.getMessage());
                log.error(msg);
                throw new KeyManagementException(msg, e);
            }
        }
    }

    public boolean containsKeyAlias(String alias) {
        try {
            return keyStore != null && !StringUtils.equalsIgnoreCase(alias, ALIAS_SALT) && keyStore.containsAlias(alias);
        } catch (KeyStoreException e) {
            String msg = String.format("Error looking up alias [%s] in keystore: %s", alias, e.getMessage());
            log.error(msg);
            throw new KeyManagementException(msg, e);
        }
    }

    void saveKeyStore() {
        if (keyStore != null && ArrayUtils.isNotEmpty(storePassword)) {
            keyStoreRepository.save(keyStore, storePassword);
        } else {
            log.info("Cannot save keystore. Keystore loaded: [{}]. Password available: [{}].",
                    keyStore != null, ArrayUtils.isNotEmpty(storePassword));
        }
    }

    private static byte[] getPrivateKey(KeyStore keyStore, String alias, char[] password) {
        try {
            KeyStore.SecretKeyEntry secretKeyEntry = (KeyStore.SecretKeyEntry) keyStore.getEntry(alias, passwordProtectionForRetrieving(password));
            SecretKey secretKey = secretKeyEntry.getSecretKey();
            return secretKey.getEncoded();
        } catch (NoSuchAlgorithmException | UnrecoverableEntryException | KeyStoreException e) {
            String msg = String.format("Cannot retrieve entry [%s]: %s", alias, e.getMessage());
            log.warn(msg);
            throw new KeyManagementException(msg, e);
        }
    }

    private static void putKey(KeyStore keyStore, String alias, SecretKey secretKey, char[] password) {
        try {
            KeyStore.SecretKeyEntry entry = new KeyStore.SecretKeyEntry(secretKey);
            keyStore.setEntry(alias, entry, passwordProtectionForStoring(password)); // override is ok
        } catch (KeyStoreException e) {
            String msg = String.format("Error adding key for [%s] to keystore: %s", alias, e.getMessage());
            log.error(msg);
            throw new KeyManagementException(msg, e);
        }
    }

    private char[] deriveKey(byte[] password, byte[] salt, String info) {
        return ConversionUtil.bytesToChars(encryptionUtility.deriveKeyBase64(password, salt, info));
    }

    private static KeyStore.PasswordProtection passwordProtectionForRetrieving(char[] password) {
        return new KeyStore.PasswordProtection(password);
    }

    static KeyStore.PasswordProtection passwordProtectionForStoring(char[] password) {
        return new KeyStore.PasswordProtection(password,
                "PBEWithHmacSHA256AndAES_256",
                new PBEParameterSpec(KeyGenerators.secureRandom(16).generateKey(), ITERATION_COUNT_PBE));
    }

    static SecretKey createSecretKey(byte[] bytes) {
        return new SecretKeySpec(bytes, "AES"); // AES is not always right but who cares
    }

    @SuppressWarnings("SameReturnValue")
    private static char[] clearAndNullify(char[] arr) {
        if (arr != null && arr.length > 0) {
            Arrays.fill(arr, 'X');
        }
        return null;
    }

    @SuppressWarnings("SameReturnValue")
    private static byte[] clearAndNullify(byte[] arr) {
        if (arr != null && arr.length > 0) {
            Arrays.fill(arr, (byte) 0);
        }
        return null;
    }
}
