package at.qubic.shell.command;

import org.junit.jupiter.api.Test;
import at.qubic.api.crypto.IdentityUtil;
import at.qubic.api.domain.qx.response.AssetOrder;
import at.qubic.api.domain.qx.response.EntityOrder;
import at.qubic.api.domain.qx.response.QxFees;
import at.qubic.shell.command.service.QxService;
import at.qubic.shell.wallet.Asset;
import at.qubic.shell.wallet.Assets;
import reactor.core.publisher.Mono;

import java.util.List;

import static org.assertj.core.api.Assertions.assertThat;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

class QxFunctionsCommandsTest {

    private final Assets assets = mock();
    private final IdentityUtil identityUtil = mock();
    private final QxService qxService = mock();
    private final QxFunctionsCommands command = new QxFunctionsCommands(qxService, identityUtil, assets);

    @Test
    void getFees() {
        when(qxService.getQxFees()).thenReturn(Mono.just(QxFees.builder()
                .assetIssuanceFee(1_000_000_000)
                .transferFee(1_000_000)
                .tradeFee(5_000_000)
                .build()));
        assertThat(command.getFees()).isEqualTo("""
                Issuance Fee: 1'000'000'000 qu
                Transfer Fee: 1'000'000 qu
                Trading Fee:  0.5 %""");
    }

    @Test
    void getEntityOrders() {
        when(identityUtil.getIdentityFromPublicKey(new byte[32])).thenReturn("AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAFXIB");
        when(identityUtil.isValidIdentity("identity")).thenReturn(true);
        when(qxService.getEntityBidOrders("identity", 0)).thenReturn(Mono.just(List.of(EntityOrder.builder()
                .issuer(new byte[32])
                        .price(1)
                        .assetName(new byte[] {'F', 'O', 'O'})
                        .numberOfShares(2)
                .build())));
        when(qxService.getEntityAskOrders("identity", 0)).thenReturn(Mono.just(List.of(EntityOrder.builder()
                .issuer(new byte[32])
                .price(2)
                .assetName(new byte[] {'F', 'O', 'O'})
                .numberOfShares(1)
                .build())));
        String result = command.getEntityOrders("identity", true, true, 10, 10, 0, 0);
        assertThat(result).contains("BID    FOO Shares: 2 Price: 1 Issuer: AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAFXIB");
        assertThat(result).contains("ASK    FOO Shares: 1 Price: 2 Issuer: AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAFXIB");
    }

    @Test
    void getEntityBidOrders() {
        when(identityUtil.getIdentityFromPublicKey(new byte[32])).thenReturn("AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAFXIB");
        when(identityUtil.isValidIdentity("identity")).thenReturn(true);
        when(qxService.getEntityBidOrders("identity", 0)).thenReturn(Mono.just(List.of(EntityOrder.builder()
                        .issuer(new byte[32])
                        .price(1)
                        .assetName(new byte[] {'F', 'O', 'O'})
                        .numberOfShares(2)
                        .build(),
                EntityOrder.builder()
                        .issuer(new byte[32])
                        .price(3)
                        .assetName(new byte[] {'B', 'A', 'R'})
                        .numberOfShares(4)
                        .build())));
        String result = command.getEntityOrders("identity", false, true, 10, 10, 0, 0);
        assertThat(result).contains("BID    FOO Shares: 2 Price: 1 Issuer: AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAFXIB");
        assertThat(result).contains("BID    BAR Shares: 4 Price: 3 Issuer: AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAFXIB");
    }

    @Test
    void getEntityAskOrders() {
        when(identityUtil.getIdentityFromPublicKey(new byte[32])).thenReturn("AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAFXIB");
        when(identityUtil.isValidIdentity("identity")).thenReturn(true);
        when(qxService.getEntityAskOrders("identity", 0)).thenReturn(Mono.just(List.of(EntityOrder.builder()
                        .issuer(new byte[32])
                        .price(2)
                        .assetName(new byte[] {'F', 'O', 'O'})
                        .numberOfShares(1)
                        .build(),
                EntityOrder.builder()
                        .issuer(new byte[32])
                        .price(3)
                        .assetName(new byte[] {'B', 'A', 'R'})
                        .numberOfShares(4)
                        .build())));
        String result = command.getEntityOrders("identity", true, false, 10, 10, 0, 0);
        assertThat(result).contains("ASK    FOO Shares: 1 Price: 2 Issuer: AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAFXIB");
        assertThat(result).contains("ASK    BAR Shares: 4 Price: 3 Issuer: AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAFXIB");
    }

    @Test
    void getEntityOrders_givenInvalidIdentity_thenMessage() {
        String result = command.getEntityOrders("foo", true, true, 1, 1, 0, 0);
        assertThat(result).containsIgnoringCase("identity [foo] is invalid");
    }

    @Test
    void getAssetOrders() {
        when(identityUtil.getIdentityFromPublicKey("ABC".getBytes())).thenReturn("ABC");
        when(identityUtil.isValidIdentity("XYZ")).thenReturn(true);
        when(qxService.getAssetBidOrders("XYZ", "FOO", 0)).thenReturn(Mono.just(List.of(AssetOrder.builder()
                        .price(2)
                        .entity("ABC".getBytes())
                        .numberOfShares(3)
                        .build(),
                AssetOrder.builder()
                        .price(1)
                        .entity("ABC".getBytes())
                        .numberOfShares(2)
                        .build())));
        when(qxService.getAssetAskOrders("XYZ", "FOO", 0)).thenReturn(Mono.just(List.of(AssetOrder.builder()
                        .price(3)
                        .entity("ABC".getBytes())
                        .numberOfShares(4)
                        .build(),
                AssetOrder.builder()
                        .price(4)
                        .entity("ABC".getBytes())
                        .numberOfShares(5)
                        .build())));
        String result = command.getAssetOrders("FOO", "XYZ", true, true, 10, 10, 0, 0);
        assertThat(result).contains("BID from ABC Price: 1 Shares: 2");
        assertThat(result).contains("BID from ABC Price: 2 Shares: 3");
        assertThat(result).contains("ASK from ABC Price: 3 Shares: 4");
        assertThat(result).contains("ASK from ABC Price: 4 Shares: 5");

        String expected = """
            BID from ABC Price: 1 Shares: 2
            BID from ABC Price: 2 Shares: 3
            -
            ASK from ABC Price: 3 Shares: 4
            ASK from ABC Price: 4 Shares: 5""";
        assertThat(result).contains(expected);
    }

    @Test
    void getAssetOrders_givenNoIssuer_thenTryToInfer() {
        when(identityUtil.getIdentityFromPublicKey("ABC".getBytes())).thenReturn("ABC");
        when(identityUtil.isValidIdentity("XYZ")).thenReturn(true);
        when(assets.getUniqueAssetForName("FOO")).thenReturn(Asset.builder().name("FOO").issuer("XYZ").build());
        when(qxService.getAssetBidOrders("XYZ", "FOO", 0)).thenReturn(Mono.just(List.of()));
        String result = command.getAssetOrders("FOO", null, false, true, 10, 10, 0, 0);
        assertThat(result).containsIgnoringWhitespaces("Orders for [FOO] of issuer[XYZ]: No BID orders.");
    }

    @Test
    void getAssetBidOrders() {
        when(identityUtil.getIdentityFromPublicKey("ABC".getBytes())).thenReturn("ABC");
        when(identityUtil.isValidIdentity("XYZ")).thenReturn(true);
        when(qxService.getAssetBidOrders("XYZ", "FOO", 0)).thenReturn(Mono.just(List.of(AssetOrder.builder()
                        .price(2)
                        .entity("ABC".getBytes())
                        .numberOfShares(3)
                        .build(),
                AssetOrder.builder()
                        .price(1)
                        .entity("ABC".getBytes())
                        .numberOfShares(2)
                        .build()
                )));
        String result = command.getAssetOrders("FOO", "XYZ", false, true, 10, 10, 0, 0);
        assertThat(result).contains("BID from ABC Price: 1 Shares: 2");
        assertThat(result).contains("BID from ABC Price: 2 Shares: 3");

        String expected = """
            BID from ABC Price: 1 Shares: 2
            BID from ABC Price: 2 Shares: 3
            """;
        assertThat(result).contains(expected);
    }

    @Test
    void getAssetAskOrders() {
        when(identityUtil.getIdentityFromPublicKey("ABC".getBytes())).thenReturn("ABC");
        when(identityUtil.isValidIdentity("XYZ")).thenReturn(true);
        when(qxService.getAssetAskOrders("XYZ", "FOO", 0)).thenReturn(Mono.just(List.of(AssetOrder.builder()
                        .price(3)
                        .entity("ABC".getBytes())
                        .numberOfShares(4)
                        .build(),
                AssetOrder.builder()
                        .price(4)
                        .entity("ABC".getBytes())
                        .numberOfShares(5)
                        .build())));
        String result = command.getAssetOrders("FOO", "XYZ", true, false, 10, 10, 0, 0);
        assertThat(result).contains("ASK from ABC Price: 3 Shares: 4");
        assertThat(result).contains("ASK from ABC Price: 4 Shares: 5");

        String expected = """
            ASK from ABC Price: 3 Shares: 4
            ASK from ABC Price: 4 Shares: 5""";
        assertThat(result).contains(expected);
    }

    @Test
    void getAssetOrders_givenInvalidIssuer_thenMessage() {
        String result = command.getAssetOrders("FOO", "XYZ", false, true, 10, 10, 0, 0);
        assertThat(result).containsIgnoringCase("identity [XYZ] is invalid");
    }

}