package at.qubic.shell.command;

import at.qubic.api.crypto.IdentityUtil;
import at.qubic.api.domain.qx.request.QxAssetOrder;
import at.qubic.shell.component.QxOrderComponent;
import at.qubic.shell.exception.InputValidationException;
import at.qubic.shell.exception.UnknownAssetException;
import at.qubic.shell.wallet.Assets;
import lombok.Builder;
import org.junit.jupiter.api.Test;
import reactor.core.publisher.Mono;

import java.util.function.Function;

import static org.assertj.core.api.Assertions.assertThat;
import static org.mockito.Mockito.*;

abstract class QxProcedureCommandsTest {

    protected final IdentityUtil identityUtil = mock();
    protected final QxOrderComponent orderComponent = mock();
    private final Assets assets = mock();
    protected final QxProcedureCommands commands = new QxProcedureCommands(identityUtil, orderComponent,  assets);

    @Test
    void order_givenInvalidIssuer_thenError() throws InputValidationException {
        doThrow(new InputValidationException("invalid issuer")).when(orderComponent).validateIdentity("issuer");
        String result = executeProcedureCall().apply(defaultArguments());
        assertThat(result).contains("invalid issuer");
    }

    @Test
    void order_givenUnknownAssetWithoutIssuer_thenError() {
        QxOrderArguments arguments = defaultArgumentsBuilder().issuer(null).build();
        when(assets.getUniqueAssetForName(arguments.assetName))
                .thenThrow(new UnknownAssetException("Unknown asset message"));
        when(identityUtil.isValidIdentity(any())).thenReturn(true);
        String result = executeProcedureCall().apply(arguments);
        assertThat(result).isEqualTo("Unknown asset message");
    }

    @Test
    void order() throws InputValidationException {
        when(orderComponent.sendQxOrder(eq("from"), any(QxAssetOrder.class), eq(3), eq(0), eq(true))).thenReturn(Mono.empty());
        String result = executeProcedureCall().apply(defaultArguments());
        assertThat(result).isBlank();
    }

    protected static QxOrderArguments defaultArguments() {
        return defaultArgumentsBuilder().build();
    }

    protected abstract Function<QxOrderArguments, String> executeProcedureCall();

    protected static QxOrderArguments.QxOrderArgumentsBuilder defaultArgumentsBuilder() {
        return QxOrderArguments.builder()
                .from("from")
                .assetName("assetName")
                .issuer("issuer")
                .offset(3)
                .retries(0)
                .validate(true);
    }

    @Builder
    record QxOrderArguments(String from, String assetName, String issuer, long pricePerShare, long numberOfShares,
                            int offset, int retries, boolean validate) {
    }
}