package at.qubic.shell.component;

import at.qubic.api.domain.std.SignedTransaction;
import at.qubic.shell.command.service.InclusionCheck;
import at.qubic.shell.command.service.InclusionCheck.InclusionStatus;
import at.qubic.shell.component.terminal.TransactionTerminal;
import at.qubic.shell.exception.AfterSendingException;
import at.qubic.shell.exception.NotIncludedException;
import lombok.extern.slf4j.Slf4j;
import org.junit.jupiter.api.Test;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;
import reactor.test.StepVerifier;

import java.time.Duration;
import java.util.List;

import static org.mockito.ArgumentMatchers.anyInt;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

@Slf4j
class InclusionCheckComponentTest {

    private final TransactionTerminal<?> writer = mock(TransactionTerminal.class);
    private final InclusionCheck inclusionCheck = mock();
    private final InclusionCheckComponent component = new InclusionCheckComponent(inclusionCheck);

    @Test
    void waitForInclusion() {
        InclusionStatus expected = InclusionStatus.builder()
                .checkedTick(42)
                .signedTransaction(SignedTransaction.builder().build())
                .build();

        when(inclusionCheck.checkForInclusion(42, "foo")).thenReturn(Flux.fromIterable(List.of(
                expected
        )));

        when(writer.writeTransactionProgress(anyInt(), anyInt(), anyInt())).then(x -> Mono.just("x"));

        StepVerifier.create(component.waitForInclusion(writer, 41, 42, "foo"))
                .expectNext(expected)
                .verifyComplete();
    }

    @Test
    void waitForInclusion_givenEmpty_thenError() {
        when(inclusionCheck.checkForInclusion(42, "foo")).thenReturn(Flux.empty());
        when(writer.writeTransactionProgress(anyInt(), anyInt(), anyInt())).then(x -> Mono.just("x"));

        StepVerifier.create(component.waitForInclusion(writer, 41, 42, "foo"))
                .expectError(AfterSendingException.class)
                .verify(Duration.ofSeconds(1));
    }

    @Test
    void waitForInclusion_givenError_thenPropagate() {
        when(inclusionCheck.checkForInclusion(42, "foo")).thenReturn(Flux.error(() -> new NotIncludedException("boom")));
        when(writer.writeTransactionProgress(anyInt(), anyInt(), anyInt())).then(x -> Mono.just("x"));

        StepVerifier.create(component.waitForInclusion(writer, 41, 42, "foo"))
                .expectError(NotIncludedException.class)
                .verify(Duration.ofSeconds(1));



    }

}