package at.qubic.shell.file;

import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.RandomStringUtils;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.springframework.security.crypto.keygen.KeyGenerators;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;

import static org.assertj.core.api.Assertions.assertThat;

class FileUtilityIT {

    private static Path TEMP_DIR;
    private final FileUtility fileUtility = new FileUtility();

    @BeforeAll
    static void setUp() throws IOException {
       TEMP_DIR = Files.createTempDirectory("qubicj-shell-test");
    }

    @Test
    void load() throws IOException {
        Path path = TEMP_DIR.resolve(randomFilename());
        FileUtils.writeLines(path.toFile(), StandardCharsets.UTF_8.name(), List.of("abc", "def", "ghi"));
        assertThat(fileUtility.load(path)).containsExactly("abc", "def", "ghi");
    }

    @Test
    void load_givenNonexistent_thenEmpty() throws IOException {
        List<String> result = fileUtility.load(TEMP_DIR.resolve(randomFilename()));
        assertThat(result).isEmpty();
    }

    @Test
    void saveStrings() throws IOException {
        Path path = TEMP_DIR.resolve(randomFilename());
        fileUtility.save(path, List.of("abc", "def", "ghi"));
        assertThat(FileUtils.readLines(path.toFile(), StandardCharsets.UTF_8)).containsExactly("abc", "def", "ghi");
    }

    @Test
    void saveBinary() throws IOException {
        Path path = TEMP_DIR.resolve(randomFilename());
        byte[] random = KeyGenerators.secureRandom().generateKey();
        fileUtility.save(path, random);
        assertThat(FileUtils.readFileToByteArray(path.toFile())).isEqualTo(random);
    }

    @Test
    void loadBinary() throws IOException {
        Path path = TEMP_DIR.resolve(randomFilename());
        byte[] random = KeyGenerators.secureRandom().generateKey();
        FileUtils.writeByteArrayToFile(path.toFile(), random);
        byte[] loaded = fileUtility.loadBinary(path);
        assertThat(loaded).isEqualTo(random);
    }

    @Test
    void loadBinary_givenNonexistent_thenEmpty() throws IOException {
        Path path = TEMP_DIR.resolve(randomFilename());
        byte[] loaded = fileUtility.loadBinary(path);
        assertThat(loaded).isEmpty();
    }

    @Test
    void move() throws IOException {
        Path source = Files.createTempFile(TEMP_DIR, "movesource", "test");
        Path target = source.resolveSibling("movetarget.test");
        fileUtility.move(source, target);
        assertThat(target.toFile().exists()).isTrue();
        assertThat(source.toFile().exists()).isFalse();
    }

    @Test
    void delete() throws IOException {
        Path source = Files.createTempFile(TEMP_DIR, "deletesource", "test");
        fileUtility.delete(source);
        assertThat(source.toFile().exists()).isFalse();
    }

    private static String randomFilename() {
        return RandomStringUtils.insecure().nextAlphanumeric(10);
    }

}