package at.qubic.shell.repository;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import at.qubic.shell.exception.KeyManagementException;
import at.qubic.shell.file.FileUtility;
import org.springframework.security.crypto.keygen.KeyGenerators;

import javax.crypto.SecretKey;
import javax.crypto.SecretKeyFactory;
import javax.crypto.spec.PBEKeySpec;
import java.io.*;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.security.KeyStore;
import java.security.UnrecoverableKeyException;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatThrownBy;

class KeyStoreRepositoryIT {

    private static final char[] PASSWORD = "test-pwd".toCharArray();
    private static final String TEMP_DIR_NAME = System.getProperties().getProperty("java.io.tmpdir") + "/test";
    private static final String KS_FILE_NAME = "test-ks";
    private static final String SALT_FILE_NAME = "test-salt";
    public static final String TEST_ALIAS = "test-alias";
    private final FileUtility fileUtility = new FileUtility();
    private final KeyStoreRepository repository = new KeyStoreRepository(fileUtility, TEMP_DIR_NAME, KS_FILE_NAME, SALT_FILE_NAME);


    @BeforeEach
    @AfterEach
    void cleanUp() throws IOException {
        Files.deleteIfExists(Paths.get(TEMP_DIR_NAME, KS_FILE_NAME));
        Files.deleteIfExists(Paths.get(TEMP_DIR_NAME, SALT_FILE_NAME));
    }

    @Test
    void load_givenNew_thenCreateKeystore() throws Exception {
        Path path = Paths.get(TEMP_DIR_NAME, KS_FILE_NAME);
        assertThat(Files.exists(path)).isFalse();

        KeyStore loaded = repository.load(KeyStore.getInstance(KeyStore.getDefaultType()), PASSWORD);
        assertThat(loaded).isNotNull();
    }

    @Test
    void load_givenKeystore_thenReturnKeystore() throws Exception {
        Path path = Paths.get(TEMP_DIR_NAME, KS_FILE_NAME);
        KeyStore keyStore = createKeystoreWithEntry();
        repository.save(keyStore, PASSWORD);
        assertThat(Files.exists(path)).isTrue();

        KeyStore loaded = repository.load(KeyStore.getInstance(KeyStore.getDefaultType()), PASSWORD);
        assertThat(loaded.containsAlias(TEST_ALIAS)).isTrue();
        assertThat(loaded.getEntry(TEST_ALIAS, new KeyStore.PasswordProtection(new char[0]))).isNotNull();
    }

    @Test
    void loadPasswordSalt() throws Exception {
        Path path = Paths.get(TEMP_DIR_NAME, SALT_FILE_NAME);
        byte[] salt = {1, 2, 3};
        fileUtility.save(path, salt);
        assertThat(repository.loadPasswordSalt()).isEqualTo(salt);
    }

    @Test
    void loadPasswordSalt_givenNone_thenCreate() {
        Path saltPath = Paths.get(TEMP_DIR_NAME, SALT_FILE_NAME);
        assertThat(Files.exists(saltPath)).isFalse();
        assertThat(repository.loadPasswordSalt()).isNotNull();
        assertThat(Files.exists(saltPath)).isTrue();
    }

    @Test
    void loadKeyStore_givenWrongPassword_thenThrow() throws Exception {
        KeyStore keystore = repository.load(KeyStore.getInstance(KeyStore.getDefaultType()), PASSWORD);
        repository.save(keystore, PASSWORD);
        char[] wrongPassword = "wrong".toCharArray();
        assertThatThrownBy(() -> repository.load(KeyStore.getInstance(KeyStore.getDefaultType()), wrongPassword))
                .isInstanceOf(KeyManagementException.class)
                .hasRootCauseInstanceOf(UnrecoverableKeyException.class);
    }

    @Test
    void moveKeystore() throws Exception {
        KeyStore keystore = repository.load(KeyStore.getInstance(KeyStore.getDefaultType()), null);
        repository.save(keystore, null);
        Path path = repository.moveKeystore();
        assertThat(Files.exists(path)).isTrue();
        Files.delete(path); // clean up
    }

    @Test
    void moveSaltFile() throws Exception {
        repository.loadPasswordSalt();
        Path path = repository.moveSaltFile();
        assertThat(Files.exists(path)).isTrue();
        Files.delete(path); // clean up
    }

    @Test
    void delete() throws IOException {
        Path path = Files.createTempFile("to-be-deleted", "test");
        repository.delete(path);
        assertThat(Files.exists(path)).isFalse();
    }

    @Test
    void save() throws Exception {
        Path path = Paths.get(TEMP_DIR_NAME, KS_FILE_NAME);
        assertThat(Files.exists(path)).isFalse();
        KeyStore keyStore = KeyStore.getInstance(KeyStore.getDefaultType());
        keyStore.load(null, PASSWORD);
        repository.save(keyStore, PASSWORD);
        repository.load(KeyStore.getInstance(KeyStore.getDefaultType()), null).containsAlias("test");

        assertThat(Files.exists(path)).isTrue();
    }

    private KeyStore createKeystoreWithEntry() throws Exception {
        KeyStore keyStore = KeyStore.getInstance(KeyStore.getDefaultType());
        keyStore.load(null, PASSWORD);
        KeyStore.SecretKeyEntry secret = new KeyStore.SecretKeyEntry(generateRandomSecretKey());
        keyStore.setEntry(TEST_ALIAS, secret, new KeyStore.PasswordProtection(new char[0]));
        return keyStore;
    }

    public SecretKey generateRandomSecretKey() throws Exception{
        SecretKeyFactory secretKeyFactory = SecretKeyFactory.getInstance("PBKDF2WithHmacSHA1");
        PBEKeySpec keySpec = new PBEKeySpec(KeyGenerators.string().generateKey().toCharArray(),
                KeyGenerators.secureRandom().generateKey(), 1000, 16);
        return secretKeyFactory.generateSecret(keySpec);
    }

}