package at.qubic.shell.service;

import at.qubic.shell.command.service.EntityCommandService;
import lombok.extern.slf4j.Slf4j;
import org.junit.jupiter.api.Test;
import at.qubic.api.crypto.IdentityUtil;
import at.qubic.api.domain.std.response.*;
import at.qubic.api.service.ComputorService;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.nio.charset.StandardCharsets;

import static java.nio.charset.StandardCharsets.UTF_8;
import static org.assertj.core.api.Assertions.assertThat;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

@Slf4j
class EntityCommandServiceTest {

    private final IdentityUtil identityUtil = mock();
    private final ComputorService computorService = mock();
    private final EntityCommandService service = new EntityCommandService(computorService, identityUtil);

    @Test
    void getEntity_thenReturnEntityString() {
        when(identityUtil.isValidIdentity("some-id")).thenReturn(true);
        when(computorService.getEntity("some-id")).thenReturn(Mono.just(Entity.builder()
                .tick(42)
                .incomingAmount(10)
                .outgoingAmount(9)
                .numberOfIncomingTransfers(8)
                .numberOfOutgoingTransfers(7)
                .latestIncomingTransferTick(6)
                .latestOutgoingTransferTick(5)
                .build()));

        final String result = service.getEntity("some-id");
        log.info(result);

        assertThat(result).contains("Identity: some-id");
        assertThat(result).contains("Balance: 1");
        assertThat(result).contains("Incoming amount: 10");
        assertThat(result).contains("Outgoing amount: 9");
        assertThat(result).contains("Incoming transfers: 8");
        assertThat(result).contains("Outgoing transfers: 7");
        assertThat(result).contains("Latest incoming tick: 6");
        assertThat(result).contains("Latest outgoing tick: 5");
        assertThat(result).contains("Tick: 42");
    }

    @Test
    void getEntity_givenEmptyResponse_thenReturnMessage() {
        when(identityUtil.isValidIdentity("some-id")).thenReturn(true);
        when(computorService.getEntity("some-id")).thenReturn(Mono.empty());
        String result = service.getEntity("some-id");
        assertThat(result).contains("Could not retrieve entity [some-id]");
    }

    @Test
    void getEntity_givenError_thenReturnMessage() {
        when(identityUtil.isValidIdentity("some-id")).thenReturn(true);
        when(computorService.getEntity("some-id")).thenReturn(Mono.error(new RuntimeException("test")));
        String result = service.getEntity("some-id");
        assertThat(result).contains("Could not retrieve entity [some-id]");
    }

    @Test
    void getEntity_givenInvalidIdentity_thenReturnErrorMessage() {
        assertThat(service.getEntity("invalid-id")).contains("Identity [invalid-id] is invalid");
    }

    @Test
    void getOwnedAssets_thenReturnAssetsString() {
        when(identityUtil.getIdentityFromPublicKey(new byte[32])).thenReturn("AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAFXIB");
        when(identityUtil.isValidIdentity("some-id")).thenReturn(true);
        when(computorService.getOwnedAssets("some-id")).thenReturn(Flux.just(
                RespondOwnedAssets.builder()
                        .asset(AssetOwnership.builder()
                                .issuanceIndex(1)
                                .numberOfUnits(2)
                                .managingContractIndex((short) 3)
                                .build())
                        .issuanceAsset(AssetIssuance.builder()
                                .publicKey(new byte[32])
                                .name("QX".getBytes(StandardCharsets.UTF_8))
                                .build())
                        .tick(12345)
                        .build(),
                RespondOwnedAssets.builder()
                        .asset(AssetOwnership.builder()
                                .issuanceIndex(4)
                                .numberOfUnits(5)
                                .managingContractIndex((short) 6)
                                .build())
                        .issuanceAsset(AssetIssuance.builder()
                                .publicKey(new byte[32])
                                .name("RANDOM".getBytes(StandardCharsets.UTF_8))
                                .build())
                        .tick(12346)
                        .build()
        ));

        final String result = service.getOwnedAssets("some-id");
        log.info(result);

        assertThat(result).contains("Issuer: AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAFXIB");

        assertThat(result).containsOnlyOnce("Units: 2");
        assertThat(result).containsOnlyOnce("Name: QX");
        assertThat(result).containsOnlyOnce("Issuance index: 1");
        assertThat(result).containsOnlyOnce("Managing contract index: 3");
        assertThat(result).containsOnlyOnce("Tick: 12345");

        assertThat(result).containsOnlyOnce("Units: 5");
        assertThat(result).containsOnlyOnce("Name: RANDOM");
        assertThat(result).containsOnlyOnce("Issuance index: 4");
        assertThat(result).containsOnlyOnce("Managing contract index: 6");
        assertThat(result).containsOnlyOnce("Tick: 12346");
    }

    @Test
    void getOwnedAssets_givenEmptyResponse_thenReturnMessage() {
        when(identityUtil.isValidIdentity("some-id")).thenReturn(true);
        when(computorService.getOwnedAssets("some-id")).thenReturn(Flux.empty());
        String result = service.getOwnedAssets("some-id");
        assertThat(result).contains("[some-id] owns no assets.");
    }

    @Test
    void getOwnedAssets_givenError_thenReturnMessage() {
        when(identityUtil.isValidIdentity("some-id")).thenReturn(true);
        when(computorService.getOwnedAssets("some-id")).thenReturn(Flux.error(new RuntimeException("test")));
        String result = service.getOwnedAssets("some-id");
        assertThat(result).contains("Could not retrieve owned assets for [some-id].");
    }

    @Test
    void getOwnedAssets_givenInvalidIdentity_thenReturnMessage() {
        assertThat(service.getOwnedAssets("invalid-id")).contains("Identity [invalid-id] is invalid");
    }

    @Test
    void getPossessedAssets_thenReturnAssetsString() {
        byte[] issuerPublicKey = "issuer".getBytes(UTF_8);
        when(identityUtil.getIdentityFromPublicKey(issuerPublicKey)).thenReturn("ISSUER_IDENTITY");
        byte[] ownerPublicKey = "owner".getBytes(UTF_8);
        when(identityUtil.getIdentityFromPublicKey(ownerPublicKey)).thenReturn("OWNER_IDENTITY");
        when(identityUtil.isValidIdentity("some-id")).thenReturn(true);
        when(computorService.getPossessedAssets("some-id")).thenReturn(Flux.just(
                RespondPossessedAssets.builder()
                        .asset(AssetPossession.builder()
                                .numberOfUnits(1)
                                .ownershipIndex(2)
                                .managingContractIndex((short) 3)
                                .build())
                        .ownershipAsset(AssetOwnership.builder()
                                .publicKey(ownerPublicKey)
                                .build())
                        .issuanceAsset(AssetIssuance.builder()
                                .publicKey(issuerPublicKey)
                                .name("QX".getBytes(StandardCharsets.UTF_8))
                                .build())
                        .tick(12345)
                        .build(),
                RespondPossessedAssets.builder()
                        .asset(AssetPossession.builder()
                                .numberOfUnits(4)
                                .ownershipIndex(5)
                                .managingContractIndex((short) 6)
                                .build())
                        .ownershipAsset(AssetOwnership.builder()
                                .publicKey(ownerPublicKey)
                                .build())
                        .issuanceAsset(AssetIssuance.builder()
                                .publicKey(issuerPublicKey)
                                .name("RANDOM".getBytes(StandardCharsets.UTF_8))
                                .build())
                        .tick(12346)
                        .build()
        ));

        final String result = service.getPossessedAssets("some-id");
        log.info(result);

        assertThat(result).contains("Issuer: ISSUER_IDENTITY");
        assertThat(result).contains("Owner: OWNER_IDENTITY");

        assertThat(result).containsOnlyOnce("Units: 1");
        assertThat(result).containsOnlyOnce("Name: QX");
        assertThat(result).containsOnlyOnce("Owner index: 2");
        assertThat(result).containsOnlyOnce("Managing contract index: 3");
        assertThat(result).containsOnlyOnce("Tick: 12345");

        assertThat(result).containsOnlyOnce("Units: 4");
        assertThat(result).containsOnlyOnce("Name: RANDOM");
        assertThat(result).containsOnlyOnce("Owner index: 5");
        assertThat(result).containsOnlyOnce("Managing contract index: 6");
        assertThat(result).containsOnlyOnce("Tick: 12346");
    }

    @Test
    void getPossessedAssets_givenEmptyResponse_thenReturnMessage() {
        when(identityUtil.isValidIdentity("some-id")).thenReturn(true);
        when(computorService.getPossessedAssets("some-id")).thenReturn(Flux.empty());
        String result = service.getPossessedAssets("some-id");
        assertThat(result).contains("[some-id] possesses no assets.");
    }

    @Test
    void getPossessedAssets_givenError_thenReturnMessage() {
        when(identityUtil.isValidIdentity("some-id")).thenReturn(true);
        when(computorService.getPossessedAssets("some-id")).thenReturn(Flux.error(new RuntimeException("test")));
        String result = service.getPossessedAssets("some-id");
        assertThat(result).contains("Could not retrieve possessed assets for [some-id].");
    }

    @Test
    void getPossessedAssets_givenInvalidIdentity_thenReturnMessage() {
        assertThat(service.getPossessedAssets("invalid-id")).contains("Identity [invalid-id] is invalid");
    }

}