package at.qubic.shell.service;

import at.qubic.shell.command.service.InclusionCheck;
import org.junit.jupiter.api.Test;
import at.qubic.api.domain.std.SignedTransaction;
import at.qubic.api.domain.std.Transaction;
import at.qubic.api.domain.std.response.TickInfo;
import at.qubic.api.service.ComputorService;
import at.qubic.shell.exception.NotIncludedException;
import at.qubic.shell.command.service.InclusionCheck.InclusionStatus;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;
import reactor.test.StepVerifier;

import java.time.Duration;
import java.util.List;

import static org.mockito.Mockito.*;

class InclusionCheckTest {

    public static final Duration TICK_CHECK_DURATION = Duration.ofSeconds(1);
    private final ComputorService computorService = mock();
    private final InclusionCheck inclusionCheck = new InclusionCheck(computorService, TICK_CHECK_DURATION);

    @Test
    void checkForInclusion() {
        SignedTransaction signedTransaction = SignedTransaction.builder()
                .transaction(Transaction.builder().build())
                .transactionHash("foo")
                .build();

        when(computorService.getCurrentTickInfo())
                .thenReturn(Mono.just(TickInfo.builder().tick(41).build()))
                .thenReturn(Mono.just(TickInfo.builder().tick(42).build()))
                .thenReturn(Mono.just(TickInfo.builder().tick(43).build()));
        when(computorService.getTickTransactions(42)).thenReturn(Flux.fromIterable(List.of(signedTransaction)));

        StepVerifier.withVirtualTime(() -> inclusionCheck.checkForInclusion(42, "foo"))
                .expectSubscription()
                .expectNoEvent(TICK_CHECK_DURATION) // initial
                .expectNext(InclusionStatus.builder().checkedTick(40).build())
                .expectNoEvent(TICK_CHECK_DURATION) // second
                .expectNext(InclusionStatus.builder().checkedTick(41).build())
                .expectNoEvent(TICK_CHECK_DURATION) // third tick
                .expectNext(InclusionStatus.builder().checkedTick(42).signedTransaction(signedTransaction).build())
                .verifyComplete();
    }

    @Test
    void checkForInclusion_givenNotIncluded_thenError() {
        when(computorService.getTickTransactions(42)).thenReturn(Flux.fromIterable(List.of(SignedTransaction.builder().transactionHash("another").build())));
        when(computorService.getCurrentTickInfo()).thenReturn(Mono.just(TickInfo.builder().tick(43).build()));

        StepVerifier.withVirtualTime(() -> inclusionCheck.checkForInclusion(42, "not found"))
                .expectSubscription()
                .expectNoEvent(TICK_CHECK_DURATION) // initial
                .expectError(NotIncludedException.class)
                .verify(Duration.ofSeconds(1));

    }

}