package at.qubic.shell.service;

import at.qubic.shell.command.service.KeyService;
import org.junit.jupiter.api.Test;
import at.qubic.api.crypto.IdentityUtil;
import at.qubic.shell.exception.CommandFailureException;
import at.qubic.shell.exception.KeyManagementException;
import at.qubic.shell.wallet.KeyManager;
import at.qubic.shell.wallet.Wallets;

import java.security.UnrecoverableKeyException;
import java.util.List;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatThrownBy;
import static org.mockito.Mockito.*;

class KeyServiceTest {

    private final KeyManager keyManager = mock();
    private final Wallets wallets = mock();
    private final IdentityUtil identityUtil = mock();
    private final KeyService keyManagementService = new KeyService(keyManager, wallets, identityUtil);

    @Test
    void tryToUnlockKeyStore() {
        assertThat(keyManagementService.tryToUnlockKeyStore("pwd")).containsIgnoringCase("Unlocked wallet");
        verify(keyManager).unlock("pwd".toCharArray());
    }

    @Test
    void tryToUnlockKeyStore_givenWrongPassword_thenMessage() {
        doThrow(new KeyManagementException("test message", new UnrecoverableKeyException("blah")))
                .when(keyManager).unlock("pwd".toCharArray());
        assertThat(keyManagementService.tryToUnlockKeyStore("pwd")).containsIgnoringCase("wrong password");
    }

    @Test
    void tryToUnlockKeyStore_givenUnexpectedError_thenThrow() {
        doThrow(new KeyManagementException("test", new RuntimeException("root cause")))
                .when(keyManager).unlock("pwd".toCharArray());
        assertThatThrownBy(() -> keyManagementService.tryToUnlockKeyStore("pwd"))
                .isInstanceOf(CommandFailureException.class)
                .message().containsIgnoringCase("Cannot unlock wallet");
    }

    @Test
    void tryToUnlockKeyStore_givenUnknownKeys_thenDeleteThem() {
        when(keyManager.getKeyAliases()).thenReturn(List.of("a", "b", "c")); // lower case
        when(wallets.containsWalletIdentity("A")).thenReturn(true); // upper case
        when(wallets.containsWalletIdentity("B")).thenReturn(true); // upper case

        keyManagementService.tryToUnlockKeyStore("pwd");

        verify(keyManager).getKeyAliases();
        verify(keyManager).unlock(any(char[].class));
        verify(keyManager).deleteKeyEntry("C");
        verifyNoMoreInteractions(keyManager);
    }

    @Test
    void tryToUpdatePassword() {
        String result = keyManagementService.tryToUpdatePassword("pass");
        assertThat(result).contains("Password successfully updated");
        verify(keyManager).updatePassword("pass".toCharArray());
    }

    @Test
    void tryToUpdatePassword_givenError_thenThrow() {
        doThrow(new KeyManagementException("test", new RuntimeException("root cause")))
                .when(keyManager).updatePassword("pwd".toCharArray());
        assertThatThrownBy(() -> keyManagementService.tryToUpdatePassword("pwd"))
                .isInstanceOf(CommandFailureException.class)
                .message().containsIgnoringCase("Password update failed");
    }

    @Test
    void isLocked() {
        assertThat(keyManagementService.isLocked()).isFalse();
        verify(keyManager).isLocked();
    }

    @Test
    void lock() {
        keyManagementService.lock();
        verify(keyManager).lock();
    }

    @Test
    void addSeed() {
        byte[] subSeed = new byte[32];
        when(identityUtil.getSubSeedFromSeed("seed")).thenReturn(subSeed);
        byte[] publicKey = {1, 2, 3};
        when(identityUtil.getPublicKeyFromSubSeed(new byte[32])).thenReturn(publicKey);
        when(identityUtil.getIdentityFromPublicKey(publicKey)).thenReturn("identity");
        when(identityUtil.isValidSeed(anyString())).thenReturn(true);

        String identity = keyManagementService.addSeed("seed");
        assertThat(identity).isEqualTo("identity");
        verify(keyManager).setPrivateKey(eq("identity"), any(byte[].class));
    }

    @Test
    void addSeed_givenExistingWallet_thenDoNotReplaceWallet() {
        when(identityUtil.getSubSeedFromSeed("seed")).thenReturn(new byte[32]);
        when(identityUtil.getIdentityFromPublicKey(any())).thenReturn("identity");
        when(wallets.containsWalletIdentity("identity")).thenReturn(true); // wallet exists
        when(identityUtil.isValidSeed(anyString())).thenReturn(true);

        String identity = keyManagementService.addSeed("seed");
        assertThat(identity).isEqualTo("identity");
        verify(keyManager).setPrivateKey(eq("identity"), any());
        verify(wallets, never()).putWallet(any()); // don't replace wallet
    }

    @Test
    void addSeed_givenNewWallet_thenAddWallet() {
        when(identityUtil.getSubSeedFromSeed("seed")).thenReturn(new byte[32]);
        when(identityUtil.getIdentityFromPublicKey(any())).thenReturn("identity");
        when(identityUtil.isValidSeed(anyString())).thenReturn(true);

        String identity = keyManagementService.addSeed("seed");
        assertThat(identity).isEqualTo("identity");
        verify(keyManager).setPrivateKey(eq("identity"), any());
        verify(wallets).putWallet(any());
    }

    @Test
    void addSeed_givenInvalidSeed_thenMessage() {
        assertThatThrownBy(() -> keyManagementService.addSeed("invalid"))
                .isInstanceOf(IllegalArgumentException.class)
                .message().containsIgnoringCase("Invalid seed");
    }

    @Test
    void verifyIdentity_givenValid_thenTrue() {
        when(keyManager.containsKeyAlias("identity")).thenReturn(true);
        when(keyManager.getPrivateKey("identity")).thenReturn(new byte[64]);
        byte[] privateKey = new byte[32];
        byte[] publicKey = {2, 3, 4};
        when(identityUtil.getPublicKeyFromPrivateKey(privateKey)).thenReturn(publicKey);
        when(identityUtil.getIdentityFromPublicKey(publicKey)).thenReturn("identity");

        boolean valid = keyManagementService.verifyIdentity("identity");
        assertThat(valid).isTrue();
    }

    @Test
    void verifyIdentity_givenInvalid_thenFalse() {
        when(keyManager.containsKeyAlias("identity")).thenReturn(true);

        boolean valid = keyManagementService.verifyIdentity("identity");
        assertThat(valid).isFalse();
        verify(keyManager).getPrivateKey("identity");
        verify(identityUtil).getPublicKeyFromPrivateKey(any());
        verify(identityUtil).getIdentityFromPublicKey(any());

    }

    @Test
    void verifyIdentity_givenUnknownIdentity_thenFalse() {
        boolean valid = keyManagementService.verifyIdentity("identity");

        assertThat(valid).isFalse();
        verifyNoInteractions(identityUtil);
        verify(keyManager, never()).getPrivateKey("identity");
    }

    @Test
    void containsKeyForIdentity() {
        when(keyManager.containsKeyAlias("identity")).thenReturn(true);
        boolean result = keyManagementService.containsKeyForIdentity("identity");
        assertThat(result).isTrue();
    }

}