package at.qubic.shell.util;

import lombok.extern.slf4j.Slf4j;
import org.apache.commons.codec.binary.Hex;
import org.bouncycastle.crypto.DerivationParameters;
import org.bouncycastle.crypto.digests.Kangaroo;
import org.bouncycastle.crypto.generators.HKDFBytesGenerator;
import org.bouncycastle.crypto.params.HKDFParameters;
import org.junit.jupiter.api.Test;

import java.nio.charset.StandardCharsets;
import java.time.Duration;
import java.time.Instant;

import static org.assertj.core.api.Assertions.assertThat;

@Slf4j
class HkdfTest {

    private static final Kangaroo.KangarooTwelve kangarooTwelve = new Kangaroo.KangarooTwelve();
    private static final HKDFBytesGenerator generator = new HKDFBytesGenerator(kangarooTwelve);

    @Test
    void hkdf() {
        int size = 32; // max is 4064

        Instant before = Instant.now();
        // derive two keys from 'password', 'salt', and 'info'
        DerivationParameters derivationParameters = new HKDFParameters("password".getBytes(), "salt".getBytes(StandardCharsets.UTF_8), "info".getBytes(StandardCharsets.UTF_8)); // extracts on init by default
        generator.init(derivationParameters);
        byte[] key1 = new byte[size];
        byte[] key2 = new byte[size];
        generator.generateBytes(key1, 0, size); // expand
        generator.generateBytes(key2, 0, size); // expand
        assertThat(Hex.encodeHexString(key1)).isEqualTo("1052f1afd1481ae5fdc6253860722fe56f006a1f87ae3b01fd7f73c2794ee5f3");
        assertThat(Hex.encodeHexString(key2)).isEqualTo("e60596c35a3bfb88ab2d2c84c8802cf18136e42beb34a1f4698529980aac22f8");

        // derive two keys from 'password', 'salt', and 'info2'
        derivationParameters = new HKDFParameters("password".getBytes(), "salt".getBytes(StandardCharsets.UTF_8), "info2".getBytes(StandardCharsets.UTF_8));
        generator.init(derivationParameters);
        key1 = new byte[size];
        key2 = new byte[size];
        generator.generateBytes(key1, 0, size);
        generator.generateBytes(key2, 0, size);
        assertThat(Hex.encodeHexString(key1)).isEqualTo("7194caf9d65c82255850d655b85cf0fc5df051a3e9e831d5405f2d650fe37ca9");
        assertThat(Hex.encodeHexString(key2)).isEqualTo("741ea32da52da0d998bad8fef0466f67ddc08f384a84e61d0a05dbf6a8a07b83");
        Instant after = Instant.now();
        log.info("Duration: {}", Duration.between(before, after));

        // derive prk first from 'password' and 'salt'. then create keys for 'info' and 'info2'. same output
        Instant beforePrk = Instant.now();
        byte[] prk = generator.extractPRK("salt".getBytes(StandardCharsets.UTF_8), "password".getBytes());
        derivationParameters = HKDFParameters.skipExtractParameters(prk, "info".getBytes(StandardCharsets.UTF_8)); // does not extract on init (again)
        generator.init(derivationParameters);
        key1 = new byte[size];
        key2 = new byte[size];
        generator.generateBytes(key1, 0, size);
        generator.generateBytes(key2, 0, size);
        assertThat(Hex.encodeHexString(key1)).isEqualTo("1052f1afd1481ae5fdc6253860722fe56f006a1f87ae3b01fd7f73c2794ee5f3");
        assertThat(Hex.encodeHexString(key2)).isEqualTo("e60596c35a3bfb88ab2d2c84c8802cf18136e42beb34a1f4698529980aac22f8");

        derivationParameters = HKDFParameters.skipExtractParameters(prk, "info2".getBytes(StandardCharsets.UTF_8));
        generator.init(derivationParameters);
        key1 = new byte[size];
        key2 = new byte[size];
        generator.generateBytes(key1, 0, size);
        generator.generateBytes(key2, 0, size);
        assertThat(Hex.encodeHexString(key1)).isEqualTo("7194caf9d65c82255850d655b85cf0fc5df051a3e9e831d5405f2d650fe37ca9");
        assertThat(Hex.encodeHexString(key2)).isEqualTo("741ea32da52da0d998bad8fef0466f67ddc08f384a84e61d0a05dbf6a8a07b83");
        Instant afterPrk = Instant.now();
        log.info("Duration: {}", Duration.between(beforePrk, afterPrk));
    }

}