package at.qubic.shell.util;

import lombok.extern.slf4j.Slf4j;
import org.apache.commons.codec.binary.Hex;
import org.bouncycastle.crypto.generators.SCrypt;
import org.junit.jupiter.api.Test;
import org.springframework.security.crypto.bcrypt.BCrypt;

import java.nio.charset.StandardCharsets;
import java.time.Duration;
import java.time.Instant;
import java.util.Base64;

import static org.assertj.core.api.Assertions.assertThat;

@Slf4j
public class PasswordHashingTest {

    /*
        OWASP recommendations https://cheatsheetseries.owasp.org/cheatsheets/Password_Storage_Cheat_Sheet.html
        N=2^17 (128 MiB), r=8 (1024 bytes), p=1
        N=2^16 (64 MiB), r=8 (1024 bytes), p=2
        N=2^15 (32 MiB), r=8 (1024 bytes), p=3
        N=2^14 (16 MiB), r=8 (1024 bytes), p=5
        N=2^13 (8 MiB), r=8 (1024 bytes), p=10
     */
    @Test
    void scrypt() {
        Instant before = Instant.now();
        byte[] generated = SCrypt.generate("password".getBytes(StandardCharsets.UTF_8), "salt".getBytes(StandardCharsets.UTF_8),
                32768, 8, 3, 32);
        Instant after = Instant.now();
        log.info("Duration scrypt: {}", Duration.between(before, after));
        assertThat(Hex.encodeHexString(generated)).isEqualTo("d1a6c3183615dadf244b2976f92ee1bfe7a45ccf3bc7ad9de6c6fa6c6796caa3");
    }

    private static final int ROUNDS = 12;
    // $<version>$<rounds>$<salt>
    public static final String BCRYPT_SALT_PREFIX = String.format("$2a$%s$", ROUNDS);

    @Test
    void bcrypt() {
        // bcrypt is no kdf!
        byte[] salt = Base64.getDecoder().decode("DCPyuN1dPoDGUPdLcRJ8xe");
        assertThat(salt).hasSize(16); // must be 16 bytes
        String saltString = BCRYPT_SALT_PREFIX + Base64.getEncoder().encodeToString(salt);
        Instant before = Instant.now();
        String hashed = BCrypt.hashpw(ConversionUtil.charsToBytes("password".toCharArray()), saltString);
        Instant after = Instant.now();
        log.info("Duration bcrypt: {}", Duration.between(before, after));
        assertThat(hashed).hasSize(60); // 32 bytes
        assertThat(hashed).isEqualTo("$2a$12$DCPyuN1dPoDGUPdLcRJ8xOHJLZvjTVyCSf9PfKc4dUpJeSH3Ry.UO");
        assertThat(BCrypt.checkpw("password", hashed)).isTrue();
    }

}
