package at.qubic.shell.wallet;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import at.qubic.shell.exception.KeyManagementException;
import at.qubic.shell.file.FileUtility;
import at.qubic.shell.repository.KeyStoreRepository;
import at.qubic.shell.util.EncryptionUtility;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.security.UnrecoverableKeyException;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatThrownBy;

class KeyManagerIT {

    private static final char[] PASSWORD = "test-pwd".toCharArray();
    private static final char[] NEW_PASSWORD = "new-test-pwd".toCharArray();
    private static final String TEMP_DIR_NAME = System.getProperties().getProperty("java.io.tmpdir") + "/test";
    private static final String KS_FILE_NAME = "test-ks";
    private static final String SALT_FILE_NAME = "test-salt";
    private final FileUtility fileUtility = new FileUtility();
    private final KeyStoreRepository keyStoreRepository = new KeyStoreRepository(fileUtility, TEMP_DIR_NAME, KS_FILE_NAME, SALT_FILE_NAME);
    private final EncryptionUtility encryptionUtility = new EncryptionUtility();
    private final KeyManager keyManager = new KeyManager(keyStoreRepository, encryptionUtility);

    @BeforeEach
    @AfterEach
    void cleanUp() throws IOException {
        Files.deleteIfExists(Paths.get(TEMP_DIR_NAME, KS_FILE_NAME));
        Files.deleteIfExists(Paths.get(TEMP_DIR_NAME, SALT_FILE_NAME));
    }

    @Test
    void unlock() {
        keyManager.unlock(PASSWORD);
        assertThat(keyManager.isLocked()).isFalse();
        assertThat(Files.exists(Paths.get(TEMP_DIR_NAME, SALT_FILE_NAME))).isTrue();
        assertThat(Files.exists(Paths.get(TEMP_DIR_NAME, KS_FILE_NAME))).isFalse(); // not saved yet
    }

    @Test
    void unlock_givenWrongPassword_thenThrow() {
        keyManager.unlock(PASSWORD);
        keyManager.lock();
        assertThatThrownBy(() -> keyManager.unlock("wrong".toCharArray()))
                .isInstanceOf(KeyManagementException.class)
                .hasRootCauseInstanceOf(UnrecoverableKeyException.class);
    }

    @Test
    void unlock_givenEmptyPassword_thenThrow() {
        assertThatThrownBy(() -> keyManager.unlock(null)).isInstanceOf(IllegalArgumentException.class);
    }

    @Test
    void unlock_givenUnlocked_thenThrow() {
        keyManager.unlock(PASSWORD);
        assertThatThrownBy(() -> keyManager.unlock(PASSWORD)).isInstanceOf(IllegalStateException.class);
    }

    @Test
    void lock() {
        assertThat(keyManager.isLocked()).isTrue();
        keyManager.unlock(PASSWORD);
        assertThat(keyManager.isLocked()).isFalse();
        keyManager.lock();
        assertThat(keyManager.isLocked()).isTrue();
    }

    @Test
    void updatePassword() {
        keyManager.unlock(PASSWORD);
        byte[] data = {1, 2, 3};
        keyManager.setPrivateKey("foo", data);

        // update
        keyManager.updatePassword(NEW_PASSWORD);
        assertThat(keyManager.getPrivateKey("foo")).isEqualTo(data);
        assertThat(Files.exists(Paths.get(TEMP_DIR_NAME, SALT_FILE_NAME))).isTrue();
        assertThat(Files.exists(Paths.get(TEMP_DIR_NAME, KS_FILE_NAME))).isTrue();
        assertThat(Files.exists(Paths.get(TEMP_DIR_NAME, SALT_FILE_NAME + ".bak"))).isFalse();
        assertThat(Files.exists(Paths.get(TEMP_DIR_NAME, KS_FILE_NAME + ".bak"))).isFalse();

        // lock and reload updated
        keyManager.lock();
        keyManager.unlock(NEW_PASSWORD);
        assertThat(keyManager.getPrivateKey("foo")).isEqualTo(data);
    }

    @Test
    void setAndGetPrivateKey() {
        keyManager.unlock(PASSWORD);
        byte[] data = "data".getBytes(StandardCharsets.UTF_8);
        keyManager.setPrivateKey("foo", data);
        assertThat(keyManager.getPrivateKey("foo")).isEqualTo(data);
    }

    @Test
    void saveKeyStore() {
        keyManager.unlock(PASSWORD);
        keyManager.saveKeyStore();
        assertThat(Files.exists(Paths.get(TEMP_DIR_NAME, KS_FILE_NAME))).isTrue();
    }

    @Test
    void isLocked() {
        assertThat(keyManager.isLocked()).isTrue();
        keyManager.unlock(PASSWORD);
        assertThat(keyManager.isLocked()).isFalse();
    }
}