package at.qubic.shell.wallet;

import org.junit.jupiter.api.Test;
import at.qubic.shell.repository.KeyStoreRepository;
import at.qubic.shell.util.ConversionUtil;
import at.qubic.shell.util.EncryptionUtility;
import org.mockito.Mockito;

import javax.crypto.SecretKey;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.CertificateException;

import static org.assertj.core.api.Assertions.assertThat;
import static at.qubic.shell.wallet.KeyManager.*;
import static org.mockito.Mockito.*;

class KeyManagerTest {

    public static final String TEST_STORE_PASSWORD = "encoded-password";
    private static final char[] TEST_INPUT_PASSWORD = "password".toCharArray();
    public static final String TEST_SALT_PASSWORD = "saltPassword";
    public static final String TEST_HASHED_PASSWORD = "hashed";
    public static final String KEYSTORE_SALT = "keystoreSalt";
    public static final byte[] INITIAL_SALT = {1, 2, 3};
    private final KeyStoreRepository keyStoreRepository = mock();
    private final EncryptionUtility encryptionUtility = mock();
    private final KeyManager keyManager = new KeyManager(keyStoreRepository, encryptionUtility);

    @Test
    void unlock_givenNoSaltYet_thenStoreSalt() throws Exception {
        KeyStore keyStore = mockKeyStore();
        assertThat(keyManager.isLocked()).isTrue();

        keyManager.unlock(TEST_INPUT_PASSWORD);
        assertThat(keyManager.isLocked()).isFalse();

        verify(keyStore).setEntry(eq(ALIAS_SALT), any(KeyStore.Entry.class), any(KeyStore.PasswordProtection.class));
        verify(keyStore).getEntry(eq(ALIAS_SALT), any(KeyStore.PasswordProtection.class));
    }

    @Test
    void unlock_givenSalt_thenRetrieveStored() throws Exception {
        KeyStore keyStore = mockKeyStore();
        SecretKey secretKey = createSecretKey(KEYSTORE_SALT.getBytes(StandardCharsets.UTF_8));
        KeyStore.SecretKeyEntry entry = new KeyStore.SecretKeyEntry(secretKey);
        keyStore.setEntry(ALIAS_SALT, entry, passwordProtectionForStoring(TEST_SALT_PASSWORD.toCharArray()));
        Mockito.reset(keyStore); // to very there is no set
        assertThat(keyManager.isLocked()).isTrue();

        keyManager.unlock(TEST_INPUT_PASSWORD);
        assertThat(keyManager.isLocked()).isFalse();

        verify(keyStore, never()).setEntry(eq(ALIAS_SALT), any(KeyStore.Entry.class), any(KeyStore.PasswordProtection.class));
        verify(keyStore).getEntry(eq(ALIAS_SALT), any(KeyStore.PasswordProtection.class));
    }

    @Test
    void lock() throws Exception {
        KeyStore keyStore = mockKeyStore();
        keyManager.unlock(TEST_INPUT_PASSWORD);
        assertThat(keyManager.isLocked()).isFalse();

        keyManager.lock();
        verify(keyStoreRepository).save(eq(keyStore), any(char[].class)); // for some reasone matches cleared password
        assertThat(keyManager.isLocked()).isTrue();
    }

    @Test
    void save() throws Exception {
        KeyStore keyStore = mockKeyStore();
        keyManager.unlock(TEST_INPUT_PASSWORD);

        keyManager.saveKeyStore();

        verify(keyStoreRepository).save(keyStore, TEST_STORE_PASSWORD.toCharArray());
    }

    @Test
    void setAndGetPrivateKey() throws Exception {
        KeyStore keyStore = mockKeyStore();
        keyManager.unlock(TEST_INPUT_PASSWORD);

        byte[] privateKey = {1, 2, 3};
        byte[] encrypted = {4, 5, 6};
        when(encryptionUtility.encryptData(any(char[].class), any(byte[].class), eq(privateKey))).thenReturn(encrypted);
        when(encryptionUtility.decryptData(any(char[].class), any(byte[].class), eq(encrypted))).thenReturn(privateKey);

        keyManager.setPrivateKey("foo", privateKey);
        verify(keyStore).setEntry(eq("foo"), any(KeyStore.Entry.class), any(KeyStore.PasswordProtection.class));
        byte[] result = keyManager.getPrivateKey("foo");
        assertThat(result).isEqualTo(privateKey);
    }

    private KeyStore mockKeyStore() throws KeyStoreException, IOException, NoSuchAlgorithmException, CertificateException {
        when(keyStoreRepository.loadPasswordSalt()).thenReturn(INITIAL_SALT);
        byte[] hashedPassword = TEST_HASHED_PASSWORD.getBytes(StandardCharsets.UTF_8);
        when(encryptionUtility.hashPassword(ConversionUtil.charsToBytes(TEST_INPUT_PASSWORD), INITIAL_SALT)).thenReturn(hashedPassword);
        byte[] storePassword = ConversionUtil.charsToBytes(TEST_STORE_PASSWORD.toCharArray());
        when(encryptionUtility.deriveKeyBase64(eq(hashedPassword), eq(INITIAL_SALT), eq(KeyManager.INFO_KEYSTORE))).thenReturn(storePassword);
        when(encryptionUtility.deriveKeyBase64(eq(hashedPassword), eq(INITIAL_SALT), eq(KeyManager.INFO_SALT))).thenReturn(TEST_SALT_PASSWORD.getBytes(StandardCharsets.UTF_8));
        byte[] keystoreSalt = KEYSTORE_SALT.getBytes();
        when(encryptionUtility.generateSalt()).thenReturn(keystoreSalt);
        when(encryptionUtility.deriveKeyBase64(eq(hashedPassword), eq(keystoreSalt), eq(KeyManager.INFO_ENTRY))).thenReturn("entryPassword".getBytes(StandardCharsets.UTF_8));
        when(encryptionUtility.deriveKeyBase64(eq(hashedPassword), eq(keystoreSalt), eq(KeyManager.INFO_ENCRYPTION))).thenReturn("encryptionPassword".getBytes(StandardCharsets.UTF_8));
        KeyStore keyStore = spy(KeyStore.getInstance(KeyStore.getDefaultType()));
        keyStore.load(null);
        when(keyStoreRepository.load(any(), eq(TEST_STORE_PASSWORD.toCharArray()))).thenReturn(keyStore);
        return keyStore;
    }


}