package at.qubic.vanity;

import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import at.qubic.vanity.keygen.VanityIdGenerator;
import org.springframework.boot.ApplicationArguments;
import org.springframework.boot.ApplicationRunner;
import org.springframework.boot.SpringApplication;
import org.springframework.boot.autoconfigure.SpringBootApplication;
import org.springframework.util.CollectionUtils;

import java.time.Duration;
import java.time.LocalDateTime;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

import static org.springframework.util.StringUtils.tokenizeToStringArray;

@SpringBootApplication
@RequiredArgsConstructor
@Slf4j
public class QubicVanityIdApplication implements ApplicationRunner {

    private final VanityIdGenerator keyGenerator;

    @Override
    public void run(ApplicationArguments args) {

        List<String> vanity = getVanitySearchValues(args.getOptionValues("vanity"));
        for (String searchString : vanity) {
            log.info("Estimated needed number of iterations for search string [{}]: [{}].", searchString, (long) Math.pow(26, StringUtils.length(searchString)));
        }

        if (!CollectionUtils.isEmpty(vanity)) {

            final long iterations = getIterations(args.getOptionValues("iterations"));
            final int cores = getCores(args.getOptionValues("cores"));
            final boolean matchEnd = args.containsOption("matchEnd");
            final boolean matchStart = !matchEnd || args.containsOption("matchStart"); // default
            String matching = String.format("%s", matchStart ? (matchEnd ? "starting/ending" : "starting") : "ending");

            log.info("Generating vanity IDs {} with the following [{}] strings: {}.", matching, vanity.size(), vanity);
            log.info("Maximum number of iterations: [{}].", iterations);
            log.info("Using [{}] cores.", cores);
            log.info("Started at [{}].", LocalDateTime.now());
            final long start = System.nanoTime();
            keyGenerator.generate(cores, iterations, matchStart, matchEnd, vanity.toArray(new String[0]))
                    .doOnNext(match -> log.info("[{}] => [{}]", match.getT1(), match.getT2()))
                    .doOnError(t -> log.error(t.toString(), t))
                    .doOnComplete(() -> logCompleted(start, iterations))
                    .blockLast();

        } else {

            log.warn("Invalid arguments: {}.", (Object) args.getSourceArgs());
            log.info("Usage: java -jar ... --vanity=<list,of,names> [--iterations=<no-of-iterations>] [--cores=<no-of-used-cores>] [--matchStart] [--matchEnd]");

        }

    }

    private static int getCores(List<String> cores) {
        return CollectionUtils.isEmpty(cores) ? Math.max(2, (Runtime.getRuntime().availableProcessors() / 2) - 1) : Integer.parseInt(cores.getFirst());
    }

    private static List<String> getVanitySearchValues(List<String> vanity) {
        return CollectionUtils.isEmpty(vanity) ? List.of() : vanity.stream()
                .flatMap(s -> Arrays.stream(tokenizeToStringArray(s, ", ", true, true)))
                .collect(Collectors.toList());
    }

    private static long getIterations(List<String> iterationValues) {
        return CollectionUtils.isEmpty(iterationValues) ? 100_000_000 : Long.parseLong(iterationValues.getFirst());
    }

    private static void logCompleted(long start, long iterations) {
        long duration = System.nanoTime() - start;
        Duration d = Duration.ofNanos(duration);
        log.info("End at [{}]", LocalDateTime.now());
        log.info("Iterations: [{}]. Duration: {}s. {}ns/it", iterations, d.toSeconds(), duration / iterations);
        log.info("Completed.");
    }

    public static void main(String[] args) {
        SpringApplication.run(QubicVanityIdApplication.class, args);
    }

}
