package at.qubic.vanity.crypto;

import lombok.extern.slf4j.Slf4j;
import at.qubic.api.crypto.FourQ;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;

@Slf4j
public class IdentityUtil {

    public final byte[] seedBytes;
    private final byte[] idBytes;
    private final byte[] subSeedBytes;
    private final byte[] privateKeyBytes;
    private final byte[] publicKeyBytes;
    private final byte[] checksumBytes;
    private final FourQ fourQ;

    public IdentityUtil(FourQ fourQ) {
        this.fourQ = fourQ;
        seedBytes = new byte[55];
        idBytes = new byte[60];
        subSeedBytes = new byte[32];
        privateKeyBytes = new byte[32];
        publicKeyBytes = new byte[32];
        checksumBytes = new byte[4];
    }

    public byte[] identityFromSeed(String seed) {
        for (int i = 0; i < seedBytes.length; i++) {
            seedBytes[i] = (byte) (seed.charAt(i) - 'a');
        }
        KangarooTwelve.hash(seedBytes, subSeedBytes); // subseed
        KangarooTwelve.hash(subSeedBytes, privateKeyBytes); // private key
        fourQ.ecc_mul_fixed(privateKeyBytes, publicKeyBytes); // public key
        ByteBuffer publicKeyBuf = ByteBuffer.wrap(publicKeyBytes);
        publicKeyBuf.order(ByteOrder.LITTLE_ENDIAN);

        for (int i = 0; i < 4; i++) {
            // we need unsigned longs, but they are not available in java natively. negative longs won't work.
            // we can either use the Long.remainderUnsigned/divideUnsigned methods or BigInteger, but it's
            // ugly to go via string constructor new BigInteger(Long.toUnsignedString(publicKeyBuf.getLong()))
            long publicKeyFragment = publicKeyBuf.getLong();

            for (int j = 0; j < 14; j++) {
                long character = Long.remainderUnsigned(publicKeyFragment, 26);
                idBytes[i*14+j] = (byte) (character + 'A');
                publicKeyFragment = Long.divideUnsigned(publicKeyFragment, 26);
            }
        }

        // checksum
        KangarooTwelve.hash(publicKeyBytes, checksumBytes);
        // packing an array of 4 bytes to an int, little endian,
        int checksumInt = (((checksumBytes[3] & 0xFF) << 24) |
                ((checksumBytes[2] & 0xFF) << 16) |
                ((checksumBytes[1] & 0xFF) << 8) |
                (checksumBytes[0] & 0xFF)) // convert to int
                & 0x3FFFF; // then mask for processing
        // four characters are stored in the integer
        for (int i = 0; i < 4; i++) {
            idBytes[56 + i] = (byte) ((checksumInt % 26) + 'A');
            checksumInt /= 26; // next character
        }

        return idBytes;
    }

}
