package at.qubic.vanity.keygen;

import io.netty.util.AsciiString;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import at.qubic.api.crypto.FourQ;
import at.qubic.vanity.crypto.StandardSeedGenerator;
import at.qubic.vanity.crypto.IdentityUtil;
import at.qubic.vanity.crypto.RandomGeneratorFactory;
import at.qubic.vanity.crypto.SeedGenerator;
import reactor.core.publisher.Flux;
import reactor.core.publisher.SynchronousSink;
import reactor.core.scheduler.Schedulers;
import reactor.util.function.Tuple2;
import reactor.util.function.Tuples;

import java.time.Duration;
import java.time.Instant;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Function;
import java.util.function.Predicate;

@Slf4j
public class VanityIdGenerator {

    public static final long LOG_COUNT = 10_000_000;
    public static final long GC_COUNT = LOG_COUNT * 10;
    private final Map<Long, SeedGenerator> generators = new ConcurrentHashMap<>(30);
    private final Map<Long, IdentityUtil> identityUtils = new ConcurrentHashMap<>(30);
    private long count;
    private final FourQ fourQ;
    private Instant lastUpdateTime = Instant.EPOCH;
    protected final RandomGeneratorFactory randomGeneratorFactory;

    public VanityIdGenerator(RandomGeneratorFactory randomGeneratorFactory, FourQ fourQ) {
        this.randomGeneratorFactory = randomGeneratorFactory;
        this.fourQ = fourQ;
    }

    public Flux<Tuple2<String, CharSequence>> generate(int threads, long maxIterations, boolean matchStart, boolean matchEnd, String... match) {
        return Flux.generate((SynchronousSink<Long> sink) -> sink.next(++count))
                .doFirst(() -> lastUpdateTime = Instant.now())
                .take(maxIterations, true)
                .parallel(threads)
                .runOn(Schedulers.parallel())
                .map(this::generateIdentity)
                .filter(matchString(matchStart, matchEnd, match))
                .sequential()
                .publishOn(Schedulers.single());
    }

    private Tuple2<String, CharSequence> generateIdentity(long count) {
        final SeedGenerator generator = generators.computeIfAbsent(Thread.currentThread().threadId(), createSeedGenerator());
        final IdentityUtil identityUtil = identityUtils.computeIfAbsent(Thread.currentThread().threadId(), x -> new IdentityUtil(fourQ));


        if (count % LOG_COUNT == 0) {
            Instant now = Instant.now();
            log.info("[{}] iterations. Current speed: {} it/s", count, Math.round(((double) LOG_COUNT / Duration.between(lastUpdateTime, now).toMillis()) * 1000));
            lastUpdateTime = now;
            if (count % GC_COUNT == 0) {
                System.gc(); // otherwise we run out of memory
            }
        }

        final String seed = generator.seed();
        AsciiString idString = new AsciiString(identityUtil.identityFromSeed(seed));
        return Tuples.of(seed, idString);
    }

    protected Function<Long, SeedGenerator> createSeedGenerator() {
        return x -> new StandardSeedGenerator(randomGeneratorFactory.create());
    }

    private static Predicate<Tuple2<String, CharSequence>> matchString(boolean matchStart, boolean matchEnd, String[] match) {
        return tuple -> (matchStart && StringUtils.startsWithAny(tuple.getT2(), match))
                || (matchEnd && StringUtils.endsWithAny(tuple.getT2(), match));
    }

}
